//////////////////////////////////////////////////////////////
//
// Cursor.CPP
//
// This is the source file for the NewGPE display driver
// class that contains the functionality related to cursors.
//
//////////////////////////////////////////////////////////////

///////////////////////////////////
//
// Includes
//

#include <windows.h>
#include <types.h>
#include "NewGPE.h"

///////////////////////////////////
//
// Constants
//

///////////////////////////////////
//
// Globals
//

///////////////////////////////////
//
// Function Prototypes
//

///////////////////////////////////
//
// Functions
//

//////////////////////////////////////////////////////////////
//
// NewGPE::SetPointerShape
//
// Set the cursor. This function overrides the base GPE class
// pure virtual function.
//
// PARAMETERS
// pMask        cursor mask
// pColorSurf   surface for color cursor
// xHot         cursor hotspot x coordinate
// yHot         cursor hotspot y coordinate
// cx           cursor width
// cy           cursor height
//
// RETURNS
// S_OK if successful, otherwise some error code
//
// NOTES
// Default functionality is already implemented in the
// form of a software cursor. Only override or change
// to add hardware cursor support.
//
// The default software cursor implementation does not
// include support for color cursors.
//
#ifdef JEFF
SCODE NewGPE::SetPointerShape(
                        GPESurf *pMask,
                        GPESurf *pColorSurf,
                        int xHot,
                        int yHot,
                        int cx,
                        int cy)
{
    SCODE scRet = S_OK;

    if (!m_bHWCursor) {
        // default software cursor support

        // hide old cursor if showing
        if (m_bCursorShowing) {
            // hide cursor
            m_bCursorShowing = FALSE;
            // restore background
            CursorBlt(m_pBkgndSurf, NULL, m_pPrimarySurface, &m_rectCursor);
        }
        // create or resize buffers if necessary (saves, masks, temps)
        if ((m_sizeCursor.cx!=cx) || (m_sizeCursor.cy!=cy)) { // surface sizes don't match
            // delete any surface
            if (m_pAndMaskSurf!=NULL)
                delete m_pAndMaskSurf;
            if (m_pXorMaskSurf!=NULL)
                delete m_pXorMaskSurf;
            if (m_pBkgndSurf!=NULL)
                delete m_pBkgndSurf;
            if (m_pNewBkgndSurf!=NULL)
                delete m_pNewBkgndSurf;
            if (m_pTempSurf!=NULL)
                delete m_pTempSurf;

            // create surfaces for software cursor
            // (keep at same color depth same as display for better performance)
            AllocSurface(&m_pAndMaskSurf, cx, cy, m_pMode->format, GPE_PREFER_VIDEO_MEMORY);
            AllocSurface(&m_pXorMaskSurf, cx, cy, m_pMode->format, GPE_PREFER_VIDEO_MEMORY);
            AllocSurface(&m_pBkgndSurf, cx, cy, m_pMode->format, GPE_PREFER_VIDEO_MEMORY);
            AllocSurface(&m_pNewBkgndSurf, cx, cy, m_pMode->format, GPE_PREFER_VIDEO_MEMORY);
            AllocSurface(&m_pTempSurf, cx, cy, m_pMode->format, GPE_PREFER_VIDEO_MEMORY);

            m_sizeCursor.cx = cx;
            m_sizeCursor.cy = cy;
        }

        // copy cursor masks (AND, XOR)
        //
        RECT rectMaskSrc;
        rectMaskSrc.top = 0;
        rectMaskSrc.left = 0;
        rectMaskSrc.bottom = cy;
        rectMaskSrc.right = cx;
        CursorBlt(pMask, &rectMaskSrc, m_pAndMaskSurf, NULL);
        rectMaskSrc.top = rectMaskSrc.bottom;
        rectMaskSrc.bottom = rectMaskSrc.top + cy;
        CursorBlt(pMask, &rectMaskSrc, m_pXorMaskSurf, NULL);

        // save hot spot info
        m_nHotX = xHot;
        m_nHotY = yHot;
    }
    else {
        // hardware cursor
        //
        // NOTE:
        // Not implemented, device specific support needs to be added.
        //
        RETAILMSG(1, (TEXT("NewGPE::SetPointerShape ERROR - hardware cursors not supported!\r\n")));
        scRet = SC_NOTIMPL;
    }

    return scRet;
}

//////////////////////////////////////////////////////////////
//
// NewGPE::MovePointer
//
// Move the cursor. This function overrides the base GPE class
// pure virtual function.
//
// PARAMETERS
// x            cursor position x coordinate
// y            cursor position y coordinate
//
// RETURNS
// S_OK if successful, otherwise some error code
//
// NOTES
// Default functionality is already implemented in the
// form of a software cursor. Only override or change
// to add hardware cursor support.
//
// The default software cursor implementation does not
// include support for color cursors.
//
// If system performance is sufficient, this function
// may be simplified.
//
SCODE GPEMultiMon::MovePointer(int x, int y)
{
    SCODE scRet = S_OK;

    // Algorithm for software cursor:
    // if hide cursor
    //   restore background
    // else if positions overlap
    //   save new background
    //   copy overlap old background to new background
    //   build new cursor image (with new background)
    //   wait for vertical blank
    //   restore old background (hide old cursor)
    //   write new cursor
    // else - positions don't overlap or cursor previously hidden
    //   save new background
    //   build new cursor image (with new background)
    //   wait for vertical blank
    //   write new cursor
    //   if cursor previously shown
    //     restore old background (hide old cursor)

    if (!m_bHWCursor) {
        // software cursor, go ahead and handle it
        if ((x==-1) || (y==-1)) { // hide cursor
            if (m_bCursorShowing) { // cursor is showing, so ok to hide
                m_bCursorShowing = FALSE;
                // hide cursor by restoring background
                CursorBlt(m_pBkgndSurf, NULL, m_pPrimarySurface, &m_rectCursor);
            }
        }
        else {
            // get rectangle for new cursor
            RECT rectNew(x-m_nHotX, y-m_nHotY,
                x-m_nHotX+m_sizeCursor.cx, y-m_nHotY+m_sizeCursor.cy);
            RECT rectIntersect, rectOld;

            if (m_bCursorShowing && ::IntersectRect(&rectIntersect, &rectNew, &m_rectCursor)) {
                // save new background
                CursorBlt(m_pPrimarySurface, &rectNew, m_pNewBkgndSurf, NULL);
                // copy overlap of old background to new background
                if (rectNew.left>=m_rectCursor.left) { // cursor in positive x-direction
                    rectOld.left = rectIntersect.left - m_rectCursor.left;
                    rectOld.right = rectIntersect.right - m_rectCursor.left;
                    rectIntersect.left = 0;
                    rectIntersect.right = rectOld.right - rectOld.left;
                }
                else { // cursor in negative x-direction
                    rectIntersect.left -= m_rectCursor.left;
                    rectIntersect.right -= m_rectCursor.left;
                    rectOld.left = 0;
                    rectOld.right = rectIntersect.right - rectIntersect.left;
                }
                if (rectNew.top>=m_rectCursor.top) { // cursor in positive y-direction
                    rectOld.top = rectIntersect.top - m_rectCursor.top;
                    rectOld.bottom = rectIntersect.bottom - m_rectCursor.top;
                    rectIntersect.top = 0;
                    rectIntersect.bottom = rectOld.bottom - rectOld.top;
                }
                else { // cursor in negative y-direction
                    rectIntersect.top -= m_rectCursor.top;
                    rectIntersect.bottom -= m_rectCursor.top;
                    rectOld.top = 0;
                    rectOld.bottom = rectIntersect.bottom - rectIntersect.top;
                }
                CursorBlt(m_pBkgndSurf, &rectOld, m_pNewBkgndSurf, &rectIntersect);

                // build new cursor image (with new background)
                CursorBlt(m_pNewBkgndSurf, NULL, m_pTempSurf, NULL);
                CursorBlt(m_pAndMaskSurf, NULL, m_pTempSurf, NULL, SRCAND);
                CursorBlt(m_pXorMaskSurf, NULL, m_pTempSurf, NULL, SRCXOR);

                // wait for vertical blank before drawing
                WaitVBlank();
                // now restore old background
                CursorBlt(m_pBkgndSurf, NULL, m_pPrimarySurface, &m_rectCursor);
                // now draw cursor
                CursorBlt(m_pTempSurf, NULL, m_pPrimarySurface, &rectNew);
            }
            else {
                // new and previous cursor positions don't overlap or
                // cursor was previously hidden

                // save new background
                CursorBlt(m_pPrimarySurface, &rectNew, m_pNewBkgndSurf, NULL);
                // build new cursor image
                CursorBlt(m_pNewBkgndSurf, NULL, m_pTempSurf, NULL);
                CursorBlt(m_pAndMaskSurf, NULL, m_pTempSurf, NULL, SRCAND);
                CursorBlt(m_pXorMaskSurf, NULL, m_pTempSurf, NULL, SRCXOR);

                // wait for vertical blank before drawing
                WaitVBlank();
                // now draw cursor
                CursorBlt(m_pTempSurf, NULL, m_pPrimarySurface, &rectNew);

                if (m_bCursorShowing) {
                    // restore old background
                    CursorBlt(m_pBkgndSurf, NULL, m_pPrimarySurface, m_rectCursor);
                }
            }

            // update parameters for new cursor position
            m_bCursorShowing = TRUE; // set in case it wasn't before
            m_rectCursor = rectNew;
            GPESurf *pSurf = m_pBkgndSurf;
            m_pBkgndSurf = m_pNewBkgndSurf;
            m_pBkgndSurf = pSurf;
        }
    }
    else {
        // hardware cursor
        //
        // NOTE:
        // Not implemented, device specific support needs to be added.
        //
        RETAILMSG(1, (TEXT("NewGPE::MovePointer ERROR - hardware cursors not supported!\r\n")));
        scRet = SC_NOTIMPL;
    }

    return scRet;
}

//////////////////////////////////////////////////////////////
//
// NewGPE::CursorBlt
//
// Perform a blt operation for software cursor support.
//
// PARAMETERS
// pSrcSurf     source surface for blt
// pSrcRect     source rectangle for blt (NULL=use entire surface)
// pDstSurf     destination surface for blt
// pDstRect     destination rectangle for blt (NULL=use entire surface)
// dwRop        raster operation used for blt
//
// RETURNS
// S_OK if successful, otherwise some error code
//
// NOTES
//
SCODE GPEMultiMon::CursorBlt(
                        GPESurf *pSrcSurf,
                        RECT *pSrcRect,
                        GPESurf *pDstSurf,
                        RECT *pDstRect,
                        DWORD dwRop/* =SRCCOPY */)
{
    SCODE scRet;

    GPEBltParms bltParms;

    memset(&bltParms, 0, sizeof(bltParms));
    bltParms.pSrc = pSrcSurf;
    bltParms.pDst = pDstSurf;
    bltParms.prclSrc = pSrcRect;
    bltParms.prclDst = pDstRect;
    bltParms.rop4 = dwRop;

    // ### check for clipping to the source or destination surface
    RECT rectClipSrc, rectClipDst;
    if (bltParms.prclSrc) {
        RECT rectSrcSurf(0, 0, , rectDstSurf;
        rectSrcSurf.
        ::IntersectRect(&rectClipSrc, 
    }
    if (bltParms.prclDst) {
    }

    // perform blt operation
    scRet = BltPrepare(&bltParms);
    if ((scRet==S_OK) && bltParms.pBlt) {
        scRet = (bltParms.pBlt)(bltParms);

        BltComplete(&bltParms);
    }

    return scRet;
}
#endif