package keepalive.apps;

import keepalive.communication.Input;
import keepalive.communication.Output;
import keepalive.communication.Receiver;
import keepalive.communication.Sender;
import keepalive.socket.AbstractSocket;
import keepalive.ui.KeepAliveUI;
import java.io.IOException;

/**
 * This is the abstract "Application" class; subclasses are a Client application
 * and a Server application.
 *
 * @see keepalive.apps.Server
 * @see keepalive.apps.Client
 *
 * @author Joshua Fox
 */
abstract class App {
    private KeepAliveUI keepAliveUI ;
    private Input input;
    private Output output;

    /**
     *
     */
    protected App() {
         keepAliveUI = createUI();
     }

    /**
     * Using the socket, create Input and Output objects,
     * hook the Input and Output objects into their UIs,
     * start the Keep-Alive, and then wait for the
     * Input/Output to finish (i.e. for the application
     * to finish).
     *
     * @param sock Socket with which we connect to the other side.
     */
    protected void connect(AbstractSocket sock ) {
        try {
            input = new Input(sock, this.toString());
            input.setMessageDisplayer(getKeepAliveUI());
            Thread inputThread = new Thread(input, "InputThread-"  + this);

            output = new Output(sock, this.toString());

            getKeepAliveUI().setOutputObject(output);
            inputThread.start();

            sock.startKeepAlive(getReceiver(), getSender(), isServerSide());

            try {
                inputThread.join(); // Wait until input thread terminates.
            } catch (InterruptedException ie) {}

        }  finally {
            try {
                if (sock  != null) {
                    sock.close();
                }
            } catch (IOException ioe) {
                System.err.println("Exception in closing Socket in "+
                                                        this + ":" + ioe);
            }
        }
   }

   /**
    * Accessor method.
    *
    * @return The Sender through which this object sends messages. May return
    *     null.
    */
   public Sender getSender() {
       return output;
   }

   /**
    * Accessor method.
    *
    * @return The Receiver through which this object receives messages. May be
    *      null.
    */
   public Receiver getReceiver() {
       return input;
   }

   /**
    * Mutator method.
    *
    * @param keepAliveUI the user interface associated with this application.
    */
   protected void setKeepAliveUI(KeepAliveUI keepAliveUI) {
       this.keepAliveUI = keepAliveUI;
   }

   /**
    * Accessor method.
    * @return the user interface associated with this application.
    */
   public KeepAliveUI getKeepAliveUI() {
    return keepAliveUI;
   }

   /**
    * Indicates whether this is a server-side or client-side app.
    *
    * @return True iff this is a server-side app.
    */
   protected abstract boolean isServerSide();

   /**
    * The creation of the User Interface is delegated to subclasses
    * in this method.
    *
    * @return A new User Interface.
    */
   protected abstract KeepAliveUI createUI() ;
}
