package keepalive.apps;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;

/**
 * A utility class for reading command-line arguments with associated switches.
 *
 * @author Joshua Fox
 */
class ArgReader {

    /**
     * Get the argument following a given switch.
     *
     * @param args The command-line args.
     * @param switchKey The String following "-" which is used as a
     *     command-line switch.
     * @param defaultValue If the switch is not found, the default value to be
     *      returned instead of the actual command-line arg.
     */
    public static String getArgBySwitch(String [] args,
                                        String switchKey, String defaultValue) {
        for (int i =0; i < args.length; i++) {
            if (args[i].equals("-" +switchKey)) {
                if (i+1 <args.length) {
                     return args[i+1];
                }
            }
        }

        return defaultValue;
    }


    /**
     * Get the property from a propertiesy file with the given name.
     *
     * @param args The command-line arg.
     * @param workingDir The directory in which to look for the properties
     *     file.
     * @param switchKey The switchkey which indicates that the following
     *     command-line arg is the name of the properties file.
     * @param propertyKey The key, in the properties file, of the property we
     *     are looking for.
     * @param defaultPropertyValue If the property file is not found or does
     *     not contain the requested key the default value to return.
     * @return The desired properties value.
     */
    public static String getPropertyBySwitch(String [] args,
                                             String workingDir,
                                             String switchKey,
                                             String propertyKey,
                                             String defaultPropertyValue) {

        String fileName = getArgBySwitch(args, switchKey, "protocolNN.properties");
        File file = new File(workingDir, fileName);
        Properties prop = new Properties();

        try {
            prop.load(new FileInputStream(file));
        } catch (IOException ioe) {
            System.err.println("Could not load property file " + file);
        }

        return prop.getProperty(propertyKey, defaultPropertyValue);
    }


    /**
     * Show a "usage" String if the user requests it
     * by typing "?", "h", "help", "-?", "-h", "-help" as a command-line arg.
     *
     * @param args The command-line args.
     * @param usage The string to display.
     */
    public static void showUsageIfRequested(String [] args, String usage) {
        String [] helpArgs = {"?", "h", "help"};
        for (int i=0; i< args.length; i++) {
            String arg = args[i];
            for (int j=0 ; j < helpArgs.length; j++) {
                if (arg.equalsIgnoreCase(helpArgs[j]) || 
                    arg.equalsIgnoreCase("-" + helpArgs[j])) {
                    System.err.println(usage);
                    System.exit(0);
                }
            }
        }
    }
}

