package keepalive.apps;

import java.io.IOException;
import java.net.InetAddress;
import keepalive.socket.SocketFactory;
import keepalive.socket.AbstractSocket;
import keepalive.ui.ClientKeepAliveUI;
import keepalive.ui.KeepAliveUI;
import keepalive.ui.KeepAliveFrame;

/**
 * A Client application. Can run with simulated sockets or real TCP/IP sockets.
 *
 * @author Joshua Fox
 */
public class Client extends App {
    private String serverName;

    /**
     * Start the client application.
     *
     * @param serverName Name of server host to which we will connect.
     */
    public Client(String serverName) {
        super();
        this.serverName = serverName;
    }

    /**
    * Create the user interface
    * @return the user interface.
    */
    protected KeepAliveUI createUI() {
      KeepAliveUI ui = new ClientKeepAliveUI(this);
      ui.setKeepAliveDescription(SocketFactory.getClientKeepAlive());
      return ui;
    }


    /**
     * Connect to the server.
     */
    public void openConnectionToServer() {
        new OpenConnectionToServerThread().start();
    }


    /**
     *
     */
    public String toString() {
        return "Client";
    }

    /**
     * Indicates that this application is a client application.
     * Abstract in superclass.
     * @return false to indicate that this is a client application.
     */
    protected boolean isServerSide() {
        return false;
    }


   /**
    * Note that more than one of these Threads may be started
    * and more than one socket may be opened.
    * This implementation of the Client needs only one socket,
    * and so any previously created sockets will be ignored.
    */
    private class OpenConnectionToServerThread extends Thread {
        /**         
        *
         */
        OpenConnectionToServerThread() {
            super("OpenConnectionToServerThread");
        }

        /**        
         *
         */
        public void run() {
            try {
                InetAddress addr = InetAddress.getByName(serverName);
                AbstractSocket sock = SocketFactory.createSocket(addr);
                connect(sock);
            } catch (IOException ioe) {
                System.err.println(
                           "Exception in OpenConnectionToServerThread: " + ioe);
            }
        }
    }


    /**
     * Run the Client application with real TCP/IP sockets.
     *
     * @exception IOException if an I/O error occurs.
     * @param args command line arguments; use -help to get usage description.
     */
    public static void main(String[] args) throws IOException {
        ArgReader.showUsageIfRequested(args,
            "Usage: java Client [-s serverAddress (default localhost)] " + 
            "[-k protocolFile] [-d workingDir]\n");

        String serverName = ArgReader.getArgBySwitch(args, "s", "localhost");

        String workingDir = ArgReader.getArgBySwitch(args, "d", ".");
        String keepAliveClassName = 
                    ArgReader.getPropertyBySwitch(args, workingDir, 
                            "k", "client", "keepalive.keepalive.NullKeepAlive");
        SocketFactory.setClientKeepAlive(keepAliveClassName);
        SocketFactory.setSocketType(SocketFactory.SOCKET);

        Client client = new Client(serverName);

        new KeepAliveFrame(client.getKeepAliveUI(), false);
    }
}
