package keepalive.apps;

import java.io.IOException;
import keepalive.socket.ServerSocketIfc;
import keepalive.socket.SocketFactory;
import keepalive.socket.AbstractSocket;
import keepalive.ui.ServerKeepAliveUI;
import keepalive.ui.KeepAliveUI;
import keepalive.ui.KeepAliveFrame;


/**
 * A Server application. Can run with simulated sockets or real TCP/IP sockets.
 *
 * @author Joshua Fox
 */
public class Server extends App implements Runnable {
    /**
     * Run the application.
     */
    public Server() {
        new Thread(this).start(); // set the server to listening
    }

    /**
     * Create the user interface.
     */
    protected KeepAliveUI createUI () {
      KeepAliveUI ui = new ServerKeepAliveUI();
      ui.setKeepAliveDescription(SocketFactory.getServerKeepAlive());
      return ui;
    }


    /**
     * Listen for connections.
     */
    public void run() {
         listen();
    }

    /**
     * Listen for incoming socket connections.
     */
    public void listen() {
        ServerSocketIfc serverSock = null;

        try {
            serverSock = SocketFactory.createServerSocket();
            while (!Thread.interrupted()) {
                AbstractSocket sock  = serverSock.accept();
                new ConnectionHandlerThread(sock).start();
            }
        } catch (IOException ioe) {
            System.err.println("Exception in Server listening: " + ioe);
        } finally {
            if (serverSock != null) {
                try {
                    serverSock.close();
                } catch (IOException ioe2) {
                }
            }
        }
    }


    /**
    */
    public String toString() {
        return "Server";
    }


    /**
     * Indicates that this application is a server application.
     * Abstract in superclass.
     */
    protected boolean isServerSide() {
        return true;
    }


    /**
     * Run the Server application with real TCP/IP sockets.
     * @param args command line arguments; use -help to get usage description.
     * @exception IOException if an I/O error occurs in binding the ServerSocket.
     */
    public static void main(String[] args)  throws IOException {
        ArgReader.showUsageIfRequested(args,
            "Usage: java Server [-k protocolFile] [-d workingDirectory]\n");
        String workingDir = ArgReader.getArgBySwitch(args, "d", ".");
        String keepAliveClassName =
                ArgReader.getPropertyBySwitch(args, workingDir, "k",
                                "server", "keepalive.keepalive.NullKeepAlive");
        SocketFactory.setServerKeepAlive(keepAliveClassName);
        SocketFactory.setSocketType(SocketFactory.SOCKET);

        Server server = new Server();
        new KeepAliveFrame(server.getKeepAliveUI(), true);
    }


    /**
     * When a Client requests a connection with this Server, a
     * new socket and a new ConnectionHandlerThread are spun off
     * to handle the incoming socket.
     */
    private class ConnectionHandlerThread extends Thread {
        /**         
         */
        private AbstractSocket spunOffSocket;

        /**         
         * @param sock
         */
        ConnectionHandlerThread(AbstractSocket sock) {
            super("ConnectionHandlerThread-" + sock);
            spunOffSocket = sock;
        }


        /**         
         */
        public void run() {
            connect(spunOffSocket);
        }
    }
}
