package keepalive.communication;

import java.io.BufferedReader;
import java.io.PrintWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import keepalive.communication.Receiver;
import keepalive.socket.AbstractSocket;
import keepalive.ui.MessageDisplayer;

/**
 * Accept an incoming stream of characters.
 */
public class Input implements Runnable, Receiver {
    private BufferedReader bufReader;

    /**
     * The name of the "user" of this class; for debugging purposes.
     * In our case, "Server" or "Client".
     */
    private String participantName;

    /**
     * Knows how to display incoming messages.
     */
    private MessageDisplayer ui;

    /**
     * The time at which the last incoming message was received.
     */
    private long timeStamp;

    /**
     * The call-back object that will be notified whenever a message arrives.
     */
    private MessageListener messageListener;

    /**
     * Create a new Input object.
     * @param sock The socket through which we will receive the incoming stream.
     * @param participantName  The name of the "user" of this class; for 
     * debugging purposes.
     */
    public Input(AbstractSocket sock, String participantName ) {
        timeStamp = System.currentTimeMillis();
        this.participantName = participantName;

        // Set "null" MessageListener and MessageDisplayer; these are
        // placeholders until someone sets the "real" ones.
        setMessageListener(new MessageListener() {
            public boolean messageReceived(String s) {
                 return false;
            }
        });

        setMessageDisplayer(new MessageDisplayer() {
             public void display(String s1, String s2, boolean  b) {
                /*Do nothing*/
             }
        });

        try {
            bufReader = new BufferedReader(
                                  new InputStreamReader(sock.getInputStream()));
        } catch (IOException ioe) {
            System.err.println("Exception in creating bufferedReader in " + 
                                                            this + "; " + ioe );
        }
    }


    /**
     * Receive messages line by line; pass each incoming message 
     * to the MessageListener
     */
    public void run() {
        try {
            String line;
            while ((line=bufReader.readLine()) != null) {
                timeStamp = System.currentTimeMillis();
                boolean processed = messageListener.messageReceived(line);
                // In a real application could choose to ignore any message 
                // which was processed by the MessageListener--these are 
                // Keep-Alive-specific messages not relevant to the remainder of 
                // the application. Here we choose to display all messages 
                // to show how the Keep-Alive works.
                ui.display(participantName, line, false);
            }
        } catch (IOException ioe) {
             System.err.println("Exception in Input for " + participantName + 
                                                                 ": " + ioe  );
        }
    }

    /**
     * Get time stamp for incoming messages. From Receiver interface
     * @return the time stamp of the last message received.
     */
    public long lastTimeStamp() {
        return timeStamp;
    }

    /**
     * Set the call-back object for received messages. Mutator method.
     * From Receiver interface,
     */
    public void setMessageListener(MessageListener listener) {
        this.messageListener = listener;
    }

    /**
     * Set the object that displays messages. Mutator method.
     */
    public void setMessageDisplayer(MessageDisplayer ui) {
        this.ui = ui;
    }


    /**
     */
    public String toString() {
         return "Input-" + participantName;
    }
}
