package keepalive.communication;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import keepalive.socket.AbstractSocket;
import keepalive.ui.MessageDisplayer;

/**
 * Send an outgoing stream of characters.
 */
public class Output implements Sender {
    private PrintWriter socketWriter;
    private MessageDisplayer ui;
    private AbstractSocket sock;

    /**
     * The name of the "user" of this class; for debugging purposes.
     * In our case, "Server" or "Client".
     */
    private String participantName;

    private int messageCount= 0;
    private long timeStamp;

    /** 
     * Create a new Output object
     * @param sock Socket through which we send the outgoing stream.
     * @param participantName The name of the "user" of this class; for
     *     debugging purposes.
     */
    public Output(AbstractSocket sock, String participantName ) {
        timeStamp = System.currentTimeMillis();
        setMessageDisplayer(new MessageDisplayer() {
            // simple default message displayer:
            public void display(String participantName, 
                                              String message, boolean sending) {
               System.out.println(participantName + " " + 
                                  (sending?"sent":"received") + ": " + message);
            }
        }); // default UI
        this.participantName  = participantName;

        try {
            this.sock  = sock;
            socketWriter = new PrintWriter( 
                    new BufferedWriter(  
                        new OutputStreamWriter( sock.getOutputStream())), true);
        } catch (IOException ioe) {
            System.err.println("Exception in Output constructor: " +ioe);
        }
    }

    /**
     * Send a message with a serial number.
     *
     * @exception IOException if if an I/O error occurs.
     */
    public void sendMessage() throws IOException{
        String sendThis = "Message " + ++messageCount;
        send(sendThis);
    }


    /**
     * Send a message with the given String. From Sender interface.
     *
     * @param sendThis message to be sent.
     * @exception IOException if an I/O error occurs.
     */
    public synchronized void send(String sendThis) throws IOException {
        socketWriter.println(sendThis);

        // PrintWriter doesn't throw IOException; use checkError() instead
        if (socketWriter.checkError()) {
            ui.display(participantName, "Error in sending " + sendThis, true);
            throw new IOException(this.toString());
        } else {
            ui.display(participantName, sendThis, true);
            timeStamp = System.currentTimeMillis();
        }
    }


    /**
     * Close socket. From Sender interface
     */
    public void close() {
        try {
            sock.close();
        } catch (IOException ioe) {
            System.err.println("Exception in Output::close() " + ioe);
        }
    }


    /**
     * Get time that last message was sent. From Sender interface.
     */
    public long lastTimeStamp() {
        return timeStamp;
    }

    /**
    */
    public String toString() {
         return "Output-" + participantName;
    }

    /**
     * Set the object that will display all outgoing messages as
     * they are sent.
     *
     * @param ui User interface which will display messages.
     */
    public void setMessageDisplayer(MessageDisplayer ui) {
        this.ui = ui;
    }
}
