package keepalive.keepalive;

import java.io.IOException;

import keepalive.communication.Sender;
import keepalive.communication.Receiver;


/**
 * If no incoming messages have been received in a while, the remote
 * application or host may be dead or the connection may have been lost,
 * so send out an Are-You-There. If the connection has been lost, we will
 * get an IOException. If the connection is alive, the Are-You-There serves
 * as a trigger to make the other side send something back. If, despite this,
 * we have received nothing in a long time, then  we close the socket.
 *
 * @author Joshua Fox
 */
public class AreYouThere extends KeepAlive {
    /**
     * The amount of time we will allow an incoming connection to be 
     * silent before we kill it.
     */
    private static final long maxPermittedIdleTime = 9000L;
    /**
     * The amount of time we wait on a silent incoming connection before 
     * sending out an Are-You-There.
     */
    private static final long idleTimeBeforeAYT  = 
                                           (long)(0.67d * maxPermittedIdleTime);

    /**
     * Create a new Keep-Alive object.
     * @param receiver
     * @param sender
     */
    public AreYouThere(Receiver receiver, Sender sender) {
        super(receiver, sender);
    }

    /**
     * Create a new Keep-Alive object.
     * Used for dynamic classloading.
     */
    public AreYouThere() {
    }


    /**
     *
     */
    public void run() {
        while (!endRequested()) {
            sendAYTIfIdleForTooLong();
            if (closeIfIdleForTooLong()) {
                break;
            }


            try {
                Thread.sleep(1000L);
            } catch (InterruptedException ie) {
            }
        }
    }


     /**
      * Close the connection if nothing has been received for a long time.
      * The length of time since any data was received is the only criterion
      * for closing. There is no specific test for receiving the I-Hear-You.
      */
     private boolean closeIfIdleForTooLong() {
        if (getSilentTimeForIncoming() > maxPermittedIdleTime) {
            getSender().close();
            return true;
        }

        return false;
    }


  /**
   * The Are-You-There sends a message at 2/3 of the time period
   * used by closeIfIdleForTooLong(). A I-Hear-You (which is actually
   * just a dummy packet) should come back almost immediately,
   * thus updating the timestamp.
   */
  private void sendAYTIfIdleForTooLong() {
      if (getSilentTimeForIncoming() > idleTimeBeforeAYT) {

         try {
            getSender().send(ARE_YOU_THERE);
         } catch (IOException ioe) {
              getSender().close();
         }
      }
   }
}