package keepalive.keepalive;

import java.io.IOException;

import keepalive.communication.Receiver;
import keepalive.communication.Sender;


/**
 * Sends a "heartbeat" which is echoed by HeartbeatServer,
 * closing the socket if there is silence for a suspicious time period.
 * <BR>
 * Combines ActiveKeepAlive (periodically sends a message)
 * with PassiveKeepAlive (listens for suspicious silence in
 * incoming data) and AreYouThere (expects a response to
 * its messages). Unlike PassiveKeepAlive, but like AreYouThere
 * it will automatically close the socket when the timeout period
 * passes with no incoming data.
 * <BR>
 * For an alternate implementation, not as simple but more conservative of
 * bandwidth, see the JavaDocs for HeartbeatServer messageReceived(String s)
 * @see keepalive.keepalive.HeartbeatServer#messageReceived(String s)
 *
 * @author Joshua Fox
 */
public class HeartbeatClient extends KeepAlive {
    /**
     * Allow a maximum idle time that is a multiple of the time
     * between probes, to allow for network latency.
     */
    private static final long maxPermittedIdleTime =    
                                                6 * TIME_BETWEEN_ACTIVE_PROBES;

    /**
     * Create a new Keep-Alive.
     * @param receiver
     * @param sender
     */
    public HeartbeatClient(Receiver receiver, Sender sender) {
        super(receiver, sender);
    }

    /**
     * Create a new Keep-Alive.
     * Used for dynamic classloading.
     */
    public HeartbeatClient() {
    }

    /**
     *
     */
    public void run() {
        while (!endRequested()) {
            try {
                getSender().send(ARE_YOU_THERE);
            } catch (IOException ioe) {
                getSender().close();
                break;
            }

            try {
               sleep(TIME_BETWEEN_ACTIVE_PROBES);
            } catch (InterruptedException ie) { }

            closeIfIdleForTooLong();
        }
    }


    /**
     * Close the connection if there has been no incoming message in a long time.
     * Here we do not differentiate between incoming I-Hear-You and other
     * incoming message.
     */
    private void closeIfIdleForTooLong() {
        if (getSilentTimeForIncoming() > maxPermittedIdleTime) {
            getSender().close();
        }
    }
}
