package keepalive.keepalive;

import java.io.IOException;

import keepalive.communication.MessageListener;
import keepalive.communication.Receiver;
import keepalive.communication.Sender;

/**
 * Echoes the "heartbeat" which is sent regularly by HeartbeatClient,
 * closing the socket if there is silence for a suspicious time period.
 * <BR>
 * Combines PassiveKeepAlive (listens for suspicious silence in
 * incoming data) and IHearYou (responds to AreYouThere).
 * Unlike PassiveKeepAlive, but like AreYouThere,
 * it will automatically close the socket when the timeout period
 * passes with no incoming data.
 *
 * @see keepalive.keepalive.HeartbeatClient
 * @author Joshua Fox
 */
public class HeartbeatServer extends KeepAlive implements MessageListener {
    /**
     * Allow a maximum idle time that is a multiple of the time
     * between probes, to allow for network latency.
     */
    private static final long maxPermittedIdleTime =
                                                6 * TIME_BETWEEN_ACTIVE_PROBES;

    /**
     * This is the server.
     * @param sender
     */
    public HeartbeatServer(Receiver receiver, Sender sender) {
        super(receiver, sender);
    }

    /**
     * Used for dynamic classloading.
     */
    public HeartbeatServer() {
    }

    /**
     * Set the receiver object which receives messages that arrive
     * at the socket associated with this Keep-Alive.
     * @param receiver
     */
    public void setReceiver(Receiver receiver) {
        super.setReceiver(receiver);
        receiver.setMessageListener(this);
    }

   /**
    */
    public void run() {
     while (!endRequested()) {
            // Can be implemented alternately using wait/notify.
            // See PassiveKeepAlive for an example.
            try {
               sleep(TIME_BETWEEN_ACTIVE_PROBES);
            } catch (InterruptedException ie) { }

            closeIfIdleForTooLong();
        }
    }


    /**
     * This method is called by Receiver as the call-back method
     * from MessageListener
     *
     * Ignore all messages but "Are You There." Respond to Are-You-There
     * with "I Hear You."
     *
     * There is an alternate implementation, not as simple but more
     * conservative of bandwidth, which resembles somewhat
     * the AreYouThere/IHearYou pair.
     *
     * <UL>
     *    <LI> Set the HeartbeatClient to only send its Are-You-There
     *    heartbeat when there has been no outgoing message from
     *    the HeartbeatClient in a certain time period.</LI>
     *    <LI> Set this HeartbeatServer to respond with I-Hear-You to
     *    <EM>all</EM> incoming data; however, do not send I-Hear-You if there
     *    has not been an outgoing message from the HeartbeatServer in
     *    the last time period.</LI>
     * </UL>
     */
    public boolean messageReceived(String s) {
        if (s.equals(ARE_YOU_THERE)) {
            try {
                getSender().send(I_HEAR_YOU);
            } catch (IOException ioe) {
                getSender().close();
            }
            return true;
        }
        return false;
    }


    /**
     *
     */
    private void closeIfIdleForTooLong() {
        if (getSilentTimeForIncoming() > maxPermittedIdleTime) {
            getSender().close();
        }
    }
}
