package keepalive.keepalive;

import keepalive.communication.Sender;
import keepalive.communication.Receiver;

/**
 * A Keep-Alive is an object that participates in a Keep-Alive protocol.
 * A Keep-Alive protocol is a way of detecting that a connection has been
 * lost so that socket resources can be cleaned up. Similarly, Keep-Alives can
 * detect that the remote application has ceased functioning (even when the
 * remote host is still alive). This is the base class for all Keep-Alives.
 * Includes most of the utility functionality, but leaves the run() method to
 * the subclasses.
 *
 * @author Joshua Fox
 */
public abstract class KeepAlive extends Thread {
    /**
     * A String to be sent by Are-You-There protocols. I-Hear-You protocols have
     * to recognize this String and respond to it.
     */
    protected static final String ARE_YOU_THERE = "AreYouThere";
    /**
     * A String to be sent by I-Hear-You Keep-Alives in respond to 
     * Are-You-There.
     * In the examples given here, the "I Hear You" message has no special
     * function on the receiving side and could be replaced by a dummy message.
     */
    protected static final String I_HEAR_YOU = "IHearYou";

    /**
     * Time between active Keep-Alive probes.
     * Other times in the Keep-Alive protocols may be derived
     * from this to create a proper balance between the
     * various time periods.
     *
     * The time period given here is far too short for
     * most of the KeepAlive classes, since significant
     * latency on the Internet does not indicate a connection
     * failure. A sender in TCP, for example,
     * allows a time period to go by before giving up on receiving
     * an acknowledgement from the other side. This means
     * that Keep-Alives based on the TCP ACKs must not expect
     * instant feedback.
     *
     * We give a short time period in this test environment
     * to let us see promptly how KeepAlive works.
     *
     * In a real program, you should set this value according to the needs of
     * your system, balancing promptness against minimized network load.
     */
    protected static final long TIME_BETWEEN_ACTIVE_PROBES = 2000L;

    /**
     * Shutdown switch for the thread.
     */
    protected volatile boolean endRequested = false;

    /**
     * Serial-number count for dummy packets.
     */
    private int dummyPacketCount = 0;

    /**
    */
    private Sender sender;

    /**
    */
    private Receiver receiver;

    /**
     * Create the Keep-Alive object with the given receiver and sender.
     *
     * @param receiver Object for sending messages.
     * Not necessarily used by all Keep-Alive subclasses.
     * @param sender Object for receiving messages.
     * Not used by all Keep-Alive subclasses.
     */
    public KeepAlive(Receiver receiver, Sender sender) {
        this();
        setReceiver(receiver);
        setSender(sender);
    }

    /**
     * Create a Keep-Alive with no receiver or sender set.
     */
    public KeepAlive() {
        setDaemon(true);
        setName(this.getClass().getName());
    }


    /**
     * Require the subclasses to implement run(), to prevent them
     * from using the run() method in class Thread with a null body.
     */
    public abstract void run();

    /**
    */
    public void setSender(Sender sender) {
        this.sender = sender;
    }

    /**
    */
    protected Sender getSender() {
        return sender;
    }

    /**
    */
    public void setReceiver(Receiver receiver) {
        this.receiver = receiver;
    }

    /**
    */
    protected Receiver getReceiver() {
        return receiver;
    }


    /**
     * Get a dummy String for use in Keep-Alive messages probes.
     * For ease of debugging, this String will be "Dummy-" plus a
     * serial number. In a real application, you may want to
     * minimize load by sending as short a dummy String as possible.
     *
     * Recall that there is a buffer on the socket level, typically 4 kilobytes
     * in size. If there is a network partition, short outgoing messages may
     * accumulate in this outgoing buffer for a while, so that you
     * will not get an IOException immediately.
     *
     * Lower-level applications (typically written in C) will often
     * use a single byte as the dummy message.
     *
     * @return a dummy string
     */
    protected String getDummyString () {

    return "Dummy-" + ++dummyPacketCount;
    }

    /**
     * Request that the thread be shut down.
     */
    public void end() {
        interrupt();
        endRequested=true;
    }

    /**
     * Indicate if the thread is being shut down.
     *
     * @return true iff the thread is being shut down.
     */
    protected boolean endRequested() {
        return isInterrupted() || endRequested;
    }

    /**
     * How long since we heard from the other side.
     * Passive Keep-Alives and Are-You-Theres in
     * may make use of this method.
     *
     * @return the number of milliseconds since we heard from the other side,
     * or return 0 if receiver == null
     */
    protected long getSilentTimeForIncoming() {
        if (getReceiver()==null) return 0;
        return System.currentTimeMillis() - getReceiver().lastTimeStamp();
    }
}
