package keepalive.keepalive;

import java.io.IOException;
import keepalive.communication.Sender;
import keepalive.communication.Receiver;
import keepalive.communication.MessageListener;
/**
 * Resembles the ActiveKeepAlive in that it sends out dummy messages to see if it gets an IOException, but does not
 * send out dummy messages periodically; rather, only if no incoming data is forthcoming for a long time
 * do we test the connection.
 *
 * @author Joshua Fox
 */
public class PassiveKeepAlive extends KeepAlive implements MessageListener {

    /**
     * If a message was received, this is set to false, so that
     * the listening thread (the KeepAlive thread) does not bother checking
     * how long the connection has been silent.
     * Access to this field is synchronized.
     */
    private volatile boolean messageJustReceived = false;

    /**
    */
    public PassiveKeepAlive (Receiver receiver, Sender sender) {
        super(receiver, sender);
    }

    /**
     * Used for dynamic classloading
     */
    public PassiveKeepAlive() {
    }

    /**
    */
    public void setReceiver(Receiver receiver) {
        super.setReceiver(receiver);
        receiver.setMessageListener(this);
    }

    /**
    */
    public void run() {

        while (!endRequested()) {
            try {
                synchronized (this) {
                    wait(TIME_BETWEEN_ACTIVE_PROBES * 2);
                }

                if (!messageJustReceived) {
                   getSender().send(getDummyString());
                }

                messageJustReceived = false;
            } catch (InterruptedException ie) {
                break;
            } catch (IOException ioe) {
                getSender().close();
                break;
            }
       }
    }

    /**
     * Whenever we receive a message, wake up the listening thread
     * for another loop. (The alternative mechanism would have the listening
     * thread sleep() for a fixed time each loop, then check to see if
     * the connection was silent. This might mean waking up uselessly, however.
     */
    public synchronized boolean messageReceived(String s) {
        messageJustReceived = true;
        notify();
        return false;
    }

}
