package keepalive.socket;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;

import keepalive.communication.Receiver;
import keepalive.communication.Sender;
import keepalive.keepalive.KeepAlive;
import keepalive.keepalive.NullKeepAlive;

/**
 * Abstracts the socket functionality. Sockets in this system have type 
 * AbstractSocket to allow real TCP/IP sockets or simulated sockets to be 
 * used interchangeably.
 *
 * @author Joshua Fox
 */
public abstract class AbstractSocket {
    /**
    */
    private static String clientKeepAliveClassName = 
                                            "keepalive.keepalive.NullKeepAlive";
    /**
    */
    private static String serverKeepAliveClassName = 
                                            "keepalive.keepalive.NullKeepAlive";

   /**
    */
   private KeepAlive keepAlive = new NullKeepAlive();
   /**
    */
   public abstract InputStream getInputStream() throws IOException;
   /**
    */
   public abstract OutputStream getOutputStream() throws IOException;
   /**
    */
   public abstract void closeSocket() throws IOException;

   /**    
    */
   public void close() throws IOException {
      keepAlive.end();
      closeSocket();
   }


  /**    
    */
   public void startKeepAlive(Receiver receiver, Sender sender, 
                                                           boolean serverSide) {
        String className = determineKeepAliveClassName(serverSide);

        keepAlive = null;
        try {
            Class keepAliveClass = Class.forName(className);
            keepAlive = (KeepAlive) keepAliveClass.newInstance();
        } catch (Exception e) { //InstantiationException, IllegalAccessException
            System.err.println("Exception in instantiating KeepAlive " + 
                                                                 className + e);
            keepAlive = new NullKeepAlive();
        }

        keepAlive.setReceiver(receiver);
        keepAlive.setSender(sender);
        keepAlive.start();
   }

   /**    
    */
   static void setServerKeepAlive(String keepAliveClassName) {
        serverKeepAliveClassName = keepAliveClassName;
   }

   /**    
   */
   static void setClientKeepAlive(String keepAliveClassName) {
        clientKeepAliveClassName = keepAliveClassName;
   }

   /**    
   */
   static String getServerKeepAlive() {
        return serverKeepAliveClassName;
   }

   /**    
   */
   static String getClientKeepAlive() {
        return clientKeepAliveClassName;
   }


   /**    
   */
   private String determineKeepAliveClassName(boolean serverSide) {
        return serverSide?serverKeepAliveClassName:clientKeepAliveClassName;
   }
}
