package keepalive.socket;

import java.io.IOException;
import java.net.InetAddress;

import keepalive.socket.sim.SimSocket;
import keepalive.socket.sim.SimServerSocket;
import keepalive.socket.adapter.SocketAdapter;
import keepalive.socket.adapter.ServerSocketAdapter;


/**
 * Create new AbstractSockets, whether real TCP/IP sockets or simulated sockets.
 *
 */
public class SocketFactory {
    /**
    */
    public static final int PORT = 5555;

    // "enum" of types of Sockets: "real" or simulated:
    public static final int SOCKET     = 1;
    public static final int SIM_SOCKET = 2;
    private static int type = SOCKET; // default is regular sockets

    /**
     * Set the type of sockets we will be producing
     * @param a type of socket-- "real" or simulated-- using the
     * integer constants defined in this class.
     */
    public static void setSocketType(int sockType) {
        type = sockType;
    }


    /**
    * Create a new client-side socket that connects to a serverhost.
    * @param addr the InetAddress of the server host.
    * @return a new socket
    */
    public static AbstractSocket createSocket(InetAddress addr) 
                                                            throws IOException {

        AbstractSocket sock = null;
        switch (type) {
            case SOCKET:
                sock = new SocketAdapter(addr, PORT);
                break;
            case SIM_SOCKET:
                // There is no address for the simulated socket 
                // (always localhost)            
                sock =  new SimSocket(PORT); 
                break;
         }

         return sock;
    }

    /**
     * Create a ServerSocket bound to the port SocketFactory.PORT
     * @return a new ServerSocket.
     */
    public static ServerSocketIfc createServerSocket() throws IOException{
        ServerSocketIfc srvSock = null;
        switch (type) {
            case SOCKET:
                srvSock = new ServerSocketAdapter(PORT);
                break;
            case SIM_SOCKET:
                srvSock = new SimServerSocket(PORT);
                break;
        }

        return srvSock;
     }


   /**
    * Set the Keep-Alive class to be used for all server-side sockets
    * produced by this factory.
    * @param keepAliveClassName the fully-specified name of
    * the Keep-Alive class to use.
    */
   public static void setServerKeepAlive(String keepAliveClassName) {
        AbstractSocket.setServerKeepAlive(keepAliveClassName);
   }

   /**
    * Set the Keep-Alive class to be used for all client-side sockets
    * produced by this factory.
    * @param keepAliveClassName the fully-specified name of
    * the Keep-Alive class to use.
    */
   public static void setClientKeepAlive(String keepAliveClassName) {
        AbstractSocket.setClientKeepAlive(keepAliveClassName);
   }

   /**
    * Get the Keep-Alive class used for all server-side sockets
    * produced by this factory.
    * @return the Keep-Alive class used for all server-side sockets
    * produced by this factory.
    */
   public static String getServerKeepAlive() {
        return AbstractSocket.getServerKeepAlive();
   }

   /**
    * Get the Keep-Alive class used for all client-side sockets
    * produced by this factory.
    * @return the Keep-Alive class used for all client-side sockets
    * produced by this factory.
    */
   public static String getClientKeepAlive() {
        return AbstractSocket.getClientKeepAlive();
   }
}
