package keepalive.socket.sim;

import java.io.PipedOutputStream;
import java.io.PipedInputStream;
import java.io.IOException;


/**
 * A "Non-Breaking" PipedOutputStream: a PipedOutputStream which does
 * not break the pipe when it is closed when it is not "connected" to
 * the other side (where a connection is represented by two plugged-in
 * Pluggable.)
 */
class NBPipedOutputStream extends PipedOutputStream {
    /**
     * The plug for "this side", the output side.
     */
    private Pluggable plug;
    /**
     * The plug for "the other side", the side with the PipedInputStream that
     * is the sink for this PipedOutputStream.
     */
    private Pluggable otherSidesPlug;

    /**
     * Creates a piped output stream that is not yet connected to a
     * piped input stream. It must be connected to a piped input stream,
     * with a plug on the other side, before being used.
     *
     * @param plug
     */
    public NBPipedOutputStream(Pluggable plug) {
        // Call super();
        this.plug = plug;
    }

    /**
     * Connect to the given PipedInputStream, where the other side has the
     * given "plug".
     *
     * @param sink
     * @param otherSidesPlug
     * @exception IOException If an I/O error occurs.
     */
    public void connect(PipedInputStream sink, Pluggable otherSidesPlug)
                                                            throws IOException {
        this.otherSidesPlug = otherSidesPlug;
        super.connect(sink);
    }

    /**
     * Write a byte to the output stream if the connection exists.
     *
     * @exception IOException If an I/O error occurs, for example, if the 
     *     connection has been lost.
     * @param b
     */
    public void write(int b)  throws IOException {
        interruptIfNeeded();
        super.write(b);
    }

    /**
     * Write a byte array to the output stream if the connection exists.
     *
     * @exception IOException if an I/Oerror occurs, for example, if the 
     *     connection has been lost.
     * @param b
     * @param off
     * @param len
     */
    public void write(byte b[], int off, int len) throws IOException {
        interruptIfNeeded();
        super.write(b, off, len);
    }

    /**
     * Break this stream if the connection has been lost--if one plug
     * or the other has been pulled.
     *
     * @exception IOException
     */
    private void interruptIfNeeded() throws IOException {
        if (!plug.isPluggedIn() ) {
            throw new IOException("The plug was pulled.");
        }


        if (otherSidesPlug!=null && !otherSidesPlug.isPluggedIn()) {
            throw new IOException("The other side's plug was pulled.");
        }
    }

    /**
     * Close the output stream. The behavior differs from an ordinary
     * PipedOutputStream in that closing this PipedOutputStream will not
     * break the the sink (the corresponding PipedInputStream) if the
     * connection has been lost.
     *
     * @exception IOException if an I/O error occurs
     */
    public void close() throws IOException {
        // IF     (1) this machine plugged in
        //   AND (2) stream is connected
        //   AND (3) other machine plugged in
        // THEN close other side.
        if (plug.isPluggedIn() && otherSidesPlug!=null && 
                                                otherSidesPlug.isPluggedIn()) {
            super.close(); // closes other side
        } else {

            /* do nothing*/

            // close() in PipedOutputStream calls receivedLast() in the
            // connected OpenPipedInputStream. In this else statement, we
            // presume that the sockets cannot talk to each other over
            // simulated TCP/IP to conduct the 'close' protocol.
            // In this overriding method we do not include the receivedLast()
            // in the overridden method so that the closing of the
            // PipedOutputStream does not automatically close the other
            // side's PipedInputStream. Then the KeepAlive can do its thing.
        }
    }
}