package keepalive.socket.sim;

import java.util.Iterator;
import keepalive.ui.KeepAliveUI;
import keepalive.socket.sim.ui.NetstatUI;
import keepalive.socket.sim.ui.SimHostFrame;
import keepalive.apps.Server;
import keepalive.apps.Client;

/**
 * A simulated host (computer on the network).
 * @author Joshua Fox
 */
public class SimHost implements Pluggable {
    /**
     * The user interface for a netstat-like report on sockets running on
     * this host.
     */
    private NetstatUI netstatUI;
    /**
     * True iff this host is acting as a server. In fact, hosts can act as server or client or both, but since keep-alive protocols are often asymmetric with respect to the server/client distinction, we note the difference here.
     */
    private boolean serverSide;
    /**
     * Whether this machine is plugged in. The metaphor is to the Ethernet plug on the computer, although of course other types of connections are possible.
     */
    private boolean pluggedIn =true;

    /**
     * Create a new host, indicating whether it is to act as a server.
     *
     * @param serverSide
     */
    public SimHost(boolean serverSide) {
        this.serverSide = serverSide;
        if (serverSide) {
           SimSocket.setServerHost(this);
           SimServerSocket.setServerHost(this);
        } else {
           SimSocket.setClientHost(this);
        }

        netstatUI = new NetstatUI(this);

        new NetstatThread().start();
        KeepAliveUI keepAliveUI = null;

        // Set up a Client or Server Application on this "host".
        // Don't keep a reference to the Application, just
        // and set it running and forget about it.
        if (serverSide) {
            Server server = new Server();
            keepAliveUI = server.getKeepAliveUI();
        } else {
            Client client  = new Client("localhost");
            keepAliveUI = client.getKeepAliveUI();
        }

        new SimHostFrame(netstatUI, keepAliveUI, serverSide);
    }


    /**
     * Get a report on the state of the sockets on this host.
     *
     * @return the report
     */
    public String getReport() {
        StringBuffer buf = new StringBuffer();
        buf.append("ServerSockets:\n");
        Iterator serverSockIter = getServerSocketIterator();
        while (serverSockIter.hasNext()) {
            buf.append("  " + serverSockIter.next() + "\n");
        }

        buf.append("\n\n");

        buf.append("Sockets:\n");
        Iterator sockIter = getSocketIterator();
        while (sockIter.hasNext()) {
            buf.append("  " + sockIter.next() + "\n");
        }

        return buf.toString();

    }

    /**
     * Get Iterator of all Sockets (not ServerSockets) now open on this host.
     * @return Iterator of all Sockets (not ServerSockets) now open on this host.
     */
    private Iterator getSocketIterator() {
        return SimSocket.getSocketIterator(this);
    }

    /**
     * Get Iterator of all ServerSockets now open on this host.
     * @return Iterator of all ServerSockets now open on this host.
     */
    private Iterator getServerSocketIterator() {
        return SimServerSocket.getServerSocketIterator(this);
    }

    /**
     * Find whether this host is acting as a server.
     */
    public boolean isServerSide() {
        return serverSide;
    }

    /**
     * Find if this host is plugged in.
     * @return true if this host is plugged in.
     */
    public boolean isPluggedIn() {
        return pluggedIn;
    }

    /**
     * Plug this host in or unplug it.
     *
     * @param b true to plug this host in; false to unplug it.
     */
    public synchronized void setPluggedIn(boolean b) {
       pluggedIn = b;
    }


    /**
     * Inner class that continually updates the netstat-like report.
     */
    private class NetstatThread extends Thread {
        /**
         */
        NetstatThread() {
            super("NetstatThread");
            setDaemon(true);
        }

        /**
         * Run a thread that continually updates the netstat-like report.
         */
        public void run() {
            // Note that this reporting mechanism gets the complete report
            // repeatedly, rather than altering the display on the basis
            // of "deltas". This is less efficient; however, because
            // this reporting mechanism is intended for debugging
            // and for simulation, we do not want to trust the
            // the system to correctly report deltas.
            while (!interrupted()) {
                String report = getReport();
                netstatUI.display(report);
                try {sleep(1000L);} catch (InterruptedException ie) {}
            }
        }
    }
}
