package keepalive.socket.sim;

import java.io.IOException;
import java.net.BindException;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Map;
import java.util.List;
import java.util.LinkedList;
import keepalive.socket.SocketFactory;
import keepalive.socket.AbstractSocket;
import keepalive.socket.ServerSocketIfc;


/**
 * A simulated ServerSocket.
 */
public class SimServerSocket implements ServerSocketIfc {
    private static Map instances = new HashMap();
    private static SimHost serverHost;
    private SimSocket remoteSocket;
    private int port;

    /**
     * Set the simulated host on which this SimServerSocket is listening.
     *
     * @param serverHost
     */
    static void setServerHost(SimHost serverHost) {
        SimServerSocket.serverHost =serverHost;
    }

    /**
     * Create a new ServerSocket to listen on the given port
     *
     * @param port
     * @exception IOException if an I/O error occurs (for example, if the
     *     simulated port is already bound.)
     */
    public SimServerSocket(int port) throws IOException{
        if (serverHost==null) {
           throw new NullPointerException(
            "Cannot start SimServerSocket without a host.");
        }

        this.port = port;
        bind(this, port);
    }

    /**
     * Find the SimServerSocket bound to the given port.
     *
     * @param port
     */
    static SimServerSocket findSimServerSocket(int port) {
       return (SimServerSocket) instances.get(new Integer(port));
    }

    /**
     * Bind the given SimServerSocket to the given port.
     *
     * @exception IOException
     * @param sock
     * @param port
     */
    static private void bind(SimServerSocket sock, int port) throws IOException {
        if (instances.get(new Integer(port)) !=null) {
            throw new BindException();
        }

        instances.put(new Integer(port), sock);
    }


    /**
     * Get the port to which this ServerSocket is bound.
     */
    public int getPort() {
        return port;
    }

   /**
    * Get Iterator all SimServerSockets in this simulated
    * networking system (on "server" and "client" hosts).
    * @return an Iterator of all SimServerSockets in this
    * simulated networking system (on "server" and "client" hosts).
    */
   public static Iterator getServerSocketIterator() {
      return instances.values().iterator();
   }

   /**
    * Accept a new connection and spin off a new socket
    * @return the spun-off socket.
    * @exception IOException if there is an I/O error.
    */
   public synchronized AbstractSocket accept() throws IOException {
        if (!isPluggedIn()) {
          throw new IOException("Server host not connected to net.");
        }

        try {
            wait();
        } catch (InterruptedException ie) {}
        SimSocket localSocket = new SimSocket(remoteSocket);
        remoteSocket = null;
        return localSocket;

   }

   /**
    * Close this SimServerSocket.
    *
    * @exception IOException
    */
   public void close() throws IOException{
        instances.remove(new Integer(this.getPort()));

   }

   /**
    *
    * @param remoteSocket
    */
   synchronized  void connect(SimSocket remoteSocket) {
        this.remoteSocket = remoteSocket;
        notify();
   }

   /**
    * Get an Iterator of  all SimServerSockets on the given host.
    * @return an Iterator of all SimServerSockets
    * on the given host.
    *
    * @param forThisHost the host whose SimServerSockets we
    * are interested in.
    */
   static Iterator getServerSocketIterator(SimHost forThisHost) {
        List list = new LinkedList();
        Iterator it = instances.values().iterator();
        while (it.hasNext()) {
            SimServerSocket serverSock = (SimServerSocket) it.next();
            if (serverSock.getSimHost()==forThisHost) { // object identity
                list.add(serverSock);
            }
        }

        return list.iterator();
   }

   /**
    * Get the simulated host on which this SimServerSocket is listening.
    */
   SimHost getSimHost() {
       return serverHost;
   }

   /**     */
   public boolean isPluggedIn() {
       return getSimHost().isPluggedIn();
   }

   /**     */
   public String toString() {
        return "SimServerSocket-" + getPort();
   }
}
