package keepalive.socket.sim;

import java.io.InputStream;
import java.io.PipedInputStream;
import java.io.OutputStream;
import java.io.IOException;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import keepalive.socket.AbstractSocket;


/**
 * A simulated socket, within a single VM.
 *
 * @author Joshua Fox
 */
public class SimSocket extends AbstractSocket {
    /**
     * For input into this socket.
     */
    private PipedInputStream pipedInputStream;

    /**
     * For output from this socket.
     */
    private NBPipedOutputStream pipedOutputStream;

    /**
     * This SimSocket may be running on client-side or server side.
     * To allow two hosts to connect, these static fields must be set
     * with references to the simulated server host and simulated
     * client host. (In a real socket system, we'd use IP addresses to
     * let the two sides find each other; this simulation does not
     * have simulated IP addresses.)
     */
    private static SimHost serverHost = null;
    private static SimHost clientHost = null;

    /**
     * All SimSockets in our simulated networking system are registered
     in this list.
     */
    private static List instances = new ArrayList();

    /**
     * Used  for assigning new epheral ports
     */
    private static int portCounter = 0;

    /**
     * The port for this SimSocket.
     */
    private int ephemeralPort;

    /**
     * The simulated host on which this socket runs.
     */
    private SimHost simHost;


    /**
     */
    static void setServerHost(SimHost serverHost) {
           SimSocket.serverHost = serverHost;
    }

    /**
     */
    static void setClientHost(SimHost clientHost) {
           SimSocket.clientHost = clientHost;
    }

    /**
     * Connects to a SimServerSocket from Client side
     */
    public SimSocket(int port) throws IOException {
        if (clientHost==null) {
          throw new NullPointerException("Client Host is null when creating a socket.");
        }

        this.simHost = clientHost;

        if (!simHost.isPluggedIn()) {
            throw new IOException("Cannot connect; host not connected to net.");
        }

        initialize();

        // create unlinked and currently unusable streams
        pipedOutputStream = new NBPipedOutputStream(this.simHost);
        pipedInputStream = new PipedInputStream();

        SimServerSocket srvSock = SimServerSocket.findSimServerSocket(port);
        if (srvSock == null) {
            instances.remove(this);
            throw new IOException("Could not find SimSocketSocket listening on port " + port);
        }

        srvSock.connect(this);
    }


    /**
     * Use this when a SimSocket is spun off from a SimServerSocket.
     */
    public SimSocket(SimSocket theOtherOne) throws IOException {
        if (serverHost==null) {
          throw new NullPointerException(
                "Server Host is null when spinning off a socket.");
        }

        this.simHost = serverHost;

        if (!simHost.isPluggedIn()) {
            throw new IOException("Cannot connect; host not connected to net.");
        }

        if (!theOtherOne.isPluggedIn()) {
            throw new IOException("Cannot connect; other host not reachable.");
        }

        initialize();

        pipedOutputStream = new NBPipedOutputStream(this.simHost);
        pipedOutputStream.connect(theOtherOne.pipedInputStream, theOtherOne.simHost);

        pipedInputStream = new PipedInputStream();
        theOtherOne.pipedOutputStream.connect(this.pipedInputStream, this.simHost);

   }

    /**
     * Initialize this SimSocket.
     */
    private void initialize() {
        instances.add(this);
        ephemeralPort = getEphemeralPort();
    }

    /**
     * Get a new ephemeral port. Note that ServerSockets are bound to a given port,
     * whereas sockets are assigned an arbitrary port number (by this method).
     *
     * @return the new ephemeral port number.
     */
    private synchronized static int getEphemeralPort() {
        return 1024 + ++portCounter;
    }

    /**
     * Get an Iterator of all SimSockets in this simualted networking system.
     * @param forThisHost
     */
    static Iterator getSocketIterator(SimHost forThisHost) {
        List list = new LinkedList();
        Iterator it = instances.iterator();
        while (it.hasNext()) {
            SimSocket sock = (SimSocket) it.next();
            if (sock.simHost==forThisHost) { // object identity
                list.add(sock);
            }
        }

        return list.iterator();
    }

    /**
     * Get the input stream associated with this simulated socket.
     * @exception IOException
     */
    public InputStream getInputStream() throws IOException {
        return pipedInputStream;
    }

    /**
     * Get the output stream associated with this simulated socket.
     * @exception IOException
     */
    public OutputStream getOutputStream() throws IOException {
        return pipedOutputStream;
    }

    /**
     * Close this socket. This method is a hook method
     * called by close() in the abstract superclass.
     *
     * @exception IOException
     */
    public void closeSocket() throws IOException {
        instances.remove(this);
        if (pipedOutputStream == null ) {
            throw new IOException("Null Output Stream");
        }

        if (pipedInputStream == null) {
            throw new IOException("Null Input Stream");
        }

        pipedOutputStream.close();
        pipedInputStream.close();
    }


    /**
     * Whether this simulated socket's host is "plugged-in."
     * @return true iff this simulated socket's host is "plugged-in."
     */
    public boolean isPluggedIn() {
        return simHost.isPluggedIn();
    }


    /**
     */
    public String toString() {
        return "SimSocket-" + ephemeralPort;
    }
}

