//////////////////////////////////////////////////////////////
//
// Misc.CPP
//
// This is the source file for the NewGPE display driver
// class that contains miscellaneous functionality such as
// palette setting and vertical blank checking.
//
//////////////////////////////////////////////////////////////

///////////////////////////////////
//
// Includes
//

#include <windows.h>
#include <types.h>
#include "NewGPE.h"

///////////////////////////////////
//
// Constants
//

///////////////////////////////////
//
// Globals
//

///////////////////////////////////
//
// Function Prototypes
//

///////////////////////////////////
//
// Functions
//

//////////////////////////////////////////////////////////////
//
// NewGPE::SetPalette
//
// Set the display palette.
//
// PARAMETERS
// src          array of palette entries
// firstEntry   index of first entry to set
// numEntries   number of entries to set
//
// RETURNS
// S_OK if successful, otherwise some error code
//
// NOTES
// Palette support for VGA devices is already implemented.
// This function only needs to be modified or overridden
// for non-VGA devices.
//
SCODE NewGPE::SetPalette(
						const PALETTEENTRY *src,
						unsigned short firstEntry,
                        unsigned short numEntries )
{
    SCODE sc = S_OK;

    // verify that parameters are valid
    int nMaxEntries = 1<<m_pMode->Bpp;
    if ((firstEntry<0) || ((firstEntry+numEntries)>nMaxEntries))
        // invalid entry range
        return E_INVALIDARG;

    if (m_bIsVGADevice) {
        // VGA device, use standard VGA palette accesses
        // (VGA palette only the six most significant bits)
        for (int i=0; i<numEntries; i++) {
            reg_DAC_WR_AD = (BYTE) (i + firstEntry);
            reg_DAC_DATA = (BYTE) ((src[i].peRed)>>2);
            reg_DAC_DATA = (BYTE) ((src[i].peGreen)>>2);
            reg_DAC_DATA = (BYTE) ((src[i].peBlue)>>2);
        }
    }
    else {
        // Non-VGA device
        // NOTE:
        // Not implemented, device specific support needs to be added.
        //
        RETAILMSG(1, (TEXT("NewGPE::SetPalette ERROR - Non-VGA devices not supported!\r\n")));

        sc = E_NOTIMPL;
    }

    return sc;
}

//////////////////////////////////////////////////////////////
//
// NewGPE::InVBlank
//
// Check if display is in vertical blank.
//
// PARAMETERS
// None
//
// RETURNS
// TRUE if in vertical blank, otherwise FALSE
//
// NOTES
// Vertical blank detection support for VGA devices is
// already implemented. This function only needs to be
// modified or overridden for non-VGA devices.
//
int	NewGPE::InVBlank()
{
    if (m_bIsVGADevice) {
        // checking VGA Input Status Register (Address 3DA)
        //   Vertical Retrace Interval (bit 3)
        if (reg_STATUS_1&0x08)
            return TRUE; // in vertical blank
        else
            return FALSE;
    }

    // Non-VGA device
    // NOTE:
    // Not implemented, device specific support needs to be added.
    //
    RETAILMSG(1, (TEXT("NewGPE::InVBlank ERROR - Non-VGA devices not supported!\r\n")));

    return FALSE;
}
