//////////////////////////////////////////////////////////////
//
// NewGPE.CPP
//
// This is the main source file for the NewGPE display driver
// class. It is derived from the default display driver class
// GPE. Other source files for this class contain the
// implementation for specific areas of functionality. They
// are:
//
// SURF.CPP   - All surface related functions, including
//              functions for the NewGPESurf class.
// CURSOR.CPP - All cursor related functions.
// BLT.CPP    - All bit blit related functions.
// LINE.CPP   - All line drawing related functions.
// MISC.CPP   - Miscellaneous functions such as palette
//              setting and vertical blank checking.
//
//////////////////////////////////////////////////////////////

///////////////////////////////////
//
// Includes
//

#include <windows.h>
#include <types.h>
#include <winddi.h>
#include <gpe.h>
#include "ioport.h"     // I/O port class definitions
#include "NewGPE.h"

///////////////////////////////////
//
// Constants
//

///////////////////////////////////
//
// Globals
//

// 16 and 256 color palettes
// Defined in palette.h file as:
// * m_rgbIdentityPal
// * m_rgbIdentityPal16
#include "palette.h"

// bit masks used by DrvGetMasks and initialized in SetMode
ULONG m_ulBitMasks[3];

// Setup debug zones for the driver. Errors and warnings are
// currently enabled. See GPE.H for more information on other
// debug zones. If this class is derived from, you should
// move this line to the derived class CPP file.
INSTANTIATE_GPE_ZONES(0x0003,"DDI Driver","unused1","unused2")

// pointer to instance of NewGPE class
static NewGPE *m_pGPE = (NewGPE *)NULL;

//////////////////////////////////////////////////////////////
//
// m_gpeModeTable
//
// This table needs to be filled with the modes that the
// driver supports.
//
// The first entry is reserved for the default mode, that is
// the mode to use if no mode is specified in the registry.
//
// The NULL entry marks the end of the list and should not be
// removed.
//
struct GPEMode m_gpeModeTable[] = {
    // modeId    width     height    Bpp       frequency format

    // default mode (must be included)

    // This is a sample entry that works with the sample ModeInit
    // function for the CEPC platform. The CEPC bootloader, LOADCEPC,
    // initializes the VGA card to 320x200 at 8Bpp, so this mode
    // can be supported with no additional initialization required.
    {  1,        320,      200,      8,        60,       gpe8Bpp },

    // all other supported modes (these are optional)

    // this NULL entry marks the end of the list and should not be removed
    {  0,        0,        0,        0,        0,        gpeUndefined }
};

///////////////////////////////////
//
// Function Prototypes
//

// Driver initialization entry point for GPE
BOOL APIENTRY GPEEnableDriver(
	ULONG          iEngineVersion,
	ULONG          cj,
	DRVENABLEDATA *pded,
	PENGCALLBACKS  pEngCallbacks);

///////////////////////////////////
//
// Functions
//

//////////////////////////////////////////////////////////////
//
// GetGPE
//
// Main entry point for a GPE compliant driver. Returns the
// instance of the GPE derived class, creating this instance
// if necessary.
//
// PARAMETERS
// None
//
// RETURNS
// Pointer to the instance of the GPE derived class.
//
// NOTES
//
GPE *GetGPE()
{
	if( !m_pGPE )
		m_pGPE = new NewGPE(); // create the instance of the GPE class
	return (GPE *) m_pGPE;
}

//////////////////////////////////////////////////////////////
//
// DrvEnableDriver
//
// DDI main entry point. GDI calls this function to get
// pointers to the other DDI functions for the driver.
//
// PARAMETERS
// iEngineVersion   DDI version for GDI
// cj               Size of DRVENABLEDATA structure pded
// pded             Pointer to DRVENABLEDATA structure to be filled
// pEngCallbacks    Pointer to ENGCALLBACKS structure
//
// RETURNS
// TRUE if successful, FALSE otherwise.
//
// NOTES
//
BOOL APIENTRY DrvEnableDriver(
	ULONG          iEngineVersion,
	ULONG          cj,
	DRVENABLEDATA *pded,
	PENGCALLBACKS  pEngCallbacks)
{
    // just call default GPE enable routine
	return GPEEnableDriver( iEngineVersion, cj, pded, pEngCallbacks );
}

//////////////////////////////////////////////////////////////
//
// DrvGetMasks
//
// Returns the color bit masks for the display devices
// current mode.
//
// PARAMETERS
// dhpdev       Pointer to the PDEV that describes the
//              physical device.
//
// RETURNS
// Pointer to color mask for the current mode.
//
// NOTES
//
ULONG *APIENTRY DrvGetMasks(
    DHPDEV     dhpdev)
{
    return m_ulBitMasks;
}

//////////////////////////////////////////////////////////////
//
// RegisterDDHALAPI
//
// Responsible for registering the DDHAL API, if supported by
// the driver.
//
// PARAMETERS
// none.
//
// RETURNS
// none.
//
// NOTES
// This function must be included in the driver or it will
// generate a link error. However, it does not need to do
// anything as DDHAL API support is not required.
//
void RegisterDDHALAPI()
{
}

//////////////////////////////////////////////////////////////
//
// GetDisplayMemoryPtr
//
// Returns a pointer to the start of video memory currently
// being displayed.
//
// This function is intended as an example for how to extend
// the display driver model by exporting certain features for
// access by applications and DLLs.
//
// PARAMETERS
// None
//
// RETURNS
// Pointer to display memory.
//
// NOTES
//
void *GetDisplayMemoryPtr()
{
    void *pRet = NULL;

    // get GPE instance
    GPE *pGPE = GetGPE();
    if (pGPE) {
        // now get a pointer to the primary surface
        GPESurf *pSurf = pGPE->PrimarySurface();
        if (pSurf) {
            // finally get the memory pointer
            pRet = pSurf->Buffer();
        }
    }

    return pRet;
}

//////////////////////////////////////////////////////////////
//
// NewGPE::NewGPE
//
// Class constructor.
//
// PARAMETERS
// none
//
// RETURNS
// none
//
// NOTES
//
NewGPE::NewGPE()
{
    DebugBreak();

    // Flags for hardware features
    //
    // m_bIsVGADevice must be changed here or in the constructor
    // of a derived class if the device is VGA based.
    //
    // m_bHWCursor can be changed here, or in the ModeInit
    // function if hardware cursor is supported.
    //
    // m_b555Mode can be changed here, or in the ModeInit
    // function if the 16-bit per pixel mode uses 5 bits
    // each for red, green, and blue. The default is
    // considered 5 bits each for red and blue and 6 bits
    // for green.
    //
    m_bIsVGADevice = TRUE;      // default to VGA device
    m_bHWCursor = FALSE;        // default to software cursor
    m_b555Mode = FALSE;         // default to 5-6-5 mode for 16Bpp

    // NOTE:
    // The following data members are modified to their final values
    // by the default implementation of SetMode and shouldn't need to
    // modified here or in ModeInit.
    //
    m_pPrimarySurface = NULL;   // pointer to primary display surface
    m_nScreenWidth = 0;         // display width
    m_nScreenHeight = 0;        // display height
    m_pMode = NULL;             // pointer to information on current mode
    m_p2DVideoMemory = NULL;    // pointer to video memory manager
    m_pLAW = NULL;              // pointer to linear access window
    memset(&m_ulBitMasks[0], 0, sizeof(m_ulBitMasks)); // bit masks

    // NOTE:
    // The following data members MUST be modified to their final
    // values by the display hardware specific function ModeInit.
    //
    m_nLAWPhysical = 0;     // the physical address of the linear address
                            // window for accessing the frame buffer
    m_nLAWSize = 0;         // size of linear access window
    m_nVideoMemorySize = 0; // size of video memory, which can be different
                            // than the linear access window
    m_nVideoMemoryStart = 0;// offset within the linear access window to the
                            // start of video memory (usually is 0)
    m_nScreenStride = 0;    // number of bytes per display line

    // general initialization for VGA register mapping
    // (specific initialization done later)
   	m_pPortRanges = (PortRange *)NULL;
}

//////////////////////////////////////////////////////////////
//
// NewGPE::~NewGPE
//
// Class destructor.
//
// PARAMETERS
// none
//
// RETURNS
// none
//
// NOTES
//
NewGPE::~NewGPE()
{
}

//////////////////////////////////////////////////////////////
//
// NewGPE::NumModes
//
// Returns the number of supported display modes.
//
// PARAMETERS
// None
//
// RETURNS
// Number of supported modes.
//
// NOTES
// This function reads the supported modes from the data
// member m_gpeModeTable.
//
int NewGPE::NumModes()
{
    // count the number of entries in the mode table

    BOOL bDone = FALSE;
    int nIndex = 0;
    while (!bDone) {
        if (m_gpeModeTable[nIndex].Bpp==0) {
            // no more entries in the table
            bDone = TRUE;
        }
        else {
            // count entry and go to next entry
            nIndex++;
        }
    }

    if (nIndex==0) {
        // ERROR: The mode table has not been initialized. This mode table needs
        //        to be initialized for this function to work.
        RETAILMSG(1, (TEXT("NewGPE::NumModes ERROR - The mode table was not setup properly!\r\n")));
    }

    return nIndex;
}

//////////////////////////////////////////////////////////////
//
// NewGPE::GetModeInfo
//
// Returns information about a particular supported mode.
//
// PARAMETERS
// pMode        Pointer to a mode info structure to fill.
// modeNo       Mode number to get information on.
//
// RETURNS
// S_OK if successful
//
// NOTES
// This function reads the supported modes from the data
// member m_gpeModeTable.
//
SCODE NewGPE::GetModeInfo(
                GPEMode *pMode,
                int modeNo )
{
    // make sure that the mode is valid (index is zero based)
    if ((modeNo<0) || (modeNo>=NumModes()))
        return E_INVALIDARG;

    // get data from mode table
    *pMode = m_gpeModeTable[modeNo];

    return S_OK;
}

//////////////////////////////////////////////////////////////
//
// NewGPE::SetMode
//
// Sets display mode. This function overrides the base GPE
// class pure virtual function and implements the device
// independent part of mode setting.
//
// PARAMETERS
// modeId       ID of mode to set.
// pPalette     Pointer to a palette handle. This should be set
//              to a default palette if the pointer!=NULL.
//
// RETURNS
// S_OK if successful
//
// NOTES
// Default functionality is already implemented for device
// independent support. This function calls ModeInit() for
// the device specific portion of mode setting.
//
// This function reads the supported modes from the data
// member m_gpeModeTable.
//
SCODE NewGPE::SetMode(
				int modeId,
				HPALETTE *pPaletteHandle )
{
    // initialize VGA I/O mapping for VGA devices
    // NOTE:
    // Only the VGA registers required for palette
    // setting and vertical blank detection are fully
    // mapped and defined. Examples of additional VGA
    // register mapping and definitions can be found
    // in the GPEVGA.CPP and GPEVGA.H files included
    // in Platform Builder.
    //
    // taken from GPEVGA.CPP
	// Reserve / Map in - standard VGA ports
    // (make sure not already done)
    if (m_bIsVGADevice && (m_pPortRanges==NULL)) {
	    m_VGAIOPort[PortId_3BA].Map(0x3BA, m_pPortRanges);
	    m_VGAIOPort[PortId_3C0].Map(0x3C0, m_pPortRanges);
	    m_VGAIOPort[PortId_3C1].Map(0x3C1, m_pPortRanges);
	    m_VGAIOPort[PortId_3C2].Map(0x3C2, m_pPortRanges);
	    m_VGAIOPort[PortId_3C3].Map(0x3C3, m_pPortRanges);
	    m_VGAIOPort[PortId_3C4].Map(0x3C4, m_pPortRanges);
	    m_VGAIOPort[PortId_3C5].Map(0x3C5, m_pPortRanges);
	    m_VGAIOPort[PortId_3C6].Map(0x3C6, m_pPortRanges);
	    m_VGAIOPort[PortId_3C7].Map(0x3C7, m_pPortRanges);
	    m_VGAIOPort[PortId_3C8].Map(0x3C8, m_pPortRanges);
	    m_VGAIOPort[PortId_3C9].Map(0x3C9, m_pPortRanges);
	    m_VGAIOPort[PortId_3CA].Map(0x3CA, m_pPortRanges);
	    m_VGAIOPort[PortId_3CC].Map(0x3CC, m_pPortRanges);
	    m_VGAIOPort[PortId_3CE].Map(0x3CE, m_pPortRanges);
	    m_VGAIOPort[PortId_3CF].Map(0x3CF, m_pPortRanges);
	    m_VGAIOPort[PortId_3D4].Map(0x3D4, m_pPortRanges);
	    m_VGAIOPort[PortId_3D5].Map(0x3D5, m_pPortRanges);
	    m_VGAIOPort[PortId_3DA].Map(0x3DA, m_pPortRanges);
    }

    int nModeNum = 0;
    GPEMode gpeMode;
    // get mode entry that matches modeId
    BOOL bDone = FALSE;
    int nNumModes = NumModes();
    while (!bDone) {
        if (nModeNum>=nNumModes) {
            // failed to find matching mode entry
            bDone = TRUE;
        }
        else if ((GetModeInfo(&gpeMode, nModeNum)==S_OK) &&
            (gpeMode.modeId==modeId)) {
            // found matching mode entry
            bDone = TRUE;
        }
        else {
            // check next entry
            nModeNum++;
        }
    }
    // check if mode number is valid
    if (nModeNum>=nNumModes) {
        return E_INVALIDARG;
    }

    // pass mode info to the hardware specific initialization
    // function
    SCODE scInit = ModeInit(&m_gpeModeTable[nModeNum]);

    if (scInit!=S_OK) {
        // something failed here, don't go any further
        return scInit;
    }
    // verify parameters ModeInit is required to initialize
    if ((m_nLAWPhysical==0) || (m_nLAWSize==0) || (m_nVideoMemorySize==0) ||
        (m_nScreenStride==0)) {
        // ERROR: ModeInit failed to properly initialize some data
        //        members that are required. These data members need
        //        to be initialized for this function to work.
        RETAILMSG(1, (TEXT("NewGPE::SetMode ERROR - ModeInit failed to initialize some data members!\r\n")));
        return E_FAIL;
    }

    // continue with hardware independent initialization
    //
    // Using the following values initialized by ModeInit:
    // * m_nLAWPhysical
    // * m_nLAWSize
    // * m_nVideoMemorySize
    // * m_nVideoMemoryStart
    // * m_nScreenStride
    // * m_bHWCursor (optional)
    // * m_b555Mode (optional)
    //
    // Initialize the remaining values:
    // * m_pMode
    // * m_nScreenWidth
    // * m_nScreenHeight
    // * m_pLAW
    // * m_p2DVideoMemory
    // * m_pPrimanrySurface
    // * m_ulBitMasks
    //
    // And don't forget to initialize the palette parameter
    // pPaletteHandle.
    //

    // initialize remaining values
    //
    m_pMode = &m_gpeModeTable[nModeNum];  // current mode
    m_nScreenWidth = m_pMode->width;    // current display width
    m_nScreenHeight = m_pMode->height;  // current display height

    // generate pointer to linear address window
    // (can't address physical memory directly, but need to
    //  create and map pointer)
    m_pLAW = (unsigned char *) VirtualAlloc(NULL, m_nLAWSize, MEM_RESERVE, PAGE_NOACCESS);
    BOOL bCreateLAW;
    if (m_nLAWPhysical<0x20000000) { // handle <512MB address differently
        bCreateLAW = VirtualCopy(m_pLAW, (LPVOID) (m_nLAWPhysical|0x80000000), m_nLAWSize,
            PAGE_READWRITE|PAGE_NOCACHE);
    }
    else {
        bCreateLAW = VirtualCopy(m_pLAW, (LPVOID) (m_nLAWPhysical>>8), m_nLAWSize,
            PAGE_READWRITE|PAGE_NOCACHE|PAGE_PHYSICAL);
    }
    // check for error creating LAW pointer
    if (!bCreateLAW) {
        RETAILMSG(1, (TEXT("NewGPE::SetMode ERROR - Failed creating LAW pointer!\r\n")));
        return E_FAIL;
    }

    // create Node2D instance (for managing video memory)
    m_p2DVideoMemory  = new Node2D(m_nScreenWidth, m_nVideoMemorySize/m_nScreenStride,
        0, 0, 32/m_pMode->Bpp);
    // check for error creating Node2D instance
    if (m_p2DVideoMemory==NULL) {
        RETAILMSG(1, (TEXT("NewGPE::SetMode ERROR - Failed creating Node2D instance!\r\n")));
        return E_FAIL;
    }

    // create primary surface
    SCODE scAlloc = AllocSurface(
        &m_pPrimarySurface,
        m_nScreenWidth,
        m_nScreenHeight,
        m_pMode->format,
        GPE_REQUIRE_VIDEO_MEMORY);
    // check for error allocating primary surface
    if (scAlloc!=S_OK) {
        RETAILMSG(1, (TEXT("NewGPE::SetMode ERROR - Failed creating primary surface!\r\n")));
        return E_FAIL;
    }

    // initialize bitmasks based on current bits per pixel
    memset(&m_ulBitMasks[0], 0, sizeof(m_ulBitMasks));
    if (m_pMode->Bpp==16) {
        // 16Bpp has two mask options
        if (m_b555Mode) {   // 5-5-5 mode
            m_ulBitMasks[0] = 0x7c00; // red
            m_ulBitMasks[1] = 0x03e0; // green
            m_ulBitMasks[2] = 0x001f; // blue
        }
        else {              // 5-6-5 mode
            m_ulBitMasks[0] = 0xf800; // red
            m_ulBitMasks[1] = 0x07e0; // green
            m_ulBitMasks[2] = 0x001f; // blue
        }
    }

    // create a default palette, if necessary
    if (pPaletteHandle!=NULL) {
        // create palette
        switch (m_pMode->Bpp) {
        case 24:
        case 32:
            *pPaletteHandle = EngCreatePalette (PAL_BGR, 0, NULL, 0, 0, 0);
            break;
        case 16:
            *pPaletteHandle = EngCreatePalette(PAL_BITFIELDS, 0, NULL, m_ulBitMasks[0], m_ulBitMasks[1], m_ulBitMasks[2]);
            break;
        case 8:
            *pPaletteHandle = EngCreatePalette (PAL_INDEXED, 256, (ULONG *)m_rgbIdentityPal, 0, 0, 0);
            SetPalette(m_rgbIdentityPal, 0, 256); // palette needs to be set here
            break;
        case 4:
            *pPaletteHandle = EngCreatePalette (PAL_INDEXED, 16, (ULONG *)m_rgbIdentityPal16, 0, 0, 0);
            SetPalette(m_rgbIdentityPal16, 0, 16); // palette needs to be set
            break;
        default:
            RETAILMSG(1,(TEXT("NewGPE::SetMode Failed to create unknown palette type.\r\n")));
            break;
        }
    }

    // initialize m_bCursorShowing since cursor will initially be hidden
    // after mode set
    m_bCursorShowing = FALSE;

    return S_OK;
}

#ifdef TRUE
// example ModeInit function
//////////////////////////////////////////////////////////////
//
// NewGPE::ModeInit
//
// Handles the device specific portion of mode setting.
//
// PARAMETERS
// pMode        Pointer to information for mode to set
//
// RETURNS
// S_OK if successful
//
// NOTES
// This function must be implemented and certain data members
// properly initialized for the driver to set the display mode
// successfully. As well, this function must carry out the
// steps to set the hardware to the specified mode.
//
SCODE NewGPE::ModeInit(GPEMode *pMode)
{
    // This example sets the required data members for
    // support of 320x200 at 8Bpp, the mode that LOADCEPC,
    // the bootloader for the CEPC platform, initializes
    // the VGA display to.
    m_nLAWPhysical = 0x000A0000;    // LAW window for VGA
    m_nLAWSize = 0x00010000;        // LAW size for VGA
    m_nVideoMemorySize = 0x00010000;// video memory size for VGA (without banking)
    m_nVideoMemoryStart = 0;        // video memory at beginning of LAW
    m_nScreenStride = 320;          // byte width of display line for 320x200 mode
    
    // Here you would add code to actually put the display
    // device into the specified mode.

    return S_OK;
}
#endif