//////////////////////////////////////////////////////////////
//
// NewGPE.H
//
// NewGPE display driver class header file.
//
// This is the main header file for the NewGPE display driver
// class. It is derived from the default display driver class
// GPE.
//
//////////////////////////////////////////////////////////////

#ifndef _NEWGPE_H_
#define _NEWGPE_H_

#include <winddi.h>     // Windows DDI definitions
#include <gpe.h>        // GPE class definitions
#include "ioport.h"     // I/O port class definitions
// need this for

// List of I/O port IDs used as indices into an array of
// mapped port registers for standard VGA I/O registers.
// The ID name (such as PortId_3C0) matches the I/O location
// that will be mapped in the particular array entry (such as
// I/O address 3C0). The NewGPE class only uses the registers
// for default VGA support of palette setting and vertical
// blank detection.
//
// See the NewGPE class definition below and the class
// constructor in the NewGPE.CPP file.
// (taken from GPEVGA.H)
//
enum ENUM_VGAPortIDs{
	PortId_3BA,
	PortId_3C0,
	PortId_3C1,
	PortId_3C2,
	PortId_3C3,
	PortId_3C4,
	PortId_3C5,
	PortId_3C6,
	PortId_3C7,
	PortId_3C8,
	PortId_3C9,
	PortId_3CA,
	PortId_3CC,
	PortId_3CE,
	PortId_3CF,
	PortId_3D4,
	PortId_3D5,
	PortId_3DA,
	numVGAPortIDs
};

//////////////////////////////////////////////////////////////
//
// Class NewGPESurf
//
// This is the class used for system and video memory surfaces
// and is derived from GPESurf, which only supports surfaces
// in system memory. This simplifies the driver development by
// providing a base surface class that inherently supports
// video memory surfaces.
//

class NewGPESurf : public GPESurf
{
protected:
    Node2D      *m_pNode2D; // location of surface in video memory

public:
	// ------------
	//
	// Surf.cpp
	//

	// constructor NewGPESurf
	//
	// Default functionality is already implemented, including support
	// for creating surfaces in video memory. Only override or change
	// to add support for additional features.
                NewGPESurf(
                    int width,
                    int height,
                    void *pBits,
                    int stride,
                    EGPEFormat format,
                    int offset,
                    Node2D *pNode);

	// destructor NewGPESurf
	//
	// Default functionality is already implemented, including support
	// for creating surfaces in video memory. Only override or change
	// to add support for additional features.
    virtual     ~NewGPESurf();
};


//////////////////////////////////////////////////////////////
//
// Class NewGPE
//
// This class serves as the base class for device specific drivers
// Emulation functions for performing graphics operations by software
// are also placed here
//

class NewGPE : public GPE
{
protected:
    // these are declared in GPE class
//	GPESurf			*m_pPrimarySurface;
//	int				m_nScreenWidth;
//	int				m_nScreenHeight;
//	GPEMode			*m_pMode;
//	unsigned long	m_hSurf;	// hsurf of display surface

    // added these data members for video memory support
    Node2D          *m_p2DVideoMemory;  // video memory management
    int             m_nScreenStride;    // number of bytes in a display line
    unsigned long   m_nLAWPhysical;     // physical address of linear access window
    unsigned char   *m_pLAW;            // pointer to linear access window
    int             m_nLAWSize;         // linear access window size
    int             m_nVideoMemorySize; // video memory size
                                        // (linear window size can be larger than memory size)
    int             m_nVideoMemoryStart;// start of video memory in linear access window

    // added these data member for miscellaneous support
    BOOL            m_bIsVGADevice;     // flags device is VGA
    BOOL            m_bHWCursor;        // flags hardware cursor supported
    BOOL            m_b555Mode;         // flags 5 bits each for red, green, and blue
                                        // when in 16 bits per pixel mode. The default
                                        // is 5 bits each for red and blue and 6 bits
                                        // for green

    // added for software cursor support
    BOOL            m_bCursorShowing;

    // added to support VGA I/O registers
    //
    // NOTE:
    // I/O accesses are handled using the classes provided
    // in Platform Builder sample driver source code, in the
    // files IOPORT.CPP and IOPORT.H.
    // 
    PortRange       *m_pPortRanges;     // pointer to list of I/O ranges
    Port            m_VGAIOPort[numVGAPortIDs]; // array of mapped I/O ports
    // VGA I/O registers
public:
	// ------------
	//
	// Blt.cpp
	//

	// BltPrepare
	//
    // Prepare for blit operation.
    //
	// Default functionality is already implemented. Only override or
	// change for non-default behavior such as hardware acceleration.
	virtual SCODE	BltPrepare( GPEBltParms *pBltParms );

	// BltComplete
	//
    // Blit operation completed.
    //
	// Default functionality is already implemented. Only override or
	// change for non-default behavior such as hardware acceleration.
	virtual SCODE	BltComplete( GPEBltParms *pBltParms );

	// ------------
	//
	// Line.cpp
	//

	// Line
	//
    // Line drawing operation.
    //
	// Default functionality is already implemented. Only override or
	// change for non-default behavior such as hardware acceleration.
	virtual SCODE	Line(
						GPELineParms *pLineParms,
						EGPEPhase phase = gpeSingle );


	// ------------
	//
	// Surf.cpp
	//

	// AllocSurface
	//
    // Allocate system or video memory surface.
    //
	// Default functionality is already implemented, including support
	// for creating surfaces in video memory. Only override or change
	// to add support for additional features.
	virtual SCODE	AllocSurface(
						GPESurf **ppSurf,
						int width,
						int height,
						EGPEFormat format,
						int surfaceFlags );

	// ------------
	//
	// NewGPE.cpp
    //

	// NewGPE
	//
    // Constructor.
    //
    // Initializes data members to default values. Some of
    // these members may need to be initialized to different
    // values depending on the capabilities of the hardware.
					NewGPE();

	// ~NewGPE
	//
    // Destructor.
    //
    // Frees any allocated resources.
    virtual			~NewGPE();

    // NumModes
    //
    // Return number of supported modes.
    //
    // Default functionality is already implemented through the
    // m_gpeModeTable global variable.
	virtual int		NumModes();

    // GetModeInfo
    //
    // Return information on a specific supported mode.
    //
    // Default functionality is already implemented through the
    // m_gpeModeTable global variable.
	virtual SCODE	GetModeInfo(
						GPEMode *pMode,
						int modeNo );

    // SetMode
    //
    // Set the specified mode.
    //
    // Default functionality is already implemented for device
    // independent support. This function calls ModeInit() for
    // the device specific portion of mode setting.
	virtual SCODE	SetMode(
						int modeId,
						HPALETTE *pPaletteHandle );

    // ModeInit
    //
    // Handles the device specific portion of mode setting.
    //
    // See the sample ModeInit function in the NewGPE.CPP file
    // for more information on the specific data members that
    // must be initialized by this function.
    virtual SCODE   ModeInit(GPEMode *pMode);

	// ------------
	//
	// Misc.cpp
    //

    // SetPalette
    //
    // Sets the palette for display modes that support a
    // palette.
    //
    // The default implementation supports palette setting
    // for VGA devices. Non-VGA devices must add device
    // specific support or override this function in a
    // derived class.
	virtual SCODE	SetPalette(
						const PALETTEENTRY *src,
						unsigned short firstEntry,
                        unsigned short numEntries );

    // InVBlank
    //
    // Detects if display is in vertical blanking period.
    //
    // The default implementation supports vertical blank
    // detection for VGA devices. Non-VGA devices must add
    // device specific support or override this function
    // in a derived class.
    virtual int		InVBlank();

	virtual SCODE	SetPointerShape(
						GPESurf *pMask,
						GPESurf *pColorSurf,
						int xHot,
						int yHot,
						int cx,
                        int cy ) {return S_OK;};
	virtual SCODE	MovePointer(
						int x,
                        int y ) {return S_OK;};

//	virtual int		IsPaletteSettable();
};

//////////////////////////////////////////////////////////////
//
// VGA register definitions
//
// Some general support is provided for VGA based devices.
// Only the subset of register definitions required for
// default VGA palette setting and vertical blank detection
// are included below. For more information or to add support
// for additional VGA registers not included here, see the
// GPEVGA.CPP and GPEVGA.H files that are included with some
// of the VGA based drivers in Platform Builder.
//
// Direct access IO mapped registers
#define reg_DAC_AD_MK		(PORT_RW_8 m_VGAIOPort[PortId_3C6 ]) //Pixel read mask (put FF here!)
#define reg_DAC_DATA		(PORT_RW_8 m_VGAIOPort[PortId_3C9 ]) //Access palette entry (use 3 times)
#define reg_DAC_RD_AD		(PORT_WO_8 m_VGAIOPort[PortId_3C7 ]) //Read index into palette (auto-incr)
#define reg_DAC_STS			(PORT_RO_8 m_VGAIOPort[PortId_3C7 ]) //DAC status - checks palette state
#define reg_DAC_WR_AD		(PORT_WO_8 m_VGAIOPort[PortId_3C8 ]) //Write index into palette (auto-incr)
#define reg_STATUS_0		(PORT_RO_8 m_VGAIOPort[PortId_3C2 ]) //CRT Irq Pending
#define reg_STATUS_1		(PORT_RO_8 m_VGAIOPort[PortId_3DA ]) //VSync Status

#endif // _NEWGPE_H_
