//////////////////////////////////////////////////////////////
//
// Surf.CPP
//
// This is the source file for the NewGPE display driver
// class that contains the functionality related to surfaces.
//
//////////////////////////////////////////////////////////////

///////////////////////////////////
//
// Includes
//

#include <windows.h>
#include <types.h>
#include "NewGPE.h"

///////////////////////////////////
//
// Constants
//

///////////////////////////////////
//
// Globals
//

///////////////////////////////////
//
// Function Prototypes
//

///////////////////////////////////
//
// Functions
//

//////////////////////////////////////////////////////////////
//
// NewGPE::AllocSurface
//
// Create a surface. This function overrides the base GPE
// class pure virtual function.
//
// PARAMETERS
// ppSurf       pointer to pointer to return created surface in
// width        width of surface to create
// height       height of surface to create
// format       pixel format of surface to create
// surfaceFlags flags for requirements of surface creation, such as video memory
//              process
//
// RETURNS
// S_OK if successful, otherwise an error code
//
// NOTES
// Default functionality is already implemented, including
// support for creating surfaces in video memory. Only
// override or change to add support for additional features.
//
SCODE NewGPE::AllocSurface(
						GPESurf **ppSurf,
						int width,
						int height,
						EGPEFormat format,
						int surfaceFlags )
{
    SCODE scRet = S_OK;
    *ppSurf = NULL;

    // check parameters are valid:
    //   video memory surface must have same pixel format as display
    if (surfaceFlags & GPE_REQUIRE_VIDEO_MEMORY) {
        // video memory surface must have same pixel format as display
        if (format!=m_pMode->format)
           return E_INVALIDARG;
    }

	// check if video memory surface requested
    if ((surfaceFlags & GPE_REQUIRE_VIDEO_MEMORY) ||
        ((surfaceFlags & GPE_PREFER_VIDEO_MEMORY) &&
        (format==m_pMode->format))) {
        // try allocating out of video memory
        Node2D *pNode = m_p2DVideoMemory->Alloc(width, height);

        if (pNode!=NULL) {
            // get offset into video memory
            DWORD dwOffset = (m_nScreenStride * pNode->Top()) +
                ((pNode->Left() * EGPEFormatToBpp[format]) / 8);

            // now create a surface for the allocated video memory
            *ppSurf = new NewGPESurf(
                width,
                height,
                m_pLAW + dwOffset,
                m_nScreenStride,
                format,
                dwOffset,
                pNode);
            if (*ppSurf!=NULL) {
                // video memory surface allocated successfully
                return S_OK;
            }

            // failed creating surface for video memory
            delete pNode; // free video memory
        }

        // if we got here then unable to allocate out of video memory
        if (surfaceFlags & GPE_REQUIRE_VIDEO_MEMORY)
            return E_OUTOFMEMORY; // fail since surface requires video memory
    }

    // Allocate surface from system memory if not in video memory
    *ppSurf = new GPESurf(width, height, format);
    if (*ppSurf!=NULL) {
        // make sure system memory allocated
        if ((*ppSurf)->Buffer()!=NULL) {
            // system memory surface allocated successfully
            return S_OK;
        }

        // system memory surface allocation failed
        delete *ppSurf;
        // just fall through to fail
    }

    // if we got here then something failed
    return E_OUTOFMEMORY;
}

//////////////////////////////////////////////////////////////
//
// NewGPESurf::NewGPESurf
//
// Constructor for NewGPESurf class. This class overrides
// the GPESurf class to support the creation of surfaces
// in video memory.
//
// PARAMETERS
// width        width of surface to create
// height       height of surface to create
// pBits        pointer to memory allocated for surface
// stride       number of bytes per line in memory
// format       pixel format of surface to create
// pNode        video memory holder object
//
// RETURNS
// none.
//
// NOTES
// Default functionality is already implemented, including
// support for creating surfaces in video memory. Only
// override or change to add support for additional features.
//
NewGPESurf::NewGPESurf(
                    int width,
                    int height,
                    void *pBits,
                    int stride,
                    EGPEFormat format,
                    int offset,
                    Node2D *pNode)
{
    // call general GPESurf initialization
    Init(width, height, pBits, stride, format);
    m_pNode2D = pNode;

    if (pNode!=NULL) {
        // surface in video memory, set flags and parameters
        m_fInVideoMemory = TRUE;
    }
}

//////////////////////////////////////////////////////////////
//
// NewGPESurf::~NewGPESurf
//
// Destructor for NewGPESurf class. This class overrides
// the GPESurf class to support the creation of surfaces
// in video memory.
//
// PARAMETERS
// none.
//
// RETURNS
// none.
//
// NOTES
// Default functionality is already implemented, including
// support for creating surfaces in video memory. Only
// override or change to add support for additional features.
//
NewGPESurf::~NewGPESurf()
{
    // free video memory if applicable
    if (m_fInVideoMemory && (m_pNode2D!=NULL)) {
        delete m_pNode2D;
    }
}
