//
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Generic Graph Component Library
//
// You should have received a copy of the License Agreement for the
// Generic Graph Component Library along with the software;  see the
// file LICENSE.  If not, contact Office of Research, University of Notre
// Dame, Notre Dame, IN  46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//
#ifndef GGCL_ADJACENCY_LIST_H
#define GGCL_ADJACENCY_LIST_H

#include <utility>
#include <boost/graph_structure.hpp>
#include <boost/int_iterator.hpp>
#include <ggcl/utils.h>

#include <ggcl/ggcl_config.h>
#include <ggcl/directed_category_traits.h>
#include <ggcl/graph_rep_gen.h>
#include <ggcl/properties.h>

#include <ggcl/detail/vertex.h>
#include <ggcl/detail/edge.h>
#include <ggcl/detail/incidence_iterator.h>
#include <ggcl/detail/adjacency_iterator.h>
#include <ggcl/detail/edge_iterator.h>

namespace ggcl {

  template <class Derived> class adj_list_base { };

  namespace detail {
    template <class Plugin, class PluginStorageRef, class Tag>
    struct adj_list_vertex_property_accessor;
    template <class Directed, class Plugin, class Tag>
    struct adj_list_edge_property_accessor;

    template <class Tag, class Plugin> struct adj_list_choose_vertex_pa;
    template <class VertexPlugin> struct vertex_plugin_storage;
    template <class DS, class EP, class ST> struct adjlist_edge_plugin_gen;
  } // namespace detail

  struct adj_list_tag { };

  template <class RepS = vecS, 
            class DirectedS = directedS,
            class VertexPlugin = no_plugin,
            class EdgePlugin = no_plugin>
  class adjacency_list
    : public adj_list_base< 
        adjacency_list<RepS,DirectedS,VertexPlugin,EdgePlugin> >,
        public detail::vertex_plugin_storage<VertexPlugin>
  {
    typedef adjacency_list  self;
    typedef detail::vertex_plugin_storage<VertexPlugin> Storage;
  public:
    typedef std::size_t                                    size_type;
    typedef typename directed_category_traits<DirectedS>
      ::directed_category                                  directed_category;
  protected:
    typedef size_type                                      Vertex;

  public:
    // If the graph is undirected, put the edge plugins in a separate
    // linked list with pointers into it from the out-edge lists. If
    // the graph is directed, put the edge plugins directly in the
    // out-edge lists. If there are no edge plugins just store the
    // the target vertex ID.
    typedef detail::adjlist_edge_plugin_gen
      <DirectedS,EdgePlugin,size_type> EdgePluginGen;
    typedef typename EdgePluginGen::type StoredEdge;

    typedef bidir_edge<directed_category,size_type,EdgePlugin>  Edge;
  public:

    // Figure out what kind of storage we are using for the backbone
    // and for the out-edge lists.
#if defined GGCL_PARTIAL_SPECIALIZATION
    typedef graph_rep_gen<RepS, StoredEdge> RepGen;
#else
    typedef typename graph_rep_gen<RepS>:: GGCL_TEMPLATE bind<StoredEdge>
      RepGen;
#endif
    typedef typename RepGen::TwoD graph_rep_type;
    typedef typename RepGen::edge_parallel_category edge_parallel_category;

  public:
    typedef typename graph_rep_type::const_iterator        Iterator2D;
    typedef typename graph_rep_type::value_type::const_iterator Iterator1D;
  public:
    typedef typename Storage::type                        VertexPluginStorage;
    typedef const typename Storage::type*                 StoredVertexPtr;
    typedef typename EdgePluginGen::plugin_container PluginContainer;
  public:
    typedef Vertex                                         vertex_descriptor;
    typedef Edge                                           edge_descriptor;

    typedef void                                           in_edge_iterator;
    typedef bidir_incidence_iterator
      <Edge, Iterator1D, Iterator2D, out_edge_tag>         out_edge_iterator;
    typedef bidir_adjacency_iterator
      <Vertex, Iterator1D, Iterator2D, StoredVertexPtr>    adjacency_iterator;
    typedef boost::int_iterator<vertex_descriptor>         vertex_iterator;
    typedef bidir_edge_iterator
      <Iterator2D, out_edge_iterator>                      edge_iterator;

    typedef EdgePlugin edge_plugin_type;
    typedef VertexPlugin vertex_plugin_type;
    typedef adj_list_tag graph_tag;

    inline adjacency_list() 
      : Storage(0), m_out_edge_matrix(0) {}

    inline adjacency_list(size_type num_vertices) 
      : Storage(num_vertices), m_out_edge_matrix(num_vertices) { }

    template <class EdgeIterator>
    inline adjacency_list(size_type num_vertices,
                          EdgeIterator first, EdgeIterator last)
      : Storage(num_vertices), m_out_edge_matrix(num_vertices)
    {
      while (first != last) {
        add_edge(*this, (*first).first, (*first).second);
        ++first;
      }
    }

    template <class EdgeIterator, class EdgePropertyIterator>
    inline adjacency_list(size_type num_vertices,
                          EdgeIterator first, EdgeIterator last,
                          EdgePropertyIterator ep_iter)
      : Storage(num_vertices), 
        m_out_edge_matrix(num_vertices)
    {
      // this should be specialized when RepS == vecS
      while (first != last) {
        add_edge(*this, (*first).first, (*first).second, *ep_iter);
        ++first;
        ++ep_iter;
      }
    }
    //  protected:
    graph_rep_type     m_out_edge_matrix;
    PluginContainer m_edge_plugins;
  };

  struct adj_list_edge_property_selector {
    template <class Directed, class Plugin, class Tag>
    struct bind {
      typedef detail::adj_list_edge_property_accessor
         <Directed,Plugin,Tag> type;
      typedef type const_type;
    };
  };
  template <>  
  struct edge_property_selector<adj_list_tag> {
    typedef adj_list_edge_property_selector type;
  };


  struct adj_list_vertex_property_selector {
    template <class Plugin, class Tag>
    struct bind {
      typedef detail::adj_list_choose_vertex_pa<Tag,Plugin> Choice;
      typedef typename Choice::type type;
      typedef typename Choice::const_type const_type;
    };
  };
  template <>  
  struct vertex_property_selector<adj_list_tag> {
    typedef adj_list_vertex_property_selector type;
  };


  template <class Vertex, class Graph>
  inline
  std::pair<typename Graph::adjacency_iterator,
            typename Graph::adjacency_iterator>
  adj(Vertex u, const adj_list_base<Graph>& G_)
  {
    const Graph& G = static_cast<const Graph&>(G_);
    typedef typename Graph::adjacency_iterator aiter;
    return std::make_pair( aiter(G.m_out_edge_matrix[u].begin(),
                                 G.m_out_edge_matrix[u].end(),
                                 G.m_out_edge_matrix.begin(), &(G.m_vplugin)),
                           aiter(G.m_out_edge_matrix[u].end(),
                                 G.m_out_edge_matrix[u].end(),
                                 G.m_out_edge_matrix.begin(), &(G.m_vplugin)));
  }

  template <class Vertex, class Graph>
  inline
  std::pair<typename Graph::out_edge_iterator,
            typename Graph::out_edge_iterator>
  out_edges(Vertex u, const adj_list_base<Graph>& G_)
  {
    const Graph& G = static_cast<const Graph&>(G_);
    typedef typename Graph::out_edge_iterator iiter;
    return std::make_pair( iiter(G.m_out_edge_matrix[u].begin(),
                                 G.m_out_edge_matrix.begin()+u,
                                 G.m_out_edge_matrix.begin()),
                           iiter(G.m_out_edge_matrix[u].end(),
                                 G.m_out_edge_matrix.begin()+u,
                                 G.m_out_edge_matrix.begin()) );
  }

  template <class Graph>
  inline 
  std::pair<typename Graph::vertex_iterator, 
            typename Graph::vertex_iterator>
  vertices(const adj_list_base<Graph>& G_)
  {
    const Graph& G = static_cast<const Graph&>(G_);
    typedef typename Graph::vertex_iterator viter;
    return std::make_pair( viter(0), viter(G.m_out_edge_matrix.size()) );
  }
  
  //If a graph is undirected, traverse edges once will traverse all
  //the edges twice.
  template <class Graph>
  inline
  std::pair<typename Graph::edge_iterator, 
            typename Graph::edge_iterator>
  edges(const adj_list_base<Graph>& G_) {
    const Graph& G = static_cast<const Graph&>(G_);
    typedef typename Graph::edge_iterator eiter;
    return std::make_pair( eiter(G.m_out_edge_matrix.begin(), 
                                 G.m_out_edge_matrix.begin(),
                                 G.m_out_edge_matrix.end()),
                           eiter(G.m_out_edge_matrix.begin(),
                                 G.m_out_edge_matrix.end(), 
                                 G.m_out_edge_matrix.end()) );
  }

  template <class DS, class ST, class EP, class Graph>
  inline typename Graph::vertex_descriptor
  source(bidir_edge<DS,ST,EP> e, const adj_list_base<Graph>& G_)
  {
    return e.m_source;
  }

  template <class DS, class ST, class EP, class Graph>
  inline typename Graph::vertex_descriptor
  target(bidir_edge<DS,ST,EP> e, const adj_list_base<Graph>& G_)
  {
    return e.m_target;
  }
  
  template <class Graph>
  inline 
  typename Graph::size_type
  num_vertices(const adj_list_base<Graph>& G_) 
  {
    const Graph& G = static_cast<const Graph&>(G_);
    return G.m_out_edge_matrix.size();
  }

  // should replace this with using a counter so this
  // is O(1)... or was there a good reason for not doing that?
  // hmm, this implementation not very accurate for undirected...
  // but it is consistent with edges(g)
  template <class Graph>
  inline 
  typename Graph::size_type
  num_edges(const adj_list_base<Graph>& G_) 
  {
    const Graph& G = static_cast<const Graph&>(G_);
    typename Graph::size_type e = 0;
    typename Graph::edge_iterator i,end;
    for (boost::tie(i,end) = edges(G); i != end; ++i)
      ++e;
    return e;
  }

  // Deprecate this? -JGS
  template <class Graph>
  inline
  typename Graph::vertex_descriptor
  vertex(int v, const adj_list_base<Graph>&) 
  {
    return v;
  }

  template<class Vertex, class Graph>
  inline typename Graph::size_type
  out_degree(Vertex u, const adj_list_base<Graph>& G_) {
    const Graph& G = static_cast<const Graph&>(G_);
    return G.m_out_edge_matrix[u].size();
  }

} /* namespace ggcl */

#include "ggcl/detail/adjacency_list.cct"

namespace ggcl {

  template <class RS, class DS, class VP, class EP, class Tag>
  typename vertex_property_accessor< adjacency_list<RS,DS,VP,EP>, Tag>::type
  get_vertex_property(adjacency_list<RS,DS,VP,EP>& g, Tag)
  {
    typedef detail::vertex_pa_dispatcher<Tag> Dispatcher;
    return Dispatcher::get_vertex_property(g);
  }
  template <class RS, class DS, class VP, class EP, class Tag>
  typename vertex_property_accessor< adjacency_list<RS,DS,VP,EP>,
    Tag>::const_type
  get_vertex_property(const adjacency_list<RS,DS,VP,EP>& g, Tag)
  {
    typedef detail::vertex_pa_dispatcher<Tag> Dispatcher;
    return Dispatcher::const_get_vertex_property(g);
  }

  template <class RS, class DS, class VP, class EP, class Tag>
  detail::adj_list_edge_property_accessor<
   typename directed_category_traits<DS>::directed_category, EP,Tag>
  get_edge_property(const adjacency_list<RS,DS,VP,EP>&, Tag) {
    typedef typename directed_category_traits<DS>::directed_category Cat;
    return detail::adj_list_edge_property_accessor<Cat,EP,Tag>();
  }

} // namespace ggcl

#endif /*GGCL_ADJACENCY_LIST_H*/
