#ifndef GGCL_ARRAY_LISTS_H
#define GGCL_ARRAY_LISTS_H

#include <iosfwd>
#include <iterator>
#include <boost/iterator.hpp>
#include <boost/operators.hpp>

// These are circular linked lists, so iterating through them is a bit
// weird right now. See the do while loops in fibonachi_heap.h :(

// Interface Summary:
//
// L l(n);
// l.new_list(x)
// l.insert(x,head)
// l.remove(x,head)
// l.concat(x,y);
// l[x].begin()
// l[x].end()
// l.nil()
// l.size()

template <class Next,
          class T = std::iterator_traits<Next>::value_type, 
          class Ref = std::iterator_traits<Next>::reference,
          class Ptr = std::iterator_traits<Next>::pointer,
          class DiffT = std::iterator_traits<Next>::difference_type
         >
class array_list_iterator
  : public boost::forward_iterator_helper<
      array_list_iterator<Next,T,Ref,Ptr,DiffT>,
      T, DiffT, Ptr, Ref>
{
  typedef array_list_iterator self;
public:
  inline array_list_iterator(T i, Next n)
    : _i(i), _next(n) { }
  inline Ref operator*() const { return _i;  }
  inline self& operator++() { _i = _next[_i]; return *this;  }
  friend inline bool operator==(const self& x, const self& y) { 
    return x._i == y._i; }
protected:
  T _i;
  Next _next;
};

namespace ggcl { 

// multiple lists represented in an array
template <class T>
struct array_lists {
  typedef std::vector<T> Vec;
  typedef typename Vec::const_iterator Next;
  typedef array_list_iterator<Next> list_iter;
public:

  struct list_type {
    typedef list_iter iterator;
    list_type(T r, Next n, T nil_) : root(r), next(n), nil(nil_) { }
    iterator begin() const { return iterator(root, next); }
    iterator end() const { return iterator(root, next); }
    T root;
    Next next;
    T nil;
    void print(std::ostream& os) {
      if (root == nil)
	os << "is empty" << endl;
      else {
	iterator xi = begin(), xend = end();
	do {
	  os << *xi << " ";
	} while (++xi != xend);
	os << endl;
      }
    }
  };
  typedef list_type value_type;
  typedef list_type reference;
  typedef T size_type;

  inline array_lists(size_type n) : next(n), prev(n) { }

  inline void new_list(T x) {
    next[x] = x;
    prev[x] = x;
  }

  inline void insert(T x, T& head)
  {
    if (head == nil()) {
      head = x;
      new_list(x);
    } else {
      next[x] = next[head];
      prev[x] = head;
      prev[next[head]] = x;
      next[head] = x;
    }
  }
  inline void remove(T x, T& head)
  {
    if (x == prev[x])
      head = nil();
    else {
      if (head == x)
        head = prev[x];
      T next_x = next[x];
      T prev_x = prev[x];
      prev[next_x] = prev_x;
      next[prev_x] = next_x;
    }
  }

  // next[y] <--> y <--> prev[y], next[x] <--> x <--> prev[x]
  // next[y] <--> x <-...-> next[x] <--> y
  inline void concat(T x, T y)
  {
    T next_x = next[x];
    prev[next_x] = y;
    next[x] = next[y];
    prev[next[y]] = x;
    next[y] = next_x;
  }

  value_type operator[](T root) { 
    return list_type(root, next.begin(), nil()); 
  }
  inline T nil() const { return next.size(); }  
  inline T size() const { return next.size(); }
  //protected:
  std::vector<T> next, prev;  
};



} // namespace ggcl 

#endif /* GGCL_ARRAY_LISTS_H */
