//
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Generic Graph Component Library
//
// You should have received a copy of the License Agreement for the
// Generic Graph Component Library along with the software;  see the
// file LICENSE.  If not, contact Office of Research, University of Notre
// Dame, Notre Dame, IN  46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//
#ifndef GGCL_BEST_FIRST_SEARCH_H
#define GGCL_BEST_FIRST_SEARCH_H

#include <utility>

#include <boost/graph_structure.hpp>
#include <ggcl/ggcl_config.h>
#include <ggcl/indirect_cmp.h>
#include <ggcl/mutable_queue.h>
#include <ggcl/properties.h>
#include <ggcl/weighted_edge_visitor.h>
#include <ggcl/bfs.h>

namespace ggcl {

  template <class Graph, class Vertex, 
            class Distance, class Weight, class Color, class ID,
            class BinaryPredicate, class BinaryFunction,
            class Visitor>
  void
  _best_first_search(Graph& G, Vertex  s,
                     Distance d, Weight w, Color color, ID id, 
                     BinaryPredicate compare, BinaryFunction combine,
                     Visitor visit, bool init_graph)
  {
    REQUIRE(Graph, VertexListGraph);
    REQUIRE2(Distance, Vertex, ReadWritePropertyAccessor);
    typedef typename boost::graph_traits<Graph>::edge_descriptor Edge;
    REQUIRE2(Weight, Edge, ReadablePropertyAccessor);
    REQUIRE2(ID, Vertex, ReadablePropertyAccessor);
    typedef typename boost::property_traits<Distance>::value_type D_value;
    typedef typename boost::property_traits<Weight>::value_type W_value;
    typedef typename boost::property_traits<ID>::value_type ID_value;
    REQUIRE3(BinaryPredicate, D_value, D_value, BinaryPredicate);
    REQUIRE4(BinaryFunction, D_value, D_value, W_value, BinaryFunction);
    REQUIRE(ID_value, Integer);
    REQUIRE4(Visitor, Vertex, Edge, Graph, UserVisitor);
    
    typedef typename boost::property_traits<Distance>::value_type D;
    typedef indirect_cmp<Distance, BinaryPredicate > IndirectCmp;
    IndirectCmp icmp(d, compare);

    mutable_queue<Vertex, std::vector<Vertex>, IndirectCmp, ID>
      Q(num_vertices(G), icmp, id);

    typedef typename boost::property_traits<Weight>::value_type W;
    weighted_edge_visitor<Weight, Distance, Color, Visitor, 
                          BinaryFunction, BinaryPredicate>
      visitor(w, d, color, combine, compare, visit);
    
    if (init_graph)
      initialize_graph(G, visitor);

    breadth_first_search(G, s, Q, visitor, queue_update());
  }

  // Variant (3)
  template <class Graph, class Vertex, 
            class Distance, class Weight, class Color, class ID,
            class BinaryPredicate, class BinaryFunction,
            class Visitor>
  inline void
  best_first_search(Graph& G, Vertex s, 
                    Distance d, Weight w, Color c, ID id,
                    BinaryPredicate compare, BinaryFunction combine,
                    Visitor visit, bool init_graph = true)
  {
    typename boost::graph_traits<Graph>::edge_descriptor 
      e = *out_edges(s, G).first;
    _best_first_search(G, s, 
                       adapt_property_accessor(d, boost::get(d,s), s),
                       adapt_property_accessor(w, boost::get(w,e), e),
                       adapt_property_accessor(c, boost::get(c,s), s),
                       adapt_property_accessor(id, boost::get(id,s), s),
                       compare, combine,
                       visit, init_graph);
  }

  // Variant (2)
  template <class Graph, class Vertex, class Distance, class Weight,
            class BinaryPredicate, class BinaryFunction>
  inline void
  best_first_search(Graph& G, Vertex s, 
                    Distance d, Weight w, 
                    BinaryPredicate compare, BinaryFunction combine,
		    bool init_graph = true)
  {
    best_first_search(G, s, d, w, 
                      get_vertex_property(G, color_tag()),
                      get_vertex_property(G, id_tag()),
                      compare, combine, null_visitor(), init_graph);
  }

  // Variant (1)
  template <class Graph, class Vertex, 
            class BinaryPredicate, class BinaryFunction>
  inline void
  best_first_search(Graph& G, Vertex s, 
                    BinaryPredicate compare, BinaryFunction combine,
		    bool init_graph = true)
  {
    typedef typename boost::graph_traits<Graph>::edge_descriptor Edge;
    typedef typename boost::graph_traits<Graph>::vertex_descriptor V;
    best_first_search(G, s, 
                      get_vertex_property(G, distance_tag()), 
                      get_edge_property(G, weight_tag()), 
                      compare, combine, init_graph);
  }

} /*namespace*/

#endif /* GGCL_BEST_FIRST_SEARCH_H */
