//
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Generic Graph Component Library
//
// You should have received a copy of the License Agreement for the
// Generic Graph Component Library along with the software;  see the
// file LICENSE.  If not, contact Office of Research, University of Notre
// Dame, Notre Dame, IN  46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//
#ifndef GGCL_BFS_H
#define GGCL_BFS_H

/*

  Breadth First Search Algorithm (Cormen, Leiserson, and Rivest p. 470)

  implementation note: the vertex arguments are not const in
    the operator[]'s since the color and distance
    may be stored in the vertex

*/
#include <boost/graph_structure.hpp>

#include <ggcl/ggcl_config.h>
#include <ggcl/ggcl_concepts.h>
#include <boost/queue.hpp>

#include <ggcl/bfs_visitor.h>
#include <ggcl/visited.h>

namespace ggcl {

  template <class VertexListGraph, class Visitor>
  void initialize_graph(VertexListGraph& G, Visitor visit)
  {
    typedef typename boost::graph_traits<VertexListGraph>
      ::vertex_descriptor Vertex;
    typedef typename boost::graph_traits<VertexListGraph>
      ::edge_descriptor Edge;
    REQUIRE(VertexListGraph, VertexListGraph);
    REQUIRE4(Visitor, Vertex, Edge, VertexListGraph, Visitor);
    typename boost::graph_traits<VertexListGraph>::vertex_iterator ui, ui_end;
    for (boost::tie(ui, ui_end) = vertices(G); ui != ui_end; ++ui) {
      visit.initialize(*ui);
    }
  }

  /* Implementation note: Using this generic BFS with stack is able to
     traverse through graph in DFS way. However, it is not correct for
     calculating discover time and finish time. */

  // Variant (3)
  template <class IncidenceGraph, class Vertex, class Buffer, 
            class Visitor, class VisitedFunc>
  void
  breadth_first_search(IncidenceGraph& G, Vertex s, Buffer& Q, 
		       Visitor visit, VisitedFunc visited)
  {
    typedef typename boost::graph_traits<IncidenceGraph>
      ::edge_descriptor Edge;
    REQUIRE(IncidenceGraph, IncidenceGraph);
    REQUIRE4(Visitor, Vertex, Edge, IncidenceGraph, Visitor);
    REQUIRE(Buffer, Buffer);

    visit.start(s);
    Q.push(s);
    while (! Q.empty()) {
      Vertex u = Q.top();
      Q.pop(); /*pop before push to avoid problem if Q is priority_queue.*/
      visit.discover(u);
      typename boost::graph_traits<IncidenceGraph>::out_edge_iterator
        ei, ei_end;
      for ( boost::tie(ei, ei_end) = out_edges(u, G); ei != ei_end; ++ei) {
        Edge e = *ei;
        if (visit.process(e, G)) {
          Q.push(target(e, G));
        } else 
          visited(visit, Q, target(e,G));
      }
      visit.finish(u);
    }
  }

  template <class Graph, class Vertex, class Buffer, class Visitor, 
            class Color >
  inline void
  _breadth_first_search(Graph& G, Vertex s, Buffer& Q, 
			Visitor visit, Color c)
  {
    REQUIRE2(Color, Vertex, ReadWritePropertyAccessor);
    typedef typename boost::property_traits<Color>::value_type ColorValue;
    REQUIRE(ColorValue, ColorValue);
    bfs_visitor<Color,Visitor> visitor(c,visit);
    initialize_graph(G, visitor);
    breadth_first_search(G, s, Q, visitor, null_operation() );
  }

  // Variant (2)
  template <class Graph, class Vertex, class Visitor, class Color>
  inline void
  breadth_first_search(Graph& G, Vertex s, Visitor visit, Color c)
  {
    typedef typename boost::graph_traits<Graph>::edge_descriptor Edge;
    typedef typename boost::graph_traits<Graph>::vertex_descriptor V;
    REQUIRE4(Visitor, V, Edge, Graph, UserVisitor);
    boost::queue<V> Q;
    _breadth_first_search(G, s, Q, visit, 
			 adapt_property_accessor(c, boost::get(c,s), s));
  }

  // Variant (1)
  template <class Graph, class Vertex, class Visitor>
  inline void
  breadth_first_search(Graph& G, Vertex s, Visitor visit)
  {
    breadth_first_search(G, s, visit, get_vertex_property(G, color_tag()));
  }

} /* namespace bfs */

#endif /* GGCL_BFS_H */

