//
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Generic Graph Component Library
//
// You should have received a copy of the License Agreement for the
// Generic Graph Component Library along with the software;  see the
// file LICENSE.  If not, contact Office of Research, University of Notre
// Dame, Notre Dame, IN  46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//
#ifndef BIDIR_ADJ_LIST_H
#define BIDIR_ADJ_LIST_H

#include <boost/graph_structure.hpp>

#include <ggcl/ggcl_config.h>
#include <ggcl/directed_category_traits.h>
#include <ggcl/graph_rep_gen.h>
#include <ggcl/utils.h>

#include <ggcl/adjacency_list.h>

#include <ggcl/detail/adjacency_iterator.h>
#include <ggcl/detail/incidence_iterator.h>
#include <ggcl/detail/edge_iterator.h>

#include <ggcl/detail/vertex.h>
#include <ggcl/detail/edge.h>


namespace ggcl {

  namespace detail {
    template <class EdgePlugin>
    struct bidir_edge_plugin_gen;
  }

  struct bidir_adj_list_tag { };

  template <class RepS = vecS,
            class VertexPlugin = no_plugin,
            class EdgePlugin = no_plugin>
  class bidir_adj_list 
    : public adj_list_base< bidir_adj_list<RepS,VertexPlugin,EdgePlugin> >,
      public detail::vertex_plugin_storage<VertexPlugin>
  {
    typedef bidir_adj_list self;
    typedef detail::vertex_plugin_storage<VertexPlugin> VStorageGen;
  public:
    typedef std::size_t                                    size_type;

    typedef detail::bidir_edge_plugin_gen<EdgePlugin> EdgePluginGen;

    typedef typename EdgePluginGen::out_type StoredOutEdge;
    typedef typename EdgePluginGen::in_type StoredInEdge;

    typedef EdgePlugin edge_plugin_type;
    typedef VertexPlugin vertex_plugin_type;
    typedef bidir_adj_list_tag graph_tag;

#if defined GGCL_PARTIAL_SPECIALIZATION
    typedef graph_rep_gen<RepS, StoredInEdge> InRepGen;
    typedef graph_rep_gen<RepS, StoredOutEdge> OutRepGen;
#else
    typedef graph_rep_gen<RepS>:: GGCL_TEMPLATE bind<StoredInEdge> InRepGen;
    typedef graph_rep_gen<RepS>:: GGCL_TEMPLATE bind<StoredOutEdge> OutRepGen;
#endif
  public:
    typedef typename InRepGen::TwoD                        in_edge_rep_type;
    typedef typename OutRepGen::TwoD                       out_edge_rep_type;

    typedef out_edge_rep_type                              graph_rep_type;

    typedef typename VStorageGen::type                   VertexPluginStorage;
    typedef const VertexPluginStorage*                   VPtr;
  public:
    typedef typename in_edge_rep_type::const_iterator      InIter2D;
    typedef typename in_edge_rep_type::value_type::const_iterator InIter1D;
    typedef typename out_edge_rep_type::const_iterator     OutIter2D;
    typedef typename out_edge_rep_type::value_type::const_iterator OutIter1D;
  public:

    typedef typename OutRepGen::edge_parallel_category edge_parallel_category;
    typedef directed_tag                                   directed_category;

    typedef size_type                                      vertex_descriptor;
    typedef bidir_edge<directed_category,size_type,EdgePlugin> edge_descriptor;
    typedef bidir_incidence_iterator
      <edge_descriptor, InIter1D, InIter2D, in_edge_tag>   in_edge_iterator;
    typedef bidir_incidence_iterator
      <edge_descriptor, OutIter1D, OutIter2D, out_edge_tag> out_edge_iterator;
    typedef bidir_adjacency_iterator
      <vertex_descriptor, OutIter1D, OutIter2D, VPtr>      adjacency_iterator;
    typedef boost::int_iterator<vertex_descriptor>         vertex_iterator;
    typedef bidir_edge_iterator
      <OutIter2D, out_edge_iterator>                       edge_iterator;

    inline bidir_adj_list()
      : VStorageGen(0), m_in_edge_matrix(0), m_out_edge_matrix(0) { }

    inline bidir_adj_list(size_type num_vertices)
      : VStorageGen(num_vertices), m_in_edge_matrix(num_vertices),
	m_out_edge_matrix(num_vertices) { }
    
    template <class EdgeIterator>
    inline bidir_adj_list(size_type num_vertices,
			  EdgeIterator first, EdgeIterator last)
      : VStorageGen(num_vertices), m_in_edge_matrix(num_vertices),
	m_out_edge_matrix(num_vertices)
    {
      while (first != last) {
	add_edge(*this, (*first).first, (*first).second);
	++first;
      }
    }

    template <class EdgeIterator, class EdgePropertyIterator>
    inline bidir_adj_list(size_type num_vertices,
			  EdgeIterator first, EdgeIterator last,
			  EdgePropertyIterator ep_iter)
      : VStorageGen(num_vertices), m_in_edge_matrix(num_vertices),
	m_out_edge_matrix(num_vertices)
    {
      while (first != last) {
	add_edge(*this, (*first).first, (*first).second, *ep_iter);
	++first;
	++ep_iter;
      }
    }
    //  protected:
    in_edge_rep_type m_in_edge_matrix;
    out_edge_rep_type m_out_edge_matrix;
    typename EdgePluginGen::plugin_container m_edge_plugins;
  };

  template <>  
  struct edge_property_selector<bidir_adj_list_tag> {
    typedef adj_list_edge_property_selector type;
  };
  template <>  
  struct vertex_property_selector<bidir_adj_list_tag> {
    typedef adj_list_vertex_property_selector type;
  };


  template <class Vertex, class RepS, class VPlugin, class EPlugin>
  std::pair<typename bidir_adj_list<RepS,VPlugin,EPlugin>::in_edge_iterator,
            typename bidir_adj_list<RepS,VPlugin,EPlugin>::in_edge_iterator>
  in_edges(Vertex u, const bidir_adj_list<RepS, VPlugin,EPlugin>& G);
    
  template <class Vertex, class RepS, class VPlugin, class EPlugin, class T>
  std::pair<typename bidir_adj_list<RepS,VPlugin,EPlugin>
            ::edge_descriptor,bool>
  add_edge(bidir_adj_list<RepS,VPlugin,EPlugin>& g,
           Vertex u, Vertex v, const T& stored_edge);

  template <class Vertex, class RepS, class VPlugin, class EPlugin>
  std::pair<typename bidir_adj_list<RepS,VPlugin,EPlugin>
            ::edge_descriptor,bool>
  add_edge(bidir_adj_list<RepS,VPlugin,EPlugin>& g,
           Vertex u, Vertex v);

  template<class Vertex, class RepS, class VPlugin, class EPlugin>
  typename bidir_adj_list<RepS,VPlugin,EPlugin>::size_type
  in_degree(Vertex u, const bidir_adj_list<RepS,VPlugin,EPlugin>& G);
    
  template <class Vertex, class RepS, class VPlugin, class EPlugin>
  void 
  remove_edge(bidir_adj_list<RepS,VPlugin,EPlugin>& g, Vertex u, Vertex v);

  template <class Edge, class RepS, class VPlugin, class EPlugin>
  void 
  remove_edge(bidir_adj_list<RepS,VPlugin,EPlugin>& g, Edge e);

  template <class RepS, class VPlugin, class EPlugin>
  typename bidir_adj_list<RepS,VPlugin,EPlugin>::size_type
  add_vertex(bidir_adj_list<RepS,VPlugin,EPlugin>& g);

  template <class RepS, class VPlugin, class EPlugin, class Vertex>
  void clear_vertex(bidir_adj_list<RepS,VPlugin,EPlugin>& g, Vertex u);

  template <class RepS, class VPlugin, class EPlugin>
  void 
  remove_vertex(bidir_adj_list<RepS,VPlugin,EPlugin>& g, std::size_t u);

} //namespace ggcl

#include "ggcl/detail/bidir_adj_list.cct"

namespace ggcl {

  template <class RS, class VP, class EP, class Tag>
  typename vertex_property_accessor< bidir_adj_list<RS,VP,EP>, Tag>::type
  get_vertex_property(bidir_adj_list<RS,VP,EP>& g, Tag)
  {
    typedef detail::vertex_pa_dispatcher<Tag> Dispatcher;
    return Dispatcher::get_vertex_property(g);
  }
  template <class RS, class VP, class EP, class Tag>
  typename vertex_property_accessor< bidir_adj_list<RS,VP,EP>, Tag>::const_type
  get_vertex_property(const bidir_adj_list<RS,VP,EP>& g, Tag)
  {
    typedef detail::vertex_pa_dispatcher<Tag> Dispatcher;
    return Dispatcher::get_vertex_property(g);
  }

  template <class RS, class VP, class EP, class Tag>
  detail::adj_list_edge_property_accessor<directed_tag, EP,Tag>
  get_edge_property(const bidir_adj_list<RS,VP,EP>&, Tag) {
    return detail::adj_list_edge_property_accessor<directed_tag,EP,Tag>();
  }


} // namespace ggcl

#endif //BIDIR_ADJ_LIST_H
