//
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Generic Graph Component Library
//
// You should have received a copy of the License Agreement for the
// Generic Graph Component Library along with the software;  see the
// file LICENSE.  If not, contact Office of Research, University of Notre
// Dame, Notre Dame, IN  46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//
#ifndef GGCL_CONNECTED_COMPONENTS_H
#define GGCL_CONNECTED_COMPONENTS_H

#include <list>
#include <deque>
#include <queue>
#include <stack>
#include <algorithm>

#include <boost/graph_structure.hpp>
#include <boost/smart_ptr.hpp>
#include <boost/tie_pair.hpp>

#include <ggcl/ggcl_config.h>
#include <ggcl/dfs.h>
#include <ggcl/dfs_visitor.h>
#include <ggcl/timestamp_visitor.h>
#include <ggcl/transpose.h>
#include <ggcl/properties.h>
#include <ggcl/indirect_cmp.h>
#include <ggcl/functor.h>
#include <ggcl/utils.h>
#include <ggcl/disjoint_sets.h>

namespace ggcl {

  // We use this visitor to record which components a vertex is
  // attributed to during the second DFS traversal in the strongly
  // connected components algorithm.
  template < class Components>
  class components_visitor
    : public base_visitor< components_visitor<Components> >
  {
  public:
    typedef every_edge_tag category;
    typedef typename boost::property_traits<Components>::value_type
      comp_type;
    components_visitor (Components _c, comp_type* _count)
      : c(_c), count(_count) {}

    template <class Vertex>
    inline void discover(Vertex u) {
      c[u] = *count;
    }
  protected:
    Components c;
    boost::shared_ptr<comp_type> count;
  };


  /* an undirected graph */
  template < class Graph, class Visitor, class Components, class Color >
  inline typename boost::property_traits<Components>::value_type
  _connected_components(Graph& G, Visitor v, 
                        Components c, Color color, undirected_tag) {
    typedef typename boost::property_traits<Components>::value_type
      count_type;
    count_type* __count = new count_type(0);

    typedef components_visitor<Components> CompVisitor;
    CompVisitor visit(c, __count);

    typedef std::pair<CompVisitor,Visitor> SuperVisitor;
    dfs_visitor<Color, SuperVisitor> visitor(color, 
                                             std::make_pair(visit,v));
    typedef typename boost::graph_traits<Graph>::vertex_descriptor Vertex;

    initialize_graph(G, visitor);

    typedef typename boost::graph_traits<Graph>::vertex_iterator Iter;
    Iter iter, end;
    for ( boost::tie(iter,end) = vertices(G); iter != end; ++iter) {
      Vertex u = *iter;
      if  (visitor.is_undiscovered(u)) {
        depth_first_visit(G, u, visitor);
        ++(*__count); 
      }
    }    
    return *__count;
  }

  /* a directed graph */
  template < class Graph, class Visitor, class Components,
             class DiscoverTime, class FinishTime,
             class Color >
  inline typename boost::property_traits<Components>::value_type
  _connected_components(Graph& G, Visitor v, 
                        Components c, DiscoverTime d, FinishTime f,Color color,
                        directed_tag)
  {
    depth_first_search(G, std::make_pair(visit_timestamp(d,f), v), color);

    Graph G_T(num_vertices(G));
    transpose(G, G_T);

    typedef components_visitor<Components> SuperVisitor;
    typedef typename boost::property_traits<Components>::value_type count_type;

    count_type* __count = new count_type(0);

    SuperVisitor visit(c, __count);

    dfs_visitor<Color, SuperVisitor> visitor(color, visit);

    typedef typename boost::graph_traits<Graph>::vertex_descriptor Vertex;

    initialize_graph(G_T, visitor);

#if defined GGCL_MSVCPP
    typedef functor_less<FinishTime> Compare;
#else
    typedef typename boost::property_traits<FinishTime>::value_type D;
    typedef indirect_cmp< FinishTime, std::less<D> > Compare;
#endif

    Compare fl(f);
    priority_queue<Vertex, std::vector<Vertex>, Compare > Q(fl);

    typedef typename boost::graph_traits<Graph>::vertex_iterator Iter;
    Iter i, j, iend, jend;
    boost::tie(i,iend) = vertices(G_T);
    boost::tie(j,jend) = vertices(G);
    for ( ; i != iend; ++i, ++j) {
      f[*i] = f[*j];
       Q.push(*i);
    }

    while ( !Q.empty() ) {
      Vertex u = Q.top();
      Q.pop();
      if  (visitor.is_undiscovered(u)) {
        depth_first_visit(G_T, u, visitor);
        ++(*__count); 
      }
    }
    return *__count;
  }

  // Variant (3)
  template < class Graph, class Components,
             class DiscoverTime, class FinishTime,
             class Color, class Visitor >
#ifdef GGCL_MSVCPP
  inline int
#else
  inline typename boost::property_traits<Components>::value_type
#endif
  connected_components(Graph& G,
                       Components c,
                       DiscoverTime d, 
                       FinishTime f,
                       Color color, Visitor v) {

    typedef typename boost::graph_traits<Graph>::directed_category Cat;
    typedef typename boost::graph_traits<Graph>::vertex_descriptor Vertex;
    Vertex u = *vertices(G).first;

    return _connected_components(G, v, 
                adapt_property_accessor(c, boost::get(c, u), u), 
                adapt_property_accessor(d, boost::get(d, u), u), 
                adapt_property_accessor(f, boost::get(f, u), u), 
                adapt_property_accessor(color,boost::get(color,u),u), Cat());
  }
 
  // Variant (2)
  template < class Graph, class Components,
             class Color, class Visitor >
#ifdef GGCL_MSVCPP
  inline int
#else
  inline typename boost::property_traits<Components>::value_type
#endif
  connected_components(Graph& G, Components c,
                       Color color, Visitor v) {
    typedef typename boost::graph_traits<Graph>::directed_category Cat;
    typedef typename boost::graph_traits<Graph>::vertex_descriptor Vertex;
    Vertex u = *vertices(G).first;
    return _connected_components(G, v, 
                      adapt_property_accessor(c, boost::get(c,u), u),
                      adapt_property_accessor(color,boost::get(color,u),u),
                      Cat());
  }

  template < class Graph, class Visitor, class Components >
#ifdef GGCL_MSVCPP
  inline int
#else
  inline typename boost::property_traits<Components>::value_type
#endif
  _connected_components(Graph& G, Visitor v, Components c, 
                       undirected_tag) 
  {
    return _connected_components(G, v, c, get_vertex_property(G, color_tag()), 
                                 undirected_tag());
  }


  template < class Graph, class Visitor, class Components >
  inline typename boost::property_traits<Components>::value_type
  _connected_components(Graph& G, Visitor v, Components c,
                        directed_tag)
  {
    return _connected_components(G, v, c, 
                                 get_vertex_property(G, discover_time_tag()),
                                 get_vertex_property(G, finish_time_tag()),
                                 get_vertex_property(G, color_tag()),
                                 directed_tag());
  }
 
  template < class Graph, class Visitor, 
             class Components,
             class Color >
  inline typename boost::property_traits<Components>::value_type
  _connected_components(Graph& G, Visitor v, Components c,
                        Color color, directed_tag)
  {
    typedef typename boost::graph_traits<Graph>::vertex_descriptor Vertex;
    return _connected_components(G, v, c, 
                                 get_vertex_property(G, discover_time_tag()),
                                 get_vertex_property(G, finish_time_tag()),
                                 color, directed_tag());
  }

  // Variant (1)
  template < class Graph, class Components, class Visitor >
#ifdef GGCL_MSVCPP
  inline int
#else
  inline typename boost::property_traits<Components>::value_type
#endif
  connected_components(Graph& G, Components c, Visitor v) {
    typedef typename boost::graph_traits<Graph>::directed_category Cat;
    typedef typename boost::graph_traits<Graph>::vertex_discriptor Vertex;
    Vertex u = *vertics(G).first;
    return _connected_components(G, v, 
         adapt_property_accessor(c,boost::get(c,u),u), Cat());
  }

  template < class Graph, class Components >
#ifdef GGCL_MSVCPP
  inline int
#else
  inline typename boost::property_traits<Components>::value_type
#endif
  connected_components(Graph& G, Components c) {
    return connected_components(G, adapt_property_accessor(c), null_visitor());
  }

  
  // A connected component algorithm for the case when dynamically
  // adding edges is common.  This is a preparing operation. Call
  // same_component to check whether two vertices are in the same
  // component, or use disjoint_set::find_set to determine
  // the representative for a vertex. 

#if 0
  template <class VertexAndEdgeListGraph, class DisjointSets> 
  void
  dynamic_connected_components(VertexAndEdgeListGraph& G,
                               DisjointSets& ds)
  {
    typedef VertexAndEdgeListGraph Graph;
    typename boost::graph_traits<Graph>::vertex_iterator ui, uiend;
    for (boost::tie(ui, uiend) = vertices(G); ui != uiend; ++ui)
      ds.make_set(*ui);
    
    typename boost::graph_traits<Graph>::edge_iterator ei, eiend;
    for ( boost::tie(ei, eiend) = edges(G); ei != eiend; ++ei) 
      ds.union_set(source(*ei,G), target(*ei,G));
  }
#endif

  template <class VertexListGraph, class DisjointSets> 
  void
  initialize_dynamic_components(VertexListGraph& G,
                                DisjointSets& ds)
  {
    typename boost::graph_traits<VertexListGraph>
      ::vertex_iterator v, vend;
    for (boost::tie(v, vend) = vertices(G); v != vend; ++v)
      ds.make_set(*v);
  }

  template <class Vertex, class DisjointSet>
  inline bool
  same_component(Vertex u, Vertex v, DisjointSet& ds) {
    return ds.find_set(u) == ds.find_set(v);
  }

  /*****   
    Next is the implementation of the EdgeList version
    of connected components adapted from the implementation
    by Alex Stepanov
    *****/

} // namespace ggcl

#include "ggcl/detail/connected_components.cct"

namespace ggcl {

// This version of connected components does not require a full
// Graph. Instead, it just needs an edge list, where the vertices of
// each edge need to be of integer type. The edges are assumed to
// be undirected. The other difference is that the result is stored in
// a container, instead of just a decorator.  The container should be
// empty before the algorithm is called. It will grow during the
// course of the algorithm. The container must be a model of
// BackInsertionSequence and RandomAccessContainer
// (std::vector is a good choice). After running the algorithm the
// index container will map each vertex to the representative
// vertex of the component to which it belongs.
//
// Adapted from an implementation by Alex Stepanov, with parts
// inspired by Tarjan's "Data Structures and Network Algorithms"

/*
  RankContainer is a random accessable container (operator[] is defined) 
  with a value type that can represent an integer part of a binary log of
  the value type of the corresponding ParentContainer (char is always enough)
  its size_type is no less than the size_type of the corresponding
  ParentContainer
  */

template <class EdgeListGraph, class DisjointSets>
void
dynamic_connected_components(EdgeListGraph& g, DisjointSets& ds)
{
  typename boost::graph_traits<EdgeListGraph>::edge_iterator e, end;
  for (boost::tie(e,end) = edges(g); e != end; ++e)
    ds.link(source(*e,g),target(*e,g));
}

template <class ParentIterator>
void
compress_components(ParentIterator first, ParentIterator last) {
  for (ParentIterator current = first; current != last; ++current) 
    detail::find_representative_with_full_compression(first, current - first);
}

template <class ParentIterator>
#ifdef GGCL_MSVCPP
ptrdiff_t
#else
typename std::iterator_traits<ParentIterator>::difference_type
#endif
component_count(ParentIterator first, ParentIterator last)
{
  ptrdiff_t count = 0;
  for (ParentIterator current = first; current != last; ++current) 
    if (*current == current - first) ++count; 
  return count;
}

// This algorithm can be applied to the result container of the
// connected_components algorithm to normalize
// the components.
template <class ParentIterator>
void normalize_components(ParentIterator first, ParentIterator last)
{
  for (ParentIterator current = first; current != last; ++current) 
    detail::normalize_node(first, current - first);
}

// considering changing the so that it initializes with a pair of
// vertex iterators and a parent PA.

template <class IndexT>
class component_index {
  //protected: (avoid friends for now)
public:
  typedef std::vector<IndexT> MyIndexContainer;
  MyIndexContainer header;
  MyIndexContainer index;
  typedef typename MyIndexContainer::size_type SizeT;
  typedef typename MyIndexContainer::const_iterator IndexIter;
public:
  typedef detail::component_iterator<IndexIter, IndexT, SizeT> 
    component_iterator;
  class component {
    friend class component_index;
  protected:
    IndexT number;
    const component_index<IndexT>* comp_ind_ptr;
    component(IndexT i, const component_index<IndexT>* p) 
      : number(i), comp_ind_ptr(p) {}
  public:
    typedef component_iterator iterator;
    typedef component_iterator const_iterator;
    typedef IndexT value_type;
    iterator begin() const {
      return iterator( comp_ind_ptr->index.begin(),
                       (comp_ind_ptr->header)[number] );
    }
    iterator end() const {
      return iterator( comp_ind_ptr->index.begin(), 
                       comp_ind_ptr->index.size() );
    }
  };
  typedef SizeT size_type;
  typedef component value_type;

  template <class Iterator>
  component_index(Iterator first, Iterator last) 
#ifdef GGCL_MSVCPP
    : index(last - first)
#else
    : index(first, last)
#endif
  { 
#ifdef GGCL_MSVCPP
    std::copy(first, last, index.begin());
#endif
    detail::construct_component_index(index, header);
  } 
  component operator[](IndexT i) const {
    return component(i, this);
  }
  SizeT size() {
    return header.size();
  }
  
};

} /*namespace ggcl*/


#endif /*GGCL_CONNECTED_COMPONENTS_H*/
