//
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Generic Graph Component Library
//
// You should have received a copy of the License Agreement for the
// Generic Graph Component Library along with the software;  see the
// file LICENSE.  If not, contact Office of Research, University of Notre
// Dame, Notre Dame, IN  46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//
#ifndef GGCL_DISJOINT_SETS_H
#define GGCL_DISJOINT_SETS_H

#include <ggcl/properties.h>
#include <ggcl/detail/disjoint_sets.cct>

namespace ggcl {

  struct representative_with_path_halving {
    template <class Parent, class Vertex>
    Vertex operator()(Parent p, Vertex v) { 
      return detail::find_representative_with_path_halving(pc, v);
    }
  };

  struct representative_with_full_path_compression {
    template <class Parent, class Vertex>
    Vertex operator()(Parent p, Vertex v){
      return detail::find_representative_with_full_compression(p, v);
    }
  };

  namespace detail {

  } // namespace detail


  // This is a generalized functor to provide disjoint sets operations
  // with "union by rank" and "path compression".  A disjoint-set data
  // structure maintains a collection S={S1, S2, ..., Sk} of disjoint
  // sets. Each set is identified by a representative, which is some
  // member of of the set. Sets are represented by rooted trees. Two
  // heuristics: "union by rank" and "path compression" are used to
  // speed up the operations.

  // Disjoint Set requires two vertex properties for internal use.  A
  // Rank and a Parent. The Rank must map Vertex to some Integral type
  // (preferably the size_type associated with Vertex). The Parent
  // must map Vertex to Vertex.
  template <class Rank, class Parent,
    class FindCompress = representative_with_full_path_compression
    >
  class disjoint_sets {
    typedef disjoint_sets self;
    
    inline disjoint_sets() {}
  public:
    inline disjoint_sets(Rank r, Parent p) 
      : rank(r), parent(p) {}

    inline disjoint_sets(const self& c) 
      : rank(c.rank), parent(c.parent) {}
    
    // Make Set -- Create a singleton set containing vertex x
    template <class Vertex>
    inline void make_set(Vertex x)
    {
      parent[x] = x;
      rank[x]   = 0;
    }
    
    // Link - union the two sets represented by vertex x and y
    template <class Vertex>
    inline void link(Vertex x, Vertex y)
    {
      detail::link_sets(parent, rank, x, y, rep);
    }
    
    // Union-Set - union the two sets containing vertex x and y 
    template <class Vertex>
    inline void union_set(Vertex x, Vertex y)
    {
      link(find_set(x), find_set(y));
    }
    
    // Find-Set - returns the Vertex representative of the set
    // containing Vertex x and applies path compression.
    template <class Vertex>
    inline Vertex find_set(Vertex x)
    {
      return rep(parent, x);
    }

    template <class VertexIterator>
    inline std::size_t count_sets(VertexIterator first, VertexIterator last)
    {
      std::size_t count = 0;  
      for ( ; first != last; ++first)
      if (parent[*first] == *first)
        ++count;
      return count;
    }

    template <class VertexIterator>
    inline void normalize_sets(VertexIterator first, VertexIterator last)
    {
      for (; first != last; ++first) 
        detail::normalize_node(parent, *first);
    }    
    
    template <class VertexIterator>
    inline void compress_sets(VertexIterator first, VertexIterator last)
    {
      for (; first != last; ++first) 
        detail::find_representative_with_full_compression(parent, *first);
    }    
  protected:
    Rank rank;
    Parent parent;
    FindCompress rep;
  };


  

  template <class ID = id_property_accessor,
            class InverseID = id_property_accessor,
            class FindCompress = representative_with_full_path_compression
            >
  class disjoint_sets_with_storage
  {
    typedef typename boost::property_traits<ID>::value_type Index;
    typedef std::vector<Index> ParentContainer;
    typedef std::vector<unsigned char> RankContainer;
  public:
    typedef typename ParentContainer::size_type size_type;

    disjoint_sets_with_storage(size_type n = 0,
                               ID id_ = ID(),
                               InverseID inv = InverseID())
      : id(id_), id_to_vertex(inv), rank(n, 0), parent(n)
    {
      for (Index i = 0; i < n; ++i)
        parent[i] = i;
    }
    // note this is not normally needed
    template <class Vertex>
    inline void 
    make_set(Vertex x) {
      parent[x] = x;
      rank[x]   = 0;
    }
    template <class Vertex>
    inline void 
    link(Vertex x, Vertex y)
    {
      extend_sets(x,y);
      detail::link_sets(parent.begin(), rank.begin(), id[x], id[y], rep);
    }
    template <class Vertex>
    inline void 
    union_set(Vertex x, Vertex y) {
      Vertex rx = find_set(x);
      Vertex ry = find_set(y);
      link(rx, ry);
    }
    template <class Vertex>
    inline Vertex find_set(Vertex x) {
      return id_to_vertex[rep(parent.begin(), id[x])];
    }

    template <class VertexIterator>
    inline std::size_t count_sets(VertexIterator first, VertexIterator last)
    {
      std::size_t count = 0;  
      for ( ; first != last; ++first)
      if (parent[*first] == *first)
        ++count;
      return count;
    }

    template <class VertexIterator>
    inline void normalize_sets(VertexIterator first, VertexIterator last)
    {
      for (; first != last; ++first) 
        detail::normalize_node(parent.begin(), *first);
    }    
    
    template <class VertexIterator>
    inline void compress_sets(VertexIterator first, VertexIterator last)
    {
      for (; first != last; ++first) 
        detail::find_representative_with_full_compression(parent.begin(),
                                                          *first);
    }    

    const ParentContainer& parents() { return parent; }

  protected:

    template <class Vertex>
    inline void 
    extend_sets(Vertex x, Vertex y)
    {
      Index needed = id[x] > id[y] ? id[x] + 1 : id[y] + 1;
      if (needed > parent.size()) {
        rank.insert(rank.end(), needed - rank.size(), 0);
        for (Index k = parent.size(); k < needed; ++k)
        parent.push_back(k);
      } 
    }

    ID id;
    InverseID id_to_vertex;
    RankContainer rank;
    ParentContainer parent;
    FindCompress rep;
  };

} /*namespace ggcl*/

#endif /*GGCL_DISJOINT_SETS_H*/
