#ifndef GGCL_DYNAMIC_GRAPH_H
#define GGCL_DYNAMIC_GRAPH_H

#include <boost/iterator.hpp>
#include <boost/operators.hpp>
#include <ggcl/properties.h>
#include <vector>
#include <ggcl/graph_rep_gen.h> // for allow_parallel_edge_tag
#include <ggcl/detail/dynamic_graph.cct>

  // how about allocators?

namespace ggcl {

  namespace detail {
#ifdef GGCL_PARTIAL_SPECIALIZATION
    template <class DirectedTag, class EdgePlugin>
    struct dynamic_edge_storage;
#else
    template <class DirectedTag>
    struct dynamic_edge_storage;
#endif
  } // namespace detail

  struct dynamic_graph_tag { };

  template <class DirectedS = directedS,
            class VertexPlugin = no_plugin,
            class EdgePlugin = no_plugin>
  class dynamic_graph : public detail::dynamic_graph_base
  {
  public:
    typedef typename directed_category_traits<DirectedS>
      ::directed_category directed_category;
    typedef detail::dyn_edge_descriptor<directed_category> edge_descriptor;
    typedef detail::dynamic_vertex_base* vertex_descriptor;
    typedef detail::dynamic_edge_iterator<directed_category> out_edge_iterator;
    typedef allow_parallel_edge_tag edge_parallel_category;
    typedef detail::dynamic_vertex_iterator vertex_iterator;
    typedef detail::dynamic_edge_iterator<directed_category> edge_iterator;
    typedef detail::dynamic_adjacency_iterator adjacency_iterator;
    typedef void in_edge_iterator;
    typedef detail::dynamic_graph_base::size_type size_type;

    typedef dynamic_graph_tag graph_tag;
    typedef EdgePlugin edge_plugin_type;
    typedef VertexPlugin vertex_plugin_type;

    // If graph is undirected, put edge plugins in a separate
    // list.
#ifdef GGCL_PARTIAL_SPECIALIZATION
    typedef typename detail::dynamic_edge_storage<directed_category,EdgePlugin>
      EdgeStore;
#else
    typedef typename detail::dynamic_edge_storage<directed_category>
      ::GGCL_TEMPLATE bind<EdgePlugin> EdgeStore;
#endif
    typedef typename EdgeStore::container_type EdgeStorage;
    typedef typename EdgeStore::on_edge_type OnEdgeType;

    inline dynamic_graph(size_type n = 0) {
      for (size_type i = 0; i < n; ++i)
        add_vertex(*this);
    }
    template <class EdgeIterator>
    inline dynamic_graph(size_type n, EdgeIterator first, EdgeIterator last)
    {
      std::vector<detail::dynamic_vertex_base*> verts(n);
      for (size_type i = 0; i < n; ++i)
        verts[i] = add_vertex(*this);

      while (first != last) {
	add_edge(*this, verts[(*first).first], verts[(*first).second]);
	++first;
      } 
    }
    EdgeStorage m_edge_plugins;
  };


  struct dynamic_graph_edge_property_selector {
    template <class Directed, class Plugin, class PropertyTag>
    struct bind {
      typedef detail::dynamic_edge_property_accessor
        <Directed, Plugin,PropertyTag> type;
      typedef type const_type;
    };
  };
  template <>  
  struct edge_property_selector<dynamic_graph_tag> {
    typedef dynamic_graph_edge_property_selector type;
  };


  struct dynamic_graph_vertex_property_selector {
    template <class Plugin, class PropertyTag>
    struct bind {
      typedef detail::dynamic_vertex_property_accessor
        <Plugin,PropertyTag> type;
      typedef type const_type;
    };
  };
  template <>  
  struct vertex_property_selector<dynamic_graph_tag> {
    typedef dynamic_graph_vertex_property_selector type;
  };


  inline detail::dynamic_vertex_base*
  source(detail::dynamic_edge_base* e, const detail::dynamic_graph_base& )
  {
    return e->m_source;
  }

  inline detail::dynamic_vertex_base*
  target(detail::dynamic_edge_base* e, const detail::dynamic_graph_base& )
  {
    return e->m_target;
  }

  template <class DS, class VP, class EP>
  inline
  std::pair<typename dynamic_graph<DS,VP,EP>::edge_iterator, 
            typename dynamic_graph<DS,VP,EP>::edge_iterator >
  out_edges(detail::dynamic_vertex_base* v, 
	    const dynamic_graph<DS,VP,EP>& )
  {  
    typedef typename dynamic_graph<DS,VP,EP>::edge_iterator Iter;
    return std::make_pair(Iter(v->m_out_edges_start),
                          Iter(v->m_out_edges_finish));
  }

  inline
  std::pair<detail::dynamic_adjacency_iterator, 
            detail::dynamic_adjacency_iterator>
  adj(detail::dynamic_vertex_base* v, const detail::dynamic_graph_base& )
  {  
    typedef detail::dynamic_adjacency_iterator Iter;
    return std::make_pair(Iter(v->m_out_edges_start),
                          Iter(v->m_out_edges_finish));
  }

  inline
  std::pair<detail::dynamic_vertex_iterator, detail::dynamic_vertex_iterator>
  vertices(const detail::dynamic_graph_base& g)
  {  
    typedef detail::dynamic_vertex_iterator Iter;
    return std::make_pair(Iter(g.m_vertices_head.m_next), 
                          Iter(&g.m_vertices_head));
  }

  template <class DS, class VP, class EP>
  inline 
  std::pair<typename dynamic_graph<DS,VP,EP>::edge_iterator,
            typename dynamic_graph<DS,VP,EP>::edge_iterator>
  edges(const dynamic_graph<DS,VP,EP>& g)
  {  
    typedef typename dynamic_graph<DS,VP,EP>::edge_iterator Iter;
    return std::make_pair(Iter(g.m_edges_head.m_next), 
                          Iter(&g.m_edges_head));
  }

  inline detail::dynamic_graph_base::size_type
  num_vertices(const detail::dynamic_graph_base& g)
  {
    return g.m_nvertices;
  }

  inline detail::dynamic_graph_base::size_type
  num_edges(const detail::dynamic_graph_base& g)
  {
    return g.m_nedges;
  }

  template <class DS, class VP, class EP, class Tag>
  typename vertex_property_accessor<dynamic_graph<DS,VP,EP>,Tag>::type
  get_vertex_property(dynamic_graph<DS,VP,EP>& g, Tag) {
    typedef typename vertex_property_accessor<dynamic_graph<DS,VP,EP>,Tag>
      ::type PA;
    return PA();
  }
  template <class DS, class VP, class EP, class Tag>
  typename edge_property_accessor<dynamic_graph<DS,VP,EP>,Tag>::type
  get_edge_property(dynamic_graph<DS,VP,EP>& g, Tag) {
    typedef typename edge_property_accessor<dynamic_graph<DS,VP,EP>,Tag>
      ::type PA;
    return PA();
  }

  template <class DS, class VP, class EP>
  std::pair<detail::dynamic_edge<EP>*, bool>
  add_edge(dynamic_graph<DS,VP,EP>& g, 
           detail::dynamic_vertex_base* u, 
           detail::dynamic_vertex_base* v)
  {
    detail::dynamic_edge<EP>* e = new detail::dynamic_edge<EP>(u,v);
    detail::add_edge(g, u, v, e);
    return std::make_pair(e, true);
  }

  template <class DS, class VP, class EP>
  std::pair<detail::dynamic_edge<EP>*, bool>
  add_edge(dynamic_graph<DS,VP,EP>& g,
           detail::dynamic_vertex_base* u,
           detail::dynamic_vertex_base* v,
           const EP& x)
  {
    typedef typename dynamic_graph<DS,VP,EP>::directed_category Cat;
    return detail::add_edge_dispatch(g, u, v, x, Cat());
  }

  // this needs some help for undirected case... need to delete more?
  template <class DS, class VP, class EP>
  void remove_edge(dynamic_graph<DS,VP,EP>& g,
                   detail::dynamic_vertex_base* u,
                   detail::dynamic_vertex_base* v)
  {
    detail::dynamic_remove_edge(g,u,v);
  }

  template <class DS, class VP, class EP>
  void remove_edge(dynamic_graph<DS,VP,EP>& g,
                   detail::dynamic_edge_base* e)
  {
    detail::dynamic_remove_edge(g,e);
    detail::dynamic_edge<EP>* de = static_cast<detail::dynamic_edge<EP>*>(e);
    delete de;
  }

  template <class DS, class VP, class EP>
  detail::dynamic_vertex_base*
  add_vertex(dynamic_graph<DS,VP,EP>& g)
  {
    detail::dynamic_vertex_base* v = new detail::dynamic_vertex<VP>();
    detail::add_vertex(g, v);
    return v;
  }

  template <class DS, class VP, class EP>
  detail::dynamic_vertex_base*
  add_vertex(dynamic_graph<DS,VP,EP>& g, const VP& x)
  {
    detail::dynamic_vertex_base* v = new detail::dynamic_vertex<VP>(x);    
    detail::add_vertex(g, v);
    return v;
  }

  template <class DS, class VP, class EP>
  void clear_vertex(dynamic_graph<DS,VP,EP>& g, detail::dynamic_vertex_base* x)
  {
    typedef detail::dynamic_edge<EP> Edge;
    typename dynamic_graph<DS,VP,EP>::edge_iterator ei, ei_end, e;
    boost::tie(ei,ei_end) = edges(g);
    while (ei != ei_end) {
      e = ei++;
      if (target(*e,g) == x || source(*e,g) == x)
        remove_edge(g, static_cast<Edge*>((*e).m_edge));
    }
  }

  template <class DS, class VP, class EP>
  void remove_vertex(dynamic_graph<DS,VP,EP>& g, 
                     detail::dynamic_vertex_base* x)
  {
    x->m_next->m_prev = x->m_prev;
    x->m_prev->m_next = x->m_next;
    detail::dynamic_vertex<VP>* v
      = static_cast<detail::dynamic_vertex<VP>*>(x);
    --g.m_nvertices;
    delete v;
  }

} // namespace ggcl


#endif /* GGCL_DYNAMIC_GRAPH_H */
