//
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Generic Graph Component Library
//
// You should have received a copy of the License Agreement for the
// Generic Graph Component Library along with the software;  see the
// file LICENSE.  If not, contact Office of Research, University of Notre
// Dame, Notre Dame, IN  46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//
#ifndef FOUNDATION_VISITOR_AUX_H
#define FOUNDATION_VISITOR_AUX_H

#include <ggcl/meta_if.h>

namespace ggcl {
  enum { FALSE_TAG, TRUE_TAG, EVERY_EDGE, TREE_EDGE, BACK_EDGE, 
	 FORWARD_EDGE, CROSS_EDGE, FORWARD_OR_CROSS_EDGE, 
	 BACK_OR_CROSS_EDGE };
  struct true_tag { enum { num = TRUE_TAG }; };
  struct false_tag { enum { num = FALSE_TAG }; };
  struct every_edge_tag { enum { num = EVERY_EDGE }; };
  struct tree_edge_tag { enum { num = TREE_EDGE }; };
  struct back_edge_tag { enum { num = BACK_EDGE }; };
  struct forward_edge_tag { enum { num = FORWARD_EDGE}; };
  struct cross_edge_tag {enum { num = CROSS_EDGE}; };
  struct forward_or_cross_edge_tag { enum { num = FORWARD_OR_CROSS_EDGE }; };
  struct back_or_cross_edge_tag { enum { num = BACK_OR_CROSS_EDGE}; };
  
#ifdef GGCL_PARTIAL_SPECIALIZATION
  template <class U, class V>
  struct is_same {
    typedef false_tag is_same_tag; 
  };
  template <class U>
  struct is_same<U, U> {
    typedef true_tag is_same_tag;
  };
#else
  template <class U, class V>
  struct is_same {
    enum { Unum = U::num, Vnum = V::num };
    typedef typename IF< (Unum == Vnum),
      true_tag, false_tag>::RET is_same_tag;
  };
#endif

template <class Visitor, class Rest, class Vertex>
inline void 
visitor_initialize_evaluate (std::pair<Visitor, Rest>& vlist, Vertex u) {
  vlist.first.initialize(u);
  visitor_initialize_evaluate(vlist.second, u);
}

#if (defined GGCL_MSVCPP) 

template <class Visitor, class Vertex>
inline void 
visitor_initialize_evaluate (base_visitor<Visitor>& vis, Vertex u) {
  Visitor& v = static_cast<Visitor&>(vis);
  v.initialize(u);
}

#else

template <class Visitor, class Vertex>
inline void 
visitor_initialize_evaluate (Visitor& v, Vertex u) {
  v.initialize(u);
}

#endif

template <class Visitor, class Rest, class Vertex>
inline void 
visitor_start_evaluate (std::pair<Visitor, Rest>& vlist, Vertex u) {
  vlist.first.start(u);
  visitor_start_evaluate(vlist.second, u);
}

#if (defined GGCL_MSVCPP) 

template <class Visitor, class Vertex>
inline void 
visitor_start_evaluate (base_visitor<Visitor>& vis, Vertex u) {
  Visitor& v = static_cast<Visitor&>(vis);
  v.start(u);
}

#else

template <class Visitor, class Vertex>
inline void 
visitor_start_evaluate (Visitor& v, Vertex u) {
  v.start(u);
}

#endif

template <class Visitor, class Rest, class Vertex>
inline void 
visitor_discover_evaluate (std::pair<Visitor, Rest>& vlist, Vertex u) {
  vlist.first.discover(u);
  visitor_discover_evaluate(vlist.second, u);
}

#if (defined GGCL_MSVCPP) 

template <class Visitor, class Vertex>
inline void 
visitor_discover_evaluate (base_visitor<Visitor>& vis, Vertex u) {
  Visitor& v = static_cast<Visitor&>(vis);
  v.discover(u);
}

#else

template <class Visitor, class Vertex>
inline void 
visitor_discover_evaluate (Visitor& v, Vertex u) {
  v.discover(u);
}

#endif

template <class Visitor, class Rest, class Vertex>
inline void 
visitor_finish_evaluate (std::pair<Visitor, Rest>& vlist, Vertex u) {
  vlist.first.finish(u);
  visitor_finish_evaluate(vlist.second, u);
}

#if (defined GGCL_MSVCPP) 

template <class Visitor, class Vertex>
inline void 
visitor_finish_evaluate (base_visitor<Visitor>& vis, Vertex u) {
  Visitor& v = static_cast<Visitor&>(vis);
  v.finish(u);
}

#else

template <class Visitor, class Vertex>
inline void 
visitor_finish_evaluate (Visitor& v, Vertex u) {
  v.finish(u);
}

#endif

template <class Visitor, class Rest, class Edge, class Graph, class Category>
inline void
visitor_process_evaluate(std::pair<Visitor,Rest>& vlist, Edge e,
			 Graph& g, Category) {
  typedef typename Visitor::category Cat;
  typedef typename is_same<Category, Cat>::is_same_tag IsSameTag;
  process_dispatch(vlist.first, e, g, IsSameTag() );
  visitor_process_evaluate(vlist.second, e, g, Category());
}

#if (defined GGCL_MSVCPP) 

template <class Visitor, class Edge, class Graph, class Category>
inline void
visitor_process_evaluate(base_visitor<Visitor>& vis, Edge e, 
			 Graph& g, Category) {
  typedef typename Visitor::category Cat;
  typedef typename is_same<Category, Cat>::is_same_tag IsSameTag;

  Visitor& v = static_cast<Visitor&>(vis);
  process_dispatch(v, e, g, IsSameTag() );
}

#else

template <class Visitor, class Edge, class Graph, class Category>
inline void
visitor_process_evaluate(Visitor& v, Edge e, Graph& g, Category) {
  typedef typename Visitor::category Cat;
  typedef typename is_same<Category, Cat>::is_same_tag IsSameTag;
  process_dispatch(v, e, g, IsSameTag() );
}

#endif

template <class Visitor, class Edge, class Graph>
inline void
process_dispatch(Visitor& v, Edge e, Graph& g, true_tag) {
    v.process(e, g);
}

template <class Visitor, class Edge, class Graph>
inline void
process_dispatch(Visitor& , Edge , Graph& , false_tag) {
}

}

#endif
