//
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Generic Graph Component Library
//
// You should have received a copy of the License Agreement for the
// Generic Graph Component Library along with the software;  see the
// file LICENSE.  If not, contact Office of Research, University of Notre
// Dame, Notre Dame, IN  46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//
#ifndef GGCL_CONCEPTS_H
#define GGCL_CONCEPTS_H

#include <boost/graph_structure.hpp>
#include <boost/property_accessor.hpp>
#include <ggcl/properties.h>
#include <ggcl/concept_checks.h>

// where to put out_degree(v,g)?

namespace ggcl {

  template <class T>
  struct MultiPassInputIterator_concept {
    CLASS_REQUIRES(T, InputIterator);
    void constraints() { }
  };

  // not sure about iterator requirements...
  // add directed_category in a couple places

  template <class G>
  struct Graph_concept
  {
    typedef typename boost::graph_traits<G>
      ::vertex_descriptor vertex_descriptor;
    typedef typename boost::graph_traits<G>
      ::edge_descriptor edge_descriptor;
    CLASS_REQUIRES(vertex_descriptor, DefaultConstructible);
    CLASS_REQUIRES(vertex_descriptor, EqualityComparable);
    CLASS_REQUIRES(vertex_descriptor, Assignable);
    CLASS_REQUIRES(edge_descriptor, DefaultConstructible);
    CLASS_REQUIRES(edge_descriptor, EqualityComparable);
    CLASS_REQUIRES(edge_descriptor, Assignable);
    typedef typename boost::graph_traits<G>::directed_category category;
    void constraints() { }
  };

  template <class G>
  struct IncidenceGraph_concept
  {
    CLASS_REQUIRES(G, Graph);
    typedef typename boost::graph_traits<G>::out_edge_iterator
      out_edge_iterator;
    CLASS_REQUIRES(out_edge_iterator, MultiPassInputIterator);
    void constraints() {
      p = out_edges(v, g);
      v = target(e, g);
    }
    void const_constraints(const G& g) {
      p = out_edges(v, g);
      v = target(e, g);
    }
    std::pair<out_edge_iterator,out_edge_iterator> p;
    typename boost::graph_traits<G>::vertex_descriptor v;
    typename boost::graph_traits<G>::edge_descriptor e;
    G g;
  };

  template <class G>
  struct BidirectionalGraph_concept
  {
    CLASS_REQUIRES(G, IncidenceGraph);
    typedef typename boost::graph_traits<G>::in_edge_iterator
      in_edge_iterator;
    CLASS_REQUIRES(in_edge_iterator, MultiPassInputIterator);
    void constraints() {
      p = in_edges(v, g);
      const_constraints(g);
    }
    void const_constraints(const G& g) {
      p = in_edges(v, g);
    }
    std::pair<in_edge_iterator,in_edge_iterator> p;
    typename boost::graph_traits<G>::vertex_descriptor v;
    G g;
  };

  template <class G>
  struct AdjacencyGraph_concept
  {
    CLASS_REQUIRES(G, IncidenceGraph);
    typedef typename boost::graph_traits<G>::adjacency_iterator
      adjacency_iterator;
    CLASS_REQUIRES(adjacency_iterator, MultiPassInputIterator);
    void constraints() {
      p = adj(v, g);
      const_constraints(g);
    }
    void const_constraints(const G& g) {
      p = adj(v, g);
    }
    std::pair<adjacency_iterator,adjacency_iterator> p;
    typename boost::graph_traits<G>::vertex_descriptor v;
    G g;
  };

  template <class G>
  struct VertexListGraph_concept
  {
    CLASS_REQUIRES(G, AdjacencyGraph);
    typedef typename boost::graph_traits<G>::vertex_iterator vertex_iterator;
    CLASS_REQUIRES(vertex_iterator, MultiPassInputIterator);
    void constraints() {
      p = vertices(g);
      const_constraints(g);
    }
    void const_constraints(const G& g) {
      p = vertices(g);
    }
    std::pair<vertex_iterator,vertex_iterator> p;
    G g;
  };

  template <class G>
  struct EdgeListGraph_concept
  {
    CLASS_REQUIRES(G, Graph);
    typedef typename boost::graph_traits<G>::edge_iterator edge_iterator;
    CLASS_REQUIRES(edge_iterator, MultiPassInputIterator);
    void constraints() {
      p = edges(g);
      u = source(e, g);
      v = target(e, g);
      const_constraints(g);
    }
    void const_constraints(const G& g) {
      p = edges(g);
      u = source(e, g);
      v = target(e, g);
    }
    std::pair<edge_iterator,edge_iterator> p;
    typename boost::graph_traits<G>::vertex_descriptor u, v;
    typename boost::graph_traits<G>::edge_descriptor e;
    G g;
  };

  template <class G>
  struct VertexAndEdgeListGraph_concept
  {
    CLASS_REQUIRES(G, VertexListGraph);
    CLASS_REQUIRES(G, EdgeListGraph);
    void constraints() { }
  };

  template <class G>
  struct MutableGraph_concept
  {
    typedef typename boost::graph_traits<G>::edge_descriptor edge_descriptor;
    void constraints() {
      v = add_vertex(g);
      clear_vertex(g,v);
      p = add_edge(g,u,v);
      remove_edge(g,u,v);
      remove_edge(g,e);
    }
    G g;
    edge_descriptor e;
    std::pair<edge_descriptor,bool> p;
    typename boost::graph_traits<G>::vertex_descriptor u, v;
  };

  template <class G>
  struct AdjacencyMatrix_concept
  {
    CLASS_REQUIRES(G, Graph);
    typedef typename boost::graph_traits<G>::edge_descriptor edge_descriptor;
    void constraints() {
      p = edge(u, v, g);
      const_constraints(g);
    }
    void const_constraints(const G& g) {
      p = edge(u, v, g);
    }
    typename boost::graph_traits<G>::vertex_descriptor u, v;
    std::pair<edge_descriptor,bool> p;
    G g;
  };

  template <class PA, class Key>
  struct ReadablePropertyAccessor_concept
  {
    typedef typename boost::property_traits<PA>::key_type key_type;
    typedef typename boost::property_traits<PA>::category Category;
    typedef boost::readable_property_accessor_tag ReadableTag;
    CLASS_REQUIRES2(Category, ReadableTag, Convertible);
    void constraints() {
      val = boost::get(pa, k);
    }
    PA pa;
    Key k;
    typename boost::property_traits<PA>::value_type val;
  };

  template <class PA, class Key>
  struct WritablePropertyAccessor_concept
  {
    typedef typename boost::property_traits<PA>::key_type key_type;
    typedef typename boost::property_traits<PA>::category Category;
    typedef boost::writable_property_accessor_tag WritableTag;
    CLASS_REQUIRES2(Category, WritableTag, Convertible);
    void constraints() {
      boost::put(pa, k, val);
    }
    PA pa;
    Key k;
    typename boost::property_traits<PA>::value_type val;
  };

  template <class PA, class Key>
  struct ReadWritePropertyAccessor_concept
  {
    CLASS_REQUIRES2(PA, Key, ReadablePropertyAccessor);
    CLASS_REQUIRES2(PA, Key, WritablePropertyAccessor);
    typedef typename boost::property_traits<PA>::category Category;
    typedef boost::read_write_property_accessor_tag ReadWriteTag;
    CLASS_REQUIRES2(Category, ReadWriteTag, Convertible);
    void constraints() { }
  };

  template <class PA, class Key>
  struct LvaluePropertyAccessor_concept
  {
    CLASS_REQUIRES2(PA, Key, ReadWritePropertyAccessor);
    typedef typename boost::property_traits<PA>::category Category;
    typedef boost::lvalue_property_accessor_tag LvalueTag;
    CLASS_REQUIRES2(Category, LvalueTag, Convertible);
    void constraints() { 
      val = boost::at(pa, k);
      boost::at(pa, k) = val;
    }
    PA pa;
    Key k;
    typename boost::property_traits<PA>::value_type val;
  };

  template <class G, class Tag>
  struct VertexPropertyGraph_concept
  {
    CLASS_REQUIRES(G, Graph);
    typedef typename boost::graph_traits<G>::vertex_descriptor Vertex;
    typedef typename vertex_property_accessor<G,Tag>::type PA;
    typedef typename vertex_property_accessor<G,Tag>::const_type const_PA;
    CLASS_REQUIRES2(PA, Vertex, ReadWritePropertyAccessor);
    CLASS_REQUIRES2(const_PA, Vertex, ReadablePropertyAccessor);
    void constraints() {
      PA pa = get_vertex_property(g, Tag());
      ignore_unused_variable_warning(pa);
    }
    void const_constraints(const G& g) {
      const_PA pa = get_vertex_property(g, Tag());
      ignore_unused_variable_warning(pa);
    }
    G g;
  };

  template <class G, class Tag>
  struct EdgePropertyGraph_concept
  {
    CLASS_REQUIRES(G, Graph);
    typedef typename boost::graph_traits<G>::edge_descriptor Edge;
    typedef typename edge_property_accessor<G,Tag>::type PA;
    typedef typename edge_property_accessor<G,Tag>::const_type const_PA;
    CLASS_REQUIRES2(PA, Edge, ReadWritePropertyAccessor);
    CLASS_REQUIRES2(const_PA, Edge, ReadablePropertyAccessor);
    void constraints() {
      PA pa = get_edge_property(g, Tag());
    }
    void const_constraints(const G& g) {
      const_PA pa = get_edge_property(g, Tag());
    }
    G g;
  };

  template <class V, class Vertex, class Edge, class Graph>
  struct Visitor_concept
  {
    void constraints() {
      v.initialize(u);
      v.start(u);
      v.discover(u);
      bool b = v.process(e, g);
      v.finish(u);
      ignore_unused_variable_warning(b);
    }
    V v;
    Vertex u;
    Edge e;
    Graph g;
  };

  template <class V, class Vertex, class Edge, class Graph>
  struct UserVisitor_concept
  {
#ifdef GGCL_PARTIAL_SPECIALIZATION
    typedef typename V::category Category;
    void constraints() {
      v.initialize(u);
      v.start(u);
      v.discover(u);
      v.process(e, g);
      v.finish(u);
    }
    V v;
    Vertex u;
    Edge e;
    Graph g;
#else
    void constraints() { }
#endif
  };

#ifdef GGCL_PARTIAL_SPECIALIZATION
  template <class V1, class V2, class Vertex, class Edge, class Graph>
  struct UserVisitor_concept< std::pair<V1,V2>, Vertex,Edge,Graph>
  {
    CLASS_REQUIRES4(V1, Vertex, Edge, Graph, UserVisitor);
    CLASS_REQUIRES4(V2, Vertex, Edge, Graph, UserVisitor);
    void constraints() { }
  };
#endif

  template <class B>
  struct Buffer_concept
  {
    void constraints() {
      b.push(t);
      b.pop();
      typename B::value_type& v = b.top();
      const_constraints(b);
      ignore_unused_variable_warning(v);
    }
    void const_constraints(const B& b) {
      const typename B::value_type& v = b.top();
      n = b.size();
      bool e = b.empty();
      ignore_unused_variable_warning(v);
      ignore_unused_variable_warning(e);
    }
    typename B::size_type n;
    typename B::value_type t;
    B b;
  };

  template <class C>
  struct ColorValue_concept
  {
    CLASS_REQUIRES(C, EqualityComparable);
    void constraints() {
      c = white(c);
      c = gray(c);
      c = black(c);
    }
    C c;
  };

} // namespace ggcl

#endif /* GGCL_CONCEPTS_H */
