//
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Generic Graph Component Library
//
// You should have received a copy of the License Agreement for the
// Generic Graph Component Library along with the software;  see the
// file LICENSE.  If not, contact Office of Research, University of Notre
// Dame, Notre Dame, IN  46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//
#ifndef GGCL_MST_KRUSKAL_H
#define GGCL_MST_KRUSKAL_H

/*
 *Minimum Spanning Tree 
 *         Kruskal Algorithm
 *
 *Requirement:
 *      undirected graph
 */

#include <vector>
#include <queue>
#include <functional>
#include <ggcl/functor.h>
#include <ggcl/disjoint_sets.h>
#include <ggcl/properties.h>
#include <ggcl/indirect_cmp.h>
#include <ggcl/ggcl_concepts.h>

namespace ggcl {


  // Kruskal's algorithm for Minimum Spanning Tree
  //
  // This is a greedy algorithm to calculate the Minimum Spanning Tree
  // for an undirected graph with weighted edges. The output will be a
  // set of edges.
  //
  template < class Graph, class OutputIterator, 
             class Rank, class Parent, class Weight >
  inline void _kruskal_minimum_spanning_tree(Graph& G, OutputIterator out, 
                                             Rank rank, Parent p, Weight w)
  {
    typedef typename boost::graph_traits<Graph>::vertex_descriptor Vertex;
    typedef typename boost::graph_traits<Graph>::edge_descriptor Edge;
    REQUIRE(Graph, VertexAndEdgeListGraph);
    REQUIRE2(OutputIterator, Edge, OutputIterator);
    REQUIRE2(Rank, Vertex, ReadWritePropertyAccessor);
    REQUIRE2(Parent, Vertex, ReadWritePropertyAccessor);
    REQUIRE2(Weight, Edge, ReadablePropertyAccessor);
    typedef typename boost::property_traits<Weight>::value_type W_value;
    typedef typename boost::property_traits<Rank>::value_type R_value;
    typedef typename boost::property_traits<Parent>::value_type P_value;
    REQUIRE(W_value, LessThanComparable);
    REQUIRE2(P_value, Vertex, Convertible);
    REQUIRE(R_value, Integer);

    disjoint_sets<Rank, Parent>  dset(rank, p);
    
    typename boost::graph_traits<Graph>::vertex_iterator ui, uiend;
    for (boost::tie(ui, uiend) = vertices(G); ui != uiend; ++ui)
      dset.make_set(*ui);

#if defined GGCL_MSVCPP
    typedef functor_greater<Weight> weight_greater;
#else 
    typedef typename boost::property_traits<Weight>::value_type W;
    typedef indirect_cmp<Weight, std::greater<W> > weight_greater;
#endif

    weight_greater wl(w);

    typedef std::priority_queue < Edge, 
      std::vector<Edge>, weight_greater> Queue;

    Queue Q(wl);

    /*push all edge into Q*/
    typename boost::graph_traits<Graph>::edge_iterator ei, eiend;
    for (boost::tie(ei,eiend) = edges(G); ei != eiend; ++ei) 
      Q.push(*ei);
    
    while ( !Q.empty() ) {
      Edge e = Q.top();
      Q.pop();
      Vertex u = dset.find_set(source(e, G));
      Vertex v = dset.find_set(target(e, G));
      if ( u != v ) {
        *out++ = e;
        dset.link(u, v);
      }
    }
  }

  // Variant (2)
  template < class Graph, class OutputIterator, 
             class Rank, class Parent, class Weight >
  inline void kruskal_minimum_spanning_tree(Graph& G, OutputIterator out, 
                                            Rank rank, Parent p, Weight w)
  {
    typedef typename boost::graph_traits<Graph>::vertex_descriptor Vertex;
    Vertex u = *vertices(G).first;
    typedef typename boost::graph_traits<Graph>::edge_descriptor Edge;
    Edge e = *out_edges(u, G).first;

    _kruskal_minimum_spanning_tree(G, out, 
                           adapt_property_accessor(rank,boost::get(rank,u),u),
                           adapt_property_accessor(p,boost::get(p,u),u),
                           adapt_property_accessor(w,boost::get(w,e),e));
  }

  // Variant (1)
  template < class Graph, class OutputIterator, 
             class Rank, class Parent >
  inline void kruskal_minimum_spanning_tree(Graph& G, OutputIterator c, 
                                            Rank rank, Parent p ) {
    typedef typename boost::graph_traits<Graph>::edge_descriptor Edge;
    kruskal_minimum_spanning_tree(G, c, rank, p, 
				  get_edge_property(G, weight_tag()));
  }

} /*namespace*/


#endif /*GGCL_MST_KRUSKAL_H*/

