//
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Generic Graph Component Library
//
// You should have received a copy of the License Agreement for the
// Generic Graph Component Library along with the software;  see the
// file LICENSE.  If not, contact Office of Research, University of Notre
// Dame, Notre Dame, IN  46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//
#ifndef GGCL_PROPERTIES_H
#define GGCL_PROPERTIES_H

#include <iterator>
#include <boost/graph_structure.hpp>
#include <boost/property_accessor.hpp>
#include <ggcl/meta_if.h>
#include <ggcl/meta_equal.h>
#include <ggcl/utils.h>

namespace boost {
  template <class T, class PA>
  struct put_get_at_helper { };

  template <class PA, class T, class K>
  inline T get(const put_get_at_helper<T,PA>& pa, const K& k) {
    T v = static_cast<const PA&>(pa)[k];
    return v;
  }
  template <class PA, class T, class K, class V>
  inline void put(put_get_at_helper<T,PA>& pa, K k, const V& v) {
    static_cast<PA&>(pa)[k] = v;
  }
  template <class PA, class T, class K>
  inline T& at(put_get_at_helper<T,PA>& pa, K k) {
    T& v = static_cast<PA&>(pa)[k];
    return v;
  }
}

namespace ggcl {

  enum ePropertyAccessorID { OP_BRACKET_PA = boost::LAST_PA+1, 
			     RAND_ACCESS_ITER_PA };

  // ggcl additions to handle operator bracket and random access iterators
  struct readable_op_bracket_tag
    : public boost::readable_property_accessor_tag
  { enum { id = OP_BRACKET_PA }; };
  struct writable_op_bracket_tag
    : public boost::writable_property_accessor_tag
  { enum { id = OP_BRACKET_PA }; };
  struct read_write_op_bracket_tag
    : public boost::read_write_property_accessor_tag 
  { enum { id = OP_BRACKET_PA }; };
  struct lvalue_op_bracket_tag
    : public boost::lvalue_property_accessor_tag
  { enum { id = OP_BRACKET_PA }; };
  struct random_access_iterator_pa_tag
    : public boost::lvalue_property_accessor_tag
  { enum { id = RAND_ACCESS_ITER_PA }; };    

}

#include <ggcl/detail/properties.cct>
#include <ggcl/detail/pa2dec.cct>

namespace boost {
#ifndef GGCL_PARTIAL_SPECIALIZATION
  // The user will just have to create their own specializations for
  // other pointers types if the compiler does not have partial
  // specializations. Sorry!
  template <>
  struct property_traits<int*> {
    typedef int value_type;
    typedef ptrdiff_t key_type;
    typedef ggcl::random_access_iterator_pa_tag   category;
  };
  template <>
  struct property_traits<std::size_t*> {
    typedef std::size_t value_type;
    typedef ptrdiff_t key_type;
    typedef ggcl::random_access_iterator_pa_tag   category;
  };
#else
  template <class T>
  struct property_traits<T*> {
    typedef typename std::iterator_traits<T*>::value_type value_type;
    typedef ptrdiff_t key_type;
    typedef ggcl::random_access_iterator_pa_tag category;
  };
#endif

  template <class T>
  inline void put(T* pa, ptrdiff_t k, const T& val) { pa[k] = val;  }

  template <class T>
  inline const T& get(const T* pa, ptrdiff_t k) { return pa[k]; }

  template <class T>
  inline T& at(T* pa, ptrdiff_t k) { return pa[k]; }

} // namespace boost

namespace ggcl {


  class dummy_property_accessor 
    : public boost::put_get_at_helper< int, dummy_property_accessor  > 
  {
  public:
    typedef void          key_type; 
    typedef int    value_type;
    typedef lvalue_op_bracket_tag category;
    inline dummy_property_accessor() : c(0) { }
    inline dummy_property_accessor(value_type cc) : c(cc) { }
    inline dummy_property_accessor(const dummy_property_accessor& x)
      : c(x.c) { }
    template <class Vertex>
    inline value_type& operator[](Vertex) { return c; }
    template <class Vertex>
    inline const value_type& operator[](Vertex)  const { return c; }
   protected:
    value_type c;
  };

  struct id_property_accessor
    : public boost::put_get_at_helper<std::size_t, id_property_accessor>
  {
    typedef void key_type;
    typedef std::size_t value_type; // ? -JGS
    typedef readable_op_bracket_tag category;

    template <class Vertex>
    inline Vertex operator[](Vertex v) const { return v; }
  };

  template <class RandomAccessIterator, 
#ifdef GGCL_MSVCPP
     class T, class R,
#else
     class T = typename std::iterator_traits<RandomAccessIterator>::value_type,
     class R = typename std::iterator_traits<RandomAccessIterator>::reference,
#endif
     class IDfunc = id_property_accessor>
  class random_access_iterator_property_accessor 
    : public boost::put_get_at_helper< T, 
        random_access_iterator_property_accessor<RandomAccessIterator,
        T, R, IDfunc> >
  {
  public:
    typedef void key_type; 
    typedef T value_type;
    typedef lvalue_op_bracket_tag category;

    inline random_access_iterator_property_accessor(
      RandomAccessIterator cc = RandomAccessIterator(), 
      const IDfunc& _id = IDfunc() ) 
      : iter(cc), id(_id) { }
    inline random_access_iterator_property_accessor(
      const random_access_iterator_property_accessor& x) 
      : iter(x.iter), id(x.id) { }

    template <class Key>
    inline R operator[](Key v) const { return *(iter + id[v]) ; }
  protected:
    RandomAccessIterator iter;
    IDfunc id;
  };

#ifdef GGCL_PARTIAL_SPECIALIZATION
  template <class RAIter, class ID>
  inline random_access_iterator_property_accessor<
    RAIter,
    typename std::iterator_traits<RAIter>::value_type,
    typename std::iterator_traits<RAIter>::reference,
    ID >
  create_random_access_iter_pa(RAIter iter, ID id) {
    typedef random_access_iterator_property_accessor<
    RAIter,
    typename std::iterator_traits<RAIter>::value_type,
    typename std::iterator_traits<RAIter>::reference,
    ID >    PA;
    return PA(iter, id);
  }
#else
  // add VC++ workarounds... -JGS
#endif

  template <class PA, class T, class K>
  inline detail::pa2dec_adaptor<PA, T, K>
  adapt_property_accessor(PA pa, T, K) {
    return detail::pa2dec_adaptor<PA, T, K>(pa);
  }

  struct foo_edge_property_selector {
    template <class Directed, class Plugin, class Tag>
    struct bind {
      typedef void type;
      typedef void const_type;
    };
  };
  template <class GraphTag>
  struct edge_property_selector {
    typedef foo_edge_property_selector type;
  };

  struct foo_vertex_property_selector {
    template <class Plugin, class Tag>
    struct bind {
      typedef void type;
      typedef void const_type;
    };
  };
  template <class GraphTag>
  struct vertex_property_selector {
    typedef foo_vertex_property_selector type;
  };

  template <class Graph, class PropertyTag>
  struct edge_property_accessor {
    typedef typename Graph::directed_category Directed;
    typedef typename Graph::edge_plugin_type Plugin;
    typedef typename Graph::graph_tag graph_tag;
    typedef typename edge_property_selector<graph_tag>::type Selector;
    typedef typename Selector::GGCL_TEMPLATE bind<Directed,Plugin,PropertyTag>
      Bind;
    typedef typename Bind::type type;
    typedef typename Bind::const_type const_type;
  };
  template <class Graph, class PropertyTag>
  struct vertex_property_accessor {
    typedef typename Graph::vertex_plugin_type Plugin;
    typedef typename Graph::graph_tag graph_tag;
    typedef typename vertex_property_selector<graph_tag>::type Selector;
    typedef typename Selector::GGCL_TEMPLATE bind<Plugin,PropertyTag> Bind;
    typedef typename Bind::type type;
    typedef typename Bind::const_type const_type;
  };

  struct no_plugin { 
    enum { num = -1 };
    typedef no_plugin tag_type;
    typedef no_plugin next_type;
    typedef no_plugin value_type;
  };

  template <class Tag, class T, class Base = no_plugin>
  struct plugin : public Base {
    typedef Base next_type;
    typedef Tag tag_type;
    typedef T value_type;
    plugin() { }
    plugin(const T& v, const Base& b = Base()) : Base(b), _value(v) { }
    T _value;
  };

  // Wild, I thought I could make a simple member functions plugin::get(Tag)
  // but the inheritance/overloading didn't work. Instead I had
  // to do this recursive thing, traversing up the plugin inheritance
  // chain. Getting the return type right was a bit tricky. -JGS
  namespace detail {

    template <int TagMatched>
    struct plugin_value_dispatch { };

    template <>
    struct plugin_value_dispatch<1> {
      template <class Plugin, class T, class Tag>
      static T& get_value(Plugin& p, T, Tag) {
	return p._value; 
      }
      template <class Plugin, class T, class Tag>
      static const T& const_get_value(const Plugin& p, T, Tag) {
	return p._value; 
      }
    };
    template <>
    struct plugin_value_dispatch<0> {
      template <class Plugin, class T, class Tag>
      static T& get_value(Plugin& p, T t, Tag tag) {
	typedef typename Plugin::next_type Next;
	typedef typename Next::tag_type Next_tag;
	enum { match = int(Next_tag::num) == int(Tag::num) };
	return plugin_value_dispatch<match>
	  ::get_value(static_cast<Next&>(p), t, tag);
      }
      template <class Plugin, class T, class Tag>
      static const T& const_get_value(const Plugin& p, T t, Tag tag) {
	typedef typename Plugin::next_type Next;
	typedef typename Next::tag_type Next_tag;
	enum { match = int(Next_tag::num) == int(Tag::num) };
	return plugin_value_dispatch<match>
	  ::get_value(static_cast<const Next&>(p), t, tag);
      }
    };

  } // namespace detail

  template <class Tag1, class Tag2, class T1, class T2, class Base>
  T2& get_plugin_value(plugin<Tag1,T1,Base>& p, T2 t2, Tag2 tag2) {
    enum { match = int(Tag1::num) == int(Tag2::num) };
    typedef detail::plugin_value_dispatch<match> Dispatcher;
    return Dispatcher::get_value(p, t2, tag2);
  }
  template <class Tag1, class Tag2, class T1, class T2, class Base>
  const T2& get_plugin_value(const plugin<Tag1,T1,Base>& p, T2 t2, Tag2 tag2) {
    enum { match = int(Tag1::num) == int(Tag2::num) };
    typedef detail::plugin_value_dispatch<match> Dispatcher;
    return Dispatcher::const_get_value(p, t2, tag2);
  }


  // The enum's are only necessary for a workaround for compilers that
  // don't do partial specialization (like VC++).

  struct id_tag {
    enum { num = detail::ID_PLUGIN_TAG };
  };
  struct name_tag { 
    enum { num = detail::NAME_PLUGIN_TAG };
  };
  struct weight_tag { 
    enum { num = detail::WEIGHT_PLUGIN_TAG };
  };
  struct distance_tag { 
    enum { num = detail::DISTANCE_PLUGIN_TAG };
  };
  struct color_tag { 
    enum { num = detail::COLOR_PLUGIN_TAG };
  };
  struct degree_tag { 
    enum { num = detail::DEGREE_PLUGIN_TAG };
  };
  struct out_degree_tag { 
    enum { num = detail::OUT_DEGREE_PLUGIN_TAG };
  };
  struct in_degree_tag { 
    enum { num = detail::IN_DEGREE_PLUGIN_TAG };
  };
  struct discover_time_tag { 
    enum { num = detail::DISCOVER_TIME_PLUGIN_TAG };
  };
  struct finish_time_tag { 
    enum { num = detail::FINISH_TIME_PLUGIN_TAG };
  };

} // namespace ggcl


#endif /* GGCL_PROPERTIES_H */



