//
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Generic Graph Component Library
//
// You should have received a copy of the License Agreement for the
// Generic Graph Component Library along with the software;  see the
// file LICENSE.  If not, contact Office of Research, University of Notre
// Dame, Notre Dame, IN  46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//
#ifndef GGCL_UTILS_H
#define GGCL_UTILS_H

#include <stdlib.h>
#include <iosfwd>
#include <assert.h>
#include <ggcl/ggcl_config.h>
#ifdef GGCL_HAVE_SLIST
#include <slist> // shouldn't have to include this... -JGS
#endif
#include <boost/graph_structure.hpp>
#include <ggcl/properties.h>
#include <ggcl/directed_category_traits.h>

#ifdef GGCL_MSVCPP
#include <vector>
#include <list>
#include <map>
#include <set>
#else
namespace std {
  template <class T, class Alloc> class vector;
#ifdef GGCL_HAVE_SLIST
  template <class T, class Alloc> class slist;
#endif
  template <class T, class Alloc> class list;
  template <class Key, class Cmp, class Alloc> class set;
  template <class Key, class T, class Cmp, class Alloc> class map;
  template <class Key, class Eq, class Hash, class Alloc> class hash_set;
  template <class Key, class T, class Eq, class Hash, class Alloc>
  class hash_map;
}
#endif

#if defined GGCL_HAVE_LIMITS
#include <limits>
#else
#include <contrib/ggcl_limits.h>
#endif

#include <ggcl/base_visitor.h>

namespace ggcl {

  struct id_property_accessor;

  typedef short default_color_type;

  inline default_color_type white(default_color_type) { return 0; }

  inline default_color_type gray(default_color_type) { return 1; }

  inline default_color_type black(default_color_type) { return -1; }

  template <class T>
  inline T numeric_limits_max(T) { return std::numeric_limits<T>::max(); }

  struct every_edge_tag;
  //
  // This is a visitor to provide only the standard visitor interface.
  // All methods are emtpy. 
  //
  class null_visitor : public base_visitor<null_visitor> {
  public:
    typedef every_edge_tag category;
    inline null_visitor() { }
    inline null_visitor(const null_visitor& x) { }
  };

#if 0
#if defined GGCL_PARTIAL_SPECIALIZATION
  template <typename SizeType, class EdgePlugin>
  struct elemtype_traits {
    typedef std::pair<SizeType, EdgePlugin> type;
  };

  struct no_plugin;

  template <typename SizeType>
  struct elemtype_traits <SizeType, no_plugin> {
    typedef SizeType type;
  };
  template <typename SizeType>
  struct elemtype_traits <SizeType, no_plugin*> {
    typedef SizeType type;
  };
  template <typename SizeType>
  struct elemtype_traits <SizeType, const no_plugin*> {
    typedef SizeType type;
  };
#else

  // This is horribly convoluted thanks to VC++
  struct elemtype_pair_maker {
    template <class EdgePlugin, typename SizeType>
    struct bind { typedef std::pair<SizeType, EdgePlugin> type; };
  };
  struct elemtype_sizet_maker {
    template <class EdgePlugin, typename SizeType>
    struct bind { typedef SizeType type; };
  };

  template <class EdgePlugin>
  struct elemtype_helper {
    typedef elemtype_pair_maker type;
  };
  template <>
  struct elemtype_helper<no_plugin> {
    typedef elemtype_sizet_maker type;
  };
  template <>
  struct elemtype_helper<no_plugin*> {
    typedef elemtype_sizet_maker type;
  };
  template <>
  struct elemtype_helper<const no_plugin*> {
    typedef elemtype_sizet_maker type;
  };

  template <typename SizeType, class EdgePlugin>
  struct elemtype_traits {
    typedef typename elemtype_helper<EdgePlugin>::type Selector;
    typedef typename Selector
      :: GGCL_TEMPLATE bind<EdgePlugin,SizeType>::type type;
  };
#endif
#endif
  
  template <class T, class Alloc, class V>
  std::pair<typename std::vector<T, Alloc>::iterator,bool>
  push(std::vector<T, Alloc>& c, const V& v)
  {
    c.push_back(v);
    typename std::vector<T, Alloc>::iterator i = c.end();
    --i;
    return std::make_pair(i,true);
  }
   
  template <class T, class Alloc, class V>
  std::pair<typename std::list<T, Alloc>::iterator,bool>
  push(std::list<T, Alloc>& c, const V& v)
  {
    c.push_front(v);
    return std::make_pair(c.begin(), true);
  }

#ifdef GGCL_HAVE_SLIST
  template <class T, class Alloc, class V>
  std::pair<typename std::slist<T, Alloc>::iterator,bool>
  push(std::slist<T, Alloc>& c, const V& v)
  {
    c.push_front(v);
    return std::make_pair(c.begin(), true);
  }
#endif

  template <class Key, class Value_type, class Compare, class Alloc>
  std::pair<typename std::map<Key, Value_type, Compare, Alloc>::iterator,bool>
  push(std::map<Key, Value_type, Compare, Alloc>& c,
       const std::pair<Key, Value_type>& v)
  {
    return c.insert(v);
  }

  template <class Key, class Compare, class Alloc, class ValueType>
  std::pair<typename std::set<Key, Compare, Alloc>::iterator,bool>
  push(std::set<Key, Compare, Alloc>& c, const ValueType& v)
  {
    return c.insert(v);
  }

#ifdef GGCL_HAVE_HASH
  template <class Key, class Value_type, class Hash, class Equal, class Alloc>
  std::pair<
    typename std::hash_map<Key, Value_type, Hash, Equal, Alloc>::iterator,
    bool>
  push(std::hash_map<Key, Value_type, Hash, Equal, Alloc>& c,
       const std::pair<Key, Value_type>& v)
  {
    return c.insert(v);
  }

  template <class Key, class Hash, class Eq, class Alloc>
  std::pair<typename std::hash_set<Key, Hash, Eq, Alloc>::iterator,bool>
  push(std::hash_set<Key, Hash, Eq, Alloc>& c, const Key& v)
  {
    return c.insert(v);
  }
#endif

  //...... for other containers


  template <class IncidenceGraph, class Name>
  void print_in_edges(const IncidenceGraph& G, Name name)
  {
    typename boost::graph_traits<IncidenceGraph>::vertex_iterator ui,ui_end;
    for (boost::tie(ui,ui_end) = vertices(G); ui != ui_end; ++ui) {
      std::cout << name[*ui] << " <-- " << std::flush;
      typename boost::graph_traits<IncidenceGraph>
        ::in_edge_iterator ei, ei_end;
      for(boost::tie(ei,ei_end) = in_edges(*ui,G); ei != ei_end; ++ei)
	std::cout << name[source(*ei,G)] << " " << std::flush;
      std::cout << std::endl;
    }
  }

  template <class IncidenceGraph, class Name>
  void print_graph_dispatch(const IncidenceGraph& G, Name name, directed_tag)
  {
    typename boost::graph_traits<IncidenceGraph>::vertex_iterator ui,ui_end;
    for (boost::tie(ui,ui_end) = vertices(G); ui != ui_end; ++ui) {
      std::cout << name[*ui] << " --> " << std::flush;
      typename boost::graph_traits<IncidenceGraph>
        ::out_edge_iterator ei, ei_end;
      for(boost::tie(ei,ei_end) = out_edges(*ui,G); ei != ei_end; ++ei)
	std::cout << name[target(*ei,G)] << " " << std::flush;
      std::cout << std::endl;
    }
  }
  template <class IncidenceGraph, class Name>
  void print_graph_dispatch(const IncidenceGraph& G, Name name, undirected_tag)
  {
    typename boost::graph_traits<IncidenceGraph>::vertex_iterator ui,ui_end;
    for (boost::tie(ui,ui_end) = vertices(G); ui != ui_end; ++ui) {
      std::cout << name[*ui] << " <--> " << std::flush;
      typename boost::graph_traits<IncidenceGraph>
        ::out_edge_iterator ei, ei_end;
      for(boost::tie(ei,ei_end) = out_edges(*ui,G); ei != ei_end; ++ei)
	std::cout << name[target(*ei,G)] << " " << std::flush;
      std::cout << std::endl;
    }
  }
  template <class IncidenceGraph, class Name>
  void print_graph(const IncidenceGraph& G, Name name)
  {
    typedef typename boost::graph_traits<IncidenceGraph>
      ::directed_category Cat;
    print_graph_dispatch(G, name, Cat());
  }
  template <class IncidenceGraph>
  void print_graph(const IncidenceGraph& G) {
    print_graph(G, get_vertex_property(G, id_tag()));
  }

  template <class EdgeListGraph, class Name>
  void print_edges(const EdgeListGraph& G, Name name)
  {
    typename boost::graph_traits<EdgeListGraph>::edge_iterator ei,ei_end;
    for (boost::tie(ei,ei_end) = edges(G); ei != ei_end; ++ei)
      std::cout << "(" << name[source(*ei,G)]
		<< "," << name[target(*ei,G)] << ") ";
    std::cout << std::endl;
  }

  template <class VertexListGraph, class Name>
  void print_vertices(const VertexListGraph& G, Name name)
  {
    typename boost::graph_traits<VertexListGraph>::vertex_iterator vi,vi_end;
    for (boost::tie(vi,vi_end) = vertices(G); vi != vi_end; ++vi)
      std::cout << name[*vi] << " ";
    std::cout << std::endl;
  }

  template <class _ForwardIter, class _Tp>
  void iota(_ForwardIter __first, _ForwardIter __last, _Tp __value)
  {
    while (__first != __last)
      *__first++ = __value++;
  }

  inline std::size_t random_number(std::size_t N) {
    std::size_t ret = rand() % N; 
    return ret;
  }

  // grab a random vertex from the graph's vertex set
  template <class Graph>
  typename boost::graph_traits<Graph>::vertex_descriptor
  random_vertex(Graph& g) {
    std::size_t n = random_number(num_vertices(g));
    typename boost::graph_traits<Graph>::vertex_iterator
      i = vertices(g).first;
#ifdef GGCL_MSVCPP
    while (n-- > 0) ++i;
#else
    std::advance(i, n);
#endif
    return *i;
  }

  template <class Graph>
  typename boost::graph_traits<Graph>::edge_descriptor
  random_edge(Graph& g) {
    std::size_t n = random_number(num_edges(g));
    typename boost::graph_traits<Graph>::edge_iterator
      i = edges(g).first;
#ifdef GGCL_MSVCPP
    while (n-- > 0) ++i;
#else
    std::advance(i, n);
#endif
    return *i;
  }

  template <class Graph, class Vertex>
  bool is_adj(Graph& g, Vertex a, Vertex b) {
    typename Graph::adjacency_iterator vi, viend, found;
    boost::tie(vi, viend) = adj(a, g);
    found = find(vi, viend, b);
    if ( found == viend )
      return false;  
    return true;
  }

  template <class Graph, class Edge>
  bool in_edge_set(Graph& g, Edge e)
  {
    typename Graph::edge_iterator ei, ei_end, found;
    boost::tie(ei, ei_end) = edges(g);
    found = std::find(ei, ei_end, e);
    return found != ei_end;
  }

  template <class Graph, class Vertex>
  bool in_vertex_set(Graph& g, Vertex v)
  {
    typename Graph::vertex_iterator vi, vi_end, found;
    boost::tie(vi, vi_end) = vertices(g);
    found = std::find(vi, vi_end, v);
    return found != vi_end;
  }

  template <class Graph, class Vertex>
  bool in_edge_set(Graph& g, Vertex u, Vertex v)
  {
    typename Graph::edge_iterator ei, ei_end;
    for (boost::tie(ei,ei_end) = edges(g); ei != ei_end; ++ei)
      if (source(*ei,g) == u && target(*ei,g) == v)
	return true;
    return false;
  }


} /* namespace ggcl */

#endif /* GGCL_UTILS_H */
