//
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Generic Graph Component Library
//
// You should have received a copy of the License Agreement for the
// Generic Graph Component Library along with the software;  see the
// file LICENSE.  If not, contact Office of Research, University of Notre
// Dame, Notre Dame, IN  46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//
#ifndef GGCL_VERTEX_PLUGIN_H
#define GGCL_VERTEX_PLUGIN_H

#include <vector>

#include <ggcl/properties.h>
#include <ggcl/utils.h>

namespace ggcl {

  // This class is the default stored vertex plugin and will be the
  // base of other off_vertex plugin.
  //
  template < class SizeType = size_t > 
  class off_vertex_default_plugin {

    typedef off_vertex_default_plugin self;
    
  public:
    typedef SizeType size_type;

    inline off_vertex_default_plugin() {}
    inline off_vertex_default_plugin(size_type n) {}
    
    /*the following 4 functions used during constructing a graph or
      dynamically change edges and vertices
    */

    inline void 
    add_vertex() {
    }

  protected:
  };
  typedef off_vertex_default_plugin<size_t> no_vertex_plugin;

  
  // the off-vertex distance plugin.
  //
  // This is the off-vertex distance plugin to provide the storage of
  // vertex distance properity inside a graph object. Use this for
  // adjacency_list and adjacency_matrix graphs. Use
  // distance_decorator to access the distance properity.
#if defined GGCL_MSVCPP
  template < class StoragePlugin = off_vertex_default_plugin<>, class T = int> 
//, class Container = std::vector< StoragePlugin::size_type> >
  class distance_plugin : public StoragePlugin {
    typedef std::vector< T > Container;
#else
  template < class StoragePlugin = off_vertex_default_plugin<>, class T = int, class Container = std::vector<T> >
  class distance_plugin : public StoragePlugin {
#endif

    typedef distance_plugin self;
  public:

    typedef typename Container::size_type size_type;
    typedef typename Container::value_type distance_type;

    inline distance_plugin(size_type n) : StoragePlugin(n), dis(n) {}

    inline distance_type& distance(size_type u) {
      return dis[u];    
    }

    inline const distance_type& distance(size_type u) const {
      return dis[u];    
    }

    inline void 
    add_vertex() {
      dis.push_back(std::numeric_limits<distance_type>::max());
      StoragePlugin::add_vertex();
    }
  protected:
    Container dis;
  };

  // the off-vertex color plugin
  //
  // This is the off-vertex plugin to provide the storage of
  // vertex color properity inside a graph object. Use this for
  // adjacency_list and adjacency_matrix graphs. Use
  // color_decorator to access the distance properity.
  //
#if defined GGCL_MSVCPP
  template < class StoragePlugin = off_vertex_default_plugin<> >
  class color_plugin : public StoragePlugin {
	typedef std::vector<default_color_type> Container;
#else
  template < class StoragePlugin = off_vertex_default_plugin<>, class Container = std::vector<default_color_type> >
  class color_plugin : public StoragePlugin {
#endif
    typedef color_plugin self;
  public:
    typedef typename StoragePlugin::size_type size_type;
    typedef typename Container::value_type color_type;

    inline color_plugin(size_type n) : StoragePlugin(n), col(n) {}

    inline color_type& color(size_type u) {
      return col[u];    
    }

    inline const color_type& color(size_type u) const {
      return col[u];    
    }

   inline void 
    add_vertex() {
     col.push_back(color_type());
     StoragePlugin::add_vertex();
    }
  protected:
    Container col;
  };

  // the off-vertex out degree plugin
  //
  // This is the off-vertex plugin to provide the storage of
  // vertex out degree properity inside a graph object. Use this for
  // adjacency_list and adjacency_matrix graphs. Use
  // out_degree_decorator to access the distance properity.
  //
#if defined GGCL_MSVCPP
  template < class StoragePlugin = off_vertex_default_plugin<>, class T=size_t>
  class out_degree_plugin : public StoragePlugin {
    typedef std::vector<T> Container;
  public:
    typedef size_t out_degree_type;
#else
  template < class StoragePlugin = off_vertex_default_plugin<>, class T = size_t, class Container = std::vector<T> >
  class out_degree_plugin : public StoragePlugin {
  public:
    typedef typename Container::value_type out_degree_type;
#endif
    typedef out_degree_plugin self;
    typedef typename StoragePlugin::size_type size_type;

    inline out_degree_plugin(size_type n) 
      : StoragePlugin(n), deg(n, 0)  {}

    inline const out_degree_type& out_degree(size_type u) const {
      return deg[u];
    }

    inline out_degree_type& out_degree(size_type u) {
      return deg[u];
    }
    
    inline void 
    add_vertex() {
      deg.push_back(out_degree_type());
      StoragePlugin::add_vertex();
    }

  protected:
    Container deg;
  };

  // the off-vertex in degree plugin
  //
  // This is the off-vertex plugin to provide the storage of
  // vertex in degree properity inside a graph object. Use this for
  // adjacency_list and adjacency_matrix graphs. Use
  // in_degree_decorator to access the distance properity.
  //
#if defined GGCL_MSVCPP
  template < class StoragePlugin = off_vertex_default_plugin<>, class T=size_t>
  class in_degree_plugin : public StoragePlugin {
    typedef std::vector< T > Container;
#else
  template < class StoragePlugin = off_vertex_default_plugin<>, class T=size_t, class Container = std::vector<T> >
  class in_degree_plugin : public StoragePlugin {
#endif
    typedef in_degree_plugin self;
  public:
    typedef typename StoragePlugin::size_type size_type;
    typedef typename Container::value_type in_degree_type;

    inline in_degree_plugin(size_type n) 
      : StoragePlugin(n), deg(n, 0)  {}

    inline const in_degree_type& in_degree(size_type u) const {
      return deg[u];
    }

    inline in_degree_type& in_degree(size_type u) {
      return deg[u];
    }
    
    inline void 
    add_vertex() {
      deg.push_back(in_degree_type());
      StoragePlugin::add_vertex();
    }

  protected:
    Container deg;
  };

  // the off-vertex degree plugin
  //
  // This is the off-vertex plugin to provide the storage of
  // vertex degree properity inside a graph object. Use this for
  // adjacency_list and adjacency_matrix graphs. Use
  // degree_decorator to access the distance properity.
  //
#if defined GGCL_MSVCPP
  template < class StoragePlugin = off_vertex_default_plugin<>, class T=size_t>
  class degree_plugin : public StoragePlugin {
    typedef std::vector<T> Container;
#else
  template < class StoragePlugin = off_vertex_default_plugin<>, class T=size_t, class Container = std::vector<T> >
  class degree_plugin : public StoragePlugin {
#endif
    typedef degree_plugin self;
  public:
    typedef typename StoragePlugin::size_type size_type;
    typedef typename Container::value_type degree_type;

    inline degree_plugin() {}
    inline degree_plugin(size_type n) 
      : StoragePlugin(n), deg(n, 0) {}

    inline const degree_type& degree(size_type u) const {
      return deg[u];
    }

    inline degree_type& degree(size_type u) {
      return deg[u];
    }
    
    inline void 
    add_vertex() {
      deg.push_back(degree_type());
      StoragePlugin::add_vertex();
    }

  protected:
    Container deg;
  };

  // the off-vertex discover-time plugin
  //
  // This is the off-vertex plugin to provide the storage of vertex
  // discover-time properity inside a graph object. Use this for
  // adjacency_list and adjacency_matrix graphs. Use
  // discover_time_decorator to access the distance properity.
  //
#if defined GGCL_MSVCPP
  template < class StoragePlugin = off_vertex_default_plugin<>, class T=int>
  class discover_time_plugin : public StoragePlugin {
    typedef std::vector<T> Container;
#else
  template < class StoragePlugin = off_vertex_default_plugin<>, class T=int, class Container = std::vector<T> >
  class discover_time_plugin : public StoragePlugin {
#endif
    typedef discover_time_plugin self;
  public:
    typedef typename StoragePlugin::size_type size_type;
    typedef typename Container::value_type discover_time_type;

    inline discover_time_plugin() {}
    inline discover_time_plugin(size_type n) 
      : StoragePlugin(n), disc(n, 0) {}

    inline const discover_time_type& discover_time(size_type u) const {
      return disc[u];
    }

    inline discover_time_type& discover_time(size_type u) {
      return disc[u];
    }

    inline void 
    add_vertex() {
      disc.push_back(std::numeric_limits<discover_time_type>::max());
      StoragePlugin::add_vertex();
    }

  protected:
    Container disc;
  };


  // the off-vertex finish-time plugin
  //
  // This is the off-vertex plugin to provide the storage of vertex
  // discover-time properity inside a graph object. Use this for
  // adjacency_list and adjacency_matrix graphs. Use discover_time_decorator
  // to access the distance properity.
  //
#if defined GGCL_MSVCPP
  template < class StoragePlugin = off_vertex_default_plugin<>, class T=int>
  class finish_time_plugin : public StoragePlugin {
    typedef std::vector<T> Container;
#else
  template < class StoragePlugin = off_vertex_default_plugin<>, class T=int, class Container = std::vector<T> >
  class finish_time_plugin : public StoragePlugin {
#endif
    typedef finish_time_plugin self;
  public:
    typedef typename StoragePlugin::size_type size_type;
    typedef typename Container::value_type finish_time_type;
    inline finish_time_plugin() {}
    inline finish_time_plugin(size_type n) 
      : StoragePlugin(n), disc(n, 0) {}

    inline const finish_time_type& finish_time(size_type u) const {
      return disc[u];
    }

    inline finish_time_type& finish_time(size_type u) {
      return disc[u];
    }

    inline void 
    add_vertex() {
      disc.push_back(std::numeric_limits<finish_time_type>::max());
      StoragePlugin::add_vertex();
    }

  protected:
    Container disc;
  };

} /*namespace*/


#endif /*GGCL_VERTEX_PLUGIN_H*/

