//
//=======================================================================
// Copyright 1997, 1998, 1999, 2000 University of Notre Dame.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// This file is part of the Generic Graph Component Library
//
// You should have received a copy of the License Agreement for the
// Generic Graph Component Library along with the software;  see the
// file LICENSE.  If not, contact Office of Research, University of Notre
// Dame, Notre Dame, IN  46556.
//
// Permission to modify the code and to distribute modified code is
// granted, provided the text of this NOTICE is retained, a notice that
// the code was modified is included with the above COPYRIGHT NOTICE and
// with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
// file is distributed with the modified code.
//
// LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
// By way of example, but not limitation, Licensor MAKES NO
// REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
// PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
// OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
// OR OTHER RIGHTS.
//=======================================================================
//
#ifndef GGCL_WEIGHTED_EDGE_VISITOR_H
#define GGCL_WEIGHTED_EDGE_VISITOR_H

#include <utility>
#include <ggcl/properties.h>
#include <ggcl/relax.h>
#include <ggcl/foundation_visitor_aux.h>

namespace ggcl {


  // Weighted Edge Visitor
  //
  // This is a generalization of the kind of visitor
  // used inside Dijkstra's and Prim's algorithms.
  // This is also used for the min-max paths problem.
  //
  // It is IMPORTANT that user has to make sure the following two
  // types are the same:
  // property_traits< DistanceDecorator >::value_type
  // property_traits< WeightDecorator >::value_type
  //
  template <class Weight, class Distance, class Color, class Base, 
            class BinaryOperator, class Cmp>
  class weighted_edge_visitor {
    typedef typename boost::property_traits<Distance>::value_type D;
  public:

    inline weighted_edge_visitor() 
      : NeedUpdateQueue(false)  { }
    inline weighted_edge_visitor(Weight wf, Distance df, Color c)
      : w(wf), d(df), color(c), NeedUpdateQueue(false) { }
    inline weighted_edge_visitor(Weight wf, Distance df, Color c,
				 const Base& b)
      : base(b), w(wf), d(df), color(c), NeedUpdateQueue(false) { }
    inline weighted_edge_visitor(Weight wf, Distance df, Color c,
				 BinaryOperator binop, 
				 Cmp compare, const Base& b)
      : base(b), w(wf), d(df), color(c), NeedUpdateQueue(false), 
	op(binop), cmp(compare)  { }

    template <class Vertex>
    inline void initialize(Vertex u) { 
      d[u] = numeric_limits_max(D()); // remove this?
      color[u] = white(color[u]);
      visitor_initialize_evaluate( base, u );
    }
    template <class Vertex>
    inline void start(Vertex u) {
      d[u] = 0;
      visitor_start_evaluate( base, u );
    }
    template <class Vertex>
    inline void discover(Vertex u) {
      visitor_discover_evaluate( base, u );
    }
    template <class Vertex>
    inline void finish(Vertex u) {
      color[u] = black(color[u]); // no longer on the queue
      visitor_finish_evaluate( base, u ); 
    }

    template <class Edge, class Graph>
    inline bool process(Edge e, Graph& g) {
      typedef typename boost::graph_traits<Graph>::vertex_descriptor Vertex;
      Vertex v = target(e,g);
      bool push_on_queue = false;
      bool decreased = relax(e, g, w, d, op, cmp);

      if (color[v] == white(color[target(e,g)])) {// previously undiscovered
	push_on_queue = true;
	color[v] = gray(color[v]);  // gray means it is on the queue
      } else if (decreased && color[v] == gray(color[v]))
	NeedUpdateQueue = true; // invoke Q.update in the visited function

      if (decreased)
	visitor_process_evaluate(base, e, g, tree_edge_tag());
      // put "else" here? (check use of copy visitor) -JGS
      visitor_process_evaluate(base, e, g, every_edge_tag());
      return push_on_queue;
    }

    inline bool need_update_queue() {
      return NeedUpdateQueue;
    }
    inline void reset_need_update() {
      NeedUpdateQueue = false;
    }
  protected:
    Base base;
    Weight w;
    Distance d;
    Color color;
    bool NeedUpdateQueue;
    BinaryOperator op;
    Cmp cmp;
  };

} /* namespace ggcl */

#endif
